/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "R4300.h"

#include "Debug/Debug.h"
#include "CPU.h"
#include "RSP.h"
#include "Interrupt.h"

#include "OSHLE/Patch.h"				// For g_PatchSymbols
#include "OSHLE/ultra_r4300.h"

#include "ROM.h"
#include "DynaRec/DynaRec.h"
#include "Core/Registers.h"			// For REG_?? defines
#include "R4300_Jump.inl"		// Jump table
#include "Debug/DBGConsole.h"
#include "Interface/MainWindow.h"

enum ERoundingMode
{
	RM_NEAR = 0,
	RM_CHOP,
	RM_UP,
	RM_DOWN,
	RM_NUM_MODES,
};
static ERoundingMode	gRoundingMode( RM_NEAR );

#ifdef DAEDALUS_PSP

inline void SET_ROUND_MODE( ERoundingMode mode )
{
	// XXXX
}

inline bool _isnan(double x)
{
	// XXXX
	return false;
}

#else

static const int		gNativeRoundingModes[ RM_NUM_MODES ] =
{
	_RC_NEAR,	// RM_NEAR,
	_RC_CHOP,	// RM_CHOP,
	_RC_UP,		// RM_UP,
	_RC_DOWN,	// RM_DOWN,
};

void SET_ROUND_MODE( ERoundingMode mode )
{
	_controlfp( gNativeRoundingModes[ mode ], _MCW_RC );
}

#endif



// Conveniently, the x87 status bits correspond to the MIPS fp status bits (shifted by 2!)
// _EM_INEXACT		0x01		FPCSR_FI		0x04
// _EM_UNDERFLOW	0x02		FPCSR_FU		0x08
// _EM_OVERFLOW		0x04		FPCSR_FO		0x10
// _EM_ZERODIVIDE	0x08		FPCSR_FZ		0x20
// _EM_INVALID		0x10		FPCSR_FV		0x40
#define X86_EM_MASK		(_EM_INEXACT|_EM_UNDERFLOW|_EM_OVERFLOW|_EM_ZERODIVIDE|_EM_INVALID)

/*
	_clearfp();

	f32	fR = op( fX, fY );

	// On S_CVT_W etc
	if ( _fpclass( fX ) & (_FPCLASS_NINF|_FPCLASS_PINF ) )
	{
		g_qwCCR[1][ 31 ]._u32[ 0 ] |= FPCSR_CE;

		R4300_Exception_FP();
		return;
	}

	u32		flags( (_statusfp() & (X86_EM_MASK)) );
	if ( flags )
	{
		if ( g_qwCCR[1][ 31 ]._u32[ 0 ] & ( flags << 7 ) )	// Compare with enable bits
		{
			g_qwCCR[1][ 31 ]._u32[ 0 ] |= flags << 12;		// Set cause bits
			R4300_Exception_FP();
			return;
		}
		else
		{
			g_qwCCR[1][ 31 ]._u32[ 0 ] |= flags << 2;		// Set flags
		}
	}
*/


inline void SpeedHack(bool bCond, OpCode op )
{
	if (bCond)
	{
		u32 dwNextOp;

		dwNextOp = *(u32 *)(g_pPCMemBase+4);

		// If nop, then this is a busy-wait for an interrupt
		if (dwNextOp == 0)
		{ 
			g_qwCPR[0][C0_COUNT]._u32[0] += (CPU_GetFirstEvent()->mCount - 1);
			CPU_GetFirstEvent()->mCount = 1;
		}
		// XXXX check this....need to update count....
		// This is:
		// 0x7f0d01e8: BNEL      v0 != v1 --> 0x7f0d01e8
		// 0x7f0d01ec: ADDIU     v0 = v0 + 0x0004
		else if (op._u32 == 0x5443ffff && dwNextOp == 0x24420004)
		{
			gGPR[REG_v0]._u64 = gGPR[REG_v1]._u64 - 4;
		}
		else
		{
			static bool warned = false;

			if (!warned)
			{
				DBGConsole_Msg(0, "Missed Speedhack 0x%08x", g_dwPC);
				warned = true;
			}
		}
	}
}


//
//	A bit on FPU exceptions.
//		LWC1/LDC1/SWC1/SDC1		are unformatted, so no exceptions raised
//		MTC1/MFC1/DMTC1/DMFC1	are unformatted, so no exceptions raised
//		Word accessed must be 4 byte aligned, DWord accesses must be 8 byte aligned
//

inline u64 LoadFPR_Long( u32 reg )
{
	if (g_qwCPR[0][C0_SR]._u32[0] & SR_FR)
	{
		return g_qwCPR[1][reg]._u64;
	}
	else
	{
		// This looks a bit crap, but VC++ seems to optimise it a lot better
		// (just movs onto the stack rather than shifts, ors etc)
		// This removes about 4/5 ops and a call to __allshl
		REG64 res;
		res._u32[1] = g_qwCPR[1][reg+1]._u32[0];
		res._u32[0] = g_qwCPR[1][reg+0]._u32[0];
		return res._u64;

		//return (g_qwCPR[1][reg+1]<<32) | (g_qwCPR[1][reg+0]&0xFFFFFFFF);
	}
}

inline void StoreFPR_Long( u32 reg, u64 value )
{
	if (g_qwCPR[0][C0_SR]._u32[0] & SR_FR)
	{
		g_qwCPR[1][reg]._u64 = value;
	}
	else
	{
		// TODO - Don't think these need sign extending...
		g_qwCPR[1][reg+0]._u32[0] = (u32)(value & 0xFFFFFFFF);
		g_qwCPR[1][reg+1]._u32[0] = (u32)(value>>32);
	}
}

__forceinline f64 LoadFPR_Double( u32 reg )
{
	u64 value = LoadFPR_Long( reg );

	return ToDouble( value );
}

__forceinline void StoreFPR_Double( u32 reg, f64 value )
{
	StoreFPR_Long( reg, ToQuad( value ) );
}

__forceinline s32 LoadFPR_Word( u32 reg )
{
	return g_qwCPR[1][reg]._s32[ 0 ];
}

__forceinline void StoreFPR_Word( u32 reg, s32 value )
{
	g_qwCPR[1][reg]._s32[ 0 ] = value;
}

__forceinline f32 LoadFPR_Single( u32 reg )
{
	return g_qwCPR[1][reg]._f32[ 0 ];
}

__forceinline void StoreFPR_Single( u32 reg, f32 value )
{
	g_qwCPR[1][reg]._f32[ 0 ] = value;
}

/* 
Might look at implementing the S/D instructions like this...
template< typename T > __forceinline T LoadFPR( u32 reg );
template< typename T > __forceinline void StoreFPR( u32 reg, T value );

template<>	f32 LoadFPR< f32 >( u32 reg )
{
	return g_qwCPR[1][reg]._f32[ 0 ];
}

template<>	f64 LoadFPR< f64 >( u32 reg )
{
	u64 value = LoadFPR_Long( reg );

	return ToDouble( value );
}

template<> __forceinline void StoreFPR< f32 >( u32 reg, f32 value )
{
	g_qwCPR[1][reg]._f32[ 0 ] = value;
}

template<> __forceinline void StoreFPR< f64 >( u32 reg, f64 value )
{
	StoreFPR_Long( reg, ToQuad( value ) );
}

template< typename T > __forceinline void R4300_Cop1_DIV( CALL_ARGS )
{
	// fd = fs/ft
	T dividend = LoadFPR< T >( gCurrentOp.fs );
	T divisor  = LoadFPR< T >( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	if ( divisor == 0 )
	{
		if ( g_qwCCR[1][ 31 ]._u32[ 0 ] & FPCSR_EZ )
		{
			// Exception
			DBGConsole_Msg( 0, "[MShould trigger FPU exception for /0 here" );
		}
		else
		{
			DBGConsole_Msg( 0, "Float divide by zero, setting flag" );
			g_qwCCR[1][ 31 ]._u32[ 0 ] |= FPCSR_FZ;
		}
	}

	StoreFPR< T >( gCurrentOp.fd, dividend / divisor );
}
*/


//*****************************************************************************
//
//	int -> float conversion routines
//
//*****************************************************************************
__forceinline f32 s32_to_f32( s32 x )
{
	return (f32)x;
}

__forceinline f64 s32_to_f64( s32 x )
{
	return (f64)x;
}

__forceinline f32 s64_to_f32( s64 x )
{
	return (f32)x;
}

__forceinline f64 s64_to_f64( s64 x )
{
	return (f64)x;
}

//*****************************************************************************
//
//	float -> float conversion routines
//
//*****************************************************************************
__forceinline f64 f32_to_f64( f32 x )
{
	return (f64)x;
}

__forceinline f32 f64_to_f32( f64 x )
{
	return (f32)x;
}


//*****************************************************************************
//
//	Float -> int conversion routines
//
// This was using the following code, but I don't think it deals
// with NaNs etc well:
//
//	s32 result;
//
//	result = (s32)x;
//
//	__asm
//	{
//		fld		x;
//		fistp	result;
//	}
//
//	return result;
//
//
//*****************************************************************************
__forceinline s32 f32_to_s32( f32 x )
{
	return (s32)x;
}

__forceinline s32 f64_to_s32( f64 x )
{
	return (s32)x;
}

__forceinline s64 f32_to_s64( f32 x )
{
	return (s64)x;
}

__forceinline s64 f64_to_s64( f64 x )
{
	return (s64)x;
}

//*****************************************************************************
//
//*****************************************************************************
void CALL_TYPE R4300_SetCop1Enable( bool enable )
{
	if ( enable )
	{
		R4300Instruction[OP_COPRO1] = R4300_CoPro1;
	}
	else
	{
		R4300Instruction[OP_COPRO1] = R4300_CoPro1_Disabled;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CALL_TYPE R4300_SetSR( u32 new_value )
{
	/*if (((u32)g_qwCPR[0][C0_SR] & SR_FR) != (new_value & SR_FR))
	{
		if (new_value & SR_FR)
			DBGConsole_Msg(0, "FP changed to 64bit register mode");
		else
			DBGConsole_Msg(0, "FP changed to 32bit register mode");
	}

	if (((u32)g_qwCPR[0][C0_SR] & SR_UX) != (new_value & SR_UX))
	{
		if (new_value & SR_UX)
			DBGConsole_Msg(0, "CPU changed to 64bit register mode");
		else
			DBGConsole_Msg(0, "CPU changed to 32bit register mode");
	}*/

	bool interrupts_enabled_before = (g_qwCPR[0][C0_SR]._u32[0] & SR_IE) != 0;

	g_qwCPR[0][C0_SR]._s64 = (s64)(s32)new_value;

	bool interrupts_enabled_after = (g_qwCPR[0][C0_SR]._u32[0] & SR_IE) != 0;
	
	R4300_SetCop1Enable( (g_qwCPR[0][C0_SR]._u64 & SR_CU1) != 0 );

	if ( !interrupts_enabled_before && interrupts_enabled_after )
	{
		if (g_qwCPR[0][C0_SR]._u32[0] & g_qwCPR[0][C0_CAUSE]._u32[0] & CAUSE_IPMASK)
		{
			//DBGConsole_Msg( 0, "[MShould be servicing pending interrupts here: %02x %02x?", 
			//	(g_qwCPR[0][C0_SR]._u32[0] & g_qwCPR[0][C0_CAUSE]._u32[0] & CAUSE_IPMASK) >> 8,
			//	Memory_MI_GetRegister(MI_INTR_MASK_REG) & Memory_MI_GetRegister(MI_INTR_REG)
			//	);
			CPU_AddJob( CPU_CHECK_INTERRUPTS );
		}
	}

}

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

#ifndef DAEDALUS_PSP
#define WARN_NOEXIST(inf) { CMainWindow::Get()->MessageBox("Instruction Unknown", inf); CPU_Halt("Instruction Unknown"); }
#define WARN_NOIMPL(op) { CMainWindow::Get()->MessageBox("Instruction Not Implemented", op); CPU_Halt("Instruction Not Implemented"); }
#else
#define WARN_NOEXIST(inf)	// XXXX
#define WARN_NOIMPL(op)
#endif 

void CALL_TYPE R4300_Unk( CALL_ARGS )     { WARN_NOEXIST("R4300_Unk"); }
void CALL_TYPE R4300_Special( CALL_ARGS ) { R4300SpecialInstruction[ gCurrentOp.funct ]( ); }
void CALL_TYPE R4300_RegImm( CALL_ARGS )  { R4300RegImmInstruction[ gCurrentOp.rt ]( );     }
void CALL_TYPE R4300_CoPro0( CALL_ARGS )  { R4300Cop0Instruction[ gCurrentOp.rs ]( );  }
void CALL_TYPE R4300_CoPro1( CALL_ARGS )  { R4300Cop1Instruction[ gCurrentOp.rs ]( );  }

void CALL_TYPE R4300_CoPro1_Disabled( CALL_ARGS )
{
	// Cop1 Unusable
	DBGConsole_Msg(0, "Thread accessing Cop1, throwing COP1 unusuable exception");

	// When HLE processing was enabled
	//R4300_SetCop1Enable(true);
	//OS_HLE_EnableThreadFP();
	//R4300Cop1Instruction[ gCurrentOp.rs ]( );

	R4300_Exception_CopUnusuable();
}


// These are the only unimplemented R4300 instructions now:
void CALL_TYPE R4300_LL( CALL_ARGS ) {  WARN_NOIMPL("LL"); }
void CALL_TYPE R4300_LLD( CALL_ARGS ) {  WARN_NOIMPL("LLD"); }

void CALL_TYPE R4300_SC( CALL_ARGS ) {  WARN_NOIMPL("SC"); }
void CALL_TYPE R4300_SCD( CALL_ARGS ) {  WARN_NOIMPL("SCD"); }


void CALL_TYPE R4300_LDC2( CALL_ARGS ) {  WARN_NOIMPL("LDC2"); }
void CALL_TYPE R4300_SDC2( CALL_ARGS ) {  WARN_NOIMPL("SDC2"); }

void CALL_TYPE R4300_RegImm_TGEI( CALL_ARGS ) {  WARN_NOIMPL("TGEI"); }
void CALL_TYPE R4300_RegImm_TGEIU( CALL_ARGS ) {  WARN_NOIMPL("TGEIU"); }
void CALL_TYPE R4300_RegImm_TLTI( CALL_ARGS ) {  WARN_NOIMPL("TLTI"); }
void CALL_TYPE R4300_RegImm_TLTIU( CALL_ARGS ) {  WARN_NOIMPL("TLTIU"); }
void CALL_TYPE R4300_RegImm_TEQI( CALL_ARGS ) {  WARN_NOIMPL("TEQI"); }
void CALL_TYPE R4300_RegImm_TNEI( CALL_ARGS ) {  WARN_NOIMPL("TNEI"); }

void CALL_TYPE R4300_RegImm_BLTZALL( CALL_ARGS ) {  WARN_NOIMPL("BLTZALL"); }
void CALL_TYPE R4300_RegImm_BGEZALL( CALL_ARGS ) {  WARN_NOIMPL("BGEZALL"); }

void CALL_TYPE R4300_Special_TGE( CALL_ARGS ) {  WARN_NOIMPL("TGE"); }
void CALL_TYPE R4300_Special_TGEU( CALL_ARGS ) {  WARN_NOIMPL("TGEU"); }
void CALL_TYPE R4300_Special_TLT( CALL_ARGS ) {  WARN_NOIMPL("TLT"); }
void CALL_TYPE R4300_Special_TLTU( CALL_ARGS ) {  WARN_NOIMPL("TLTU"); }
void CALL_TYPE R4300_Special_TEQ( CALL_ARGS ) {  WARN_NOIMPL("TEQ"); }
void CALL_TYPE R4300_Special_TNE( CALL_ARGS ) {  WARN_NOIMPL("TNE"); }


void CALL_TYPE R4300_DBG_Bkpt( CALL_ARGS )
{
#ifdef DAEDALUS_BREAKPOINTS_ENABLED
	// Entry is in lower 26 bits...
	u32 dwBreakPoint = gCurrentOp.bp_index;

	if (g_BreakPoints[dwBreakPoint].bEnabled &&
		!g_BreakPoints[dwBreakPoint].bTemporaryDisable)
	{
		// Set the temporary disable so we don't execute bp immediately again
		g_BreakPoints[dwBreakPoint].bTemporaryDisable = true;
		CPU_Halt("BreakPoint");
		DBGConsole_Msg(0, "[RBreakPoint at 0x%08x]", g_dwPC);

		// Decrement, so we move onto this instruction next
		DECREMENT_PC();
	}
	else
	{
		// If this was on, disable it
		g_BreakPoints[dwBreakPoint].bTemporaryDisable = false;

		gCurrentOp = g_BreakPoints[dwBreakPoint].mOriginalOp;
		R4300Instruction[ gCurrentOp.op ]( );
	}
#else

	DAEDALUS_ASSERTMSG( "How did we get here when breakpoints are disabled?" );

#endif
	
}
void CALL_TYPE R4300_SRHack_UnOpt( CALL_ARGS )
{
	// Entry is in lower 26 bits...
	u32 dwEntry = gCurrentOp.dynarec_index;

	CDynarecCode * pCode = DynaRec_GetEntry( dwEntry );
	
	if (g_nDelay == NO_DELAY)
	{
		DynaRec_ExecuteCode(pCode);
	}
	else
	{
		gCurrentOp = pCode->mOriginalOp;
		R4300Instruction[ gCurrentOp.op ]( );
	}

}

void CALL_TYPE R4300_SRHack_Opt( CALL_ARGS )
{
	// Entry is in lower 26 bits...
	u32 dwEntry = gCurrentOp.dynarec_index;

	CDynarecCode * pCode = DynaRec_GetEntry( dwEntry );
	
	if (g_nDelay == NO_DELAY)
	{
		DynaRec_ExecuteCode_AlreadyOptimised(pCode);
	}
	else
	{
		gCurrentOp = pCode->mOriginalOp;
		R4300Instruction[ gCurrentOp.op ]( );
	}

}


// Couldn't compile this code....
void CALL_TYPE R4300_SRHack_NoOpt( CALL_ARGS )
{
	// Entry is in lower 26 bits...
	u32 dwEntry = gCurrentOp.dynarec_index;

	CDynarecCode * pCode = DynaRec_GetEntry( dwEntry );

	gCurrentOp = pCode->mOriginalOp;

	R4300Instruction[ gCurrentOp.op ]( );
}

void CALL_TYPE R4300_Patch( CALL_ARGS )
{
	// Read patch info from g_PatchSymbols[]
	u32 dwPatchNum = gCurrentOp.patch_index;

	u32 dwRet = g_PatchSymbols[dwPatchNum]->pFunction();
	if (dwRet == PATCH_RET_JR_RA)
	{
		// TODO - We could call DynaRec_CompileCode here to compile the subsequent code
		CPU_SetPC(gGPR[REG_ra]._u32[0]);
		DECREMENT_PC();		// We increment PC after this call..
		g_nDelay = NO_DELAY;

	}
	else if (dwRet == PATCH_RET_ERET)
	{
		// Bit of a hack for now - we should have a seperate R4300_FNCion...
		R4300_TLB_ERET( );	
	}
	else
	{
		gCurrentOp = g_PatchSymbols[dwPatchNum]->opReplaced;
		R4300Instruction[ gCurrentOp.op ]( );
	}
}

void CALL_TYPE R4300_J( CALL_ARGS ) 				// Jump
{
	g_dwNewPC = (g_dwPC & 0xF0000000) | (gCurrentOp.target<<2);

	SpeedHack(g_dwNewPC == g_dwPC, gCurrentOp);
	CPU_Delay();
}

void CALL_TYPE R4300_JAL( CALL_ARGS ) 				// Jump And Link
{
	gGPR[REG_ra]._s64 = (s64)(s32)(g_dwPC + 8);		// Store return address
	g_dwNewPC = (g_dwPC & 0xF0000000) | (gCurrentOp.target<<2);


	CPU_Delay();
}

void CALL_TYPE R4300_BEQ( CALL_ARGS ) 		// Branch on Equal
{
	//branch if rs == rt
	if ( gGPR[gCurrentOp.rs]._u32[0] == gGPR[gCurrentOp.rt]._u32[0] )
	{
		s16 offset = (s16)gCurrentOp.immediate;

		SpeedHack(offset == -1, gCurrentOp);
			
		g_dwNewPC = g_dwPC + ((s32)offset<<2) + 4;
		CPU_Delay();
	}
	/*else
	{
		if (gGPR[dwRS] == gGPR[dwRT])
			DBGConsole_Msg(0, "BEQ Missed");
	}*/
}

void CALL_TYPE R4300_BNE( CALL_ARGS )             // Branch on Not Equal
{
	//branch if rs <> rt
	if ( gGPR[gCurrentOp.rs]._u32[0] != gGPR[gCurrentOp.rt]._u32[0] )
	{
		s16 offset = (s16)gCurrentOp.immediate;

		SpeedHack(offset == -1, gCurrentOp);

		g_dwNewPC = g_dwPC + ((s32)offset<<2) + 4;
		CPU_Delay();
	} 
	/*else
	{
		if (gGPR[dwRS] != gGPR[dwRT])
			DBGConsole_Msg(0, "BNE Missed");
	}*/
}

void CALL_TYPE R4300_BLEZ( CALL_ARGS ) 			// Branch on Less than of Equal to Zero
{
	//branch if rs <= 0
	//if ((s64)gGPR[gCurrentOp.rs] <= 0)
	if (gGPR[gCurrentOp.rs]._s32[0] <= 0)
	{
		s16 offset = (s16)gCurrentOp.immediate;

		SpeedHack(offset == -1, gCurrentOp);

		g_dwNewPC = g_dwPC + ((s32)offset<<2) + 4;
		CPU_Delay();

	} 
	/*else
	{
		if ((s64)gGPR[dwRS] <= 0)
			DBGConsole_Msg(0, "BLEZ Missed");
	}*/
}

void CALL_TYPE R4300_BGTZ( CALL_ARGS ) 			// Branch on Greater than Zero
{
	//branch if rs > 0
	//if ((s64)gGPR[gCurrentOp.rs] > 0)
	if (gGPR[gCurrentOp.rs]._s32[0] > 0)
	{
		s16 offset = (s16)gCurrentOp.immediate;

		SpeedHack(offset == -1, gCurrentOp);

		g_dwNewPC = g_dwPC + ((s32)offset<<2) + 4;
		CPU_Delay();

	} 
	/*else
	{
		if ((s64)gGPR[dwRS] > 0)
			DBGConsole_Msg(0, "BGTZ Missed");
	}*/
}


void CALL_TYPE R4300_DADDI( CALL_ARGS ) 			// Doubleword ADD Immediate
{	
	// Check for overflow
	// Reserved Instruction exception

	//rt = rs + immediate
	gGPR[gCurrentOp.rt]._s64 = gGPR[gCurrentOp.rs]._s64 + (s32)(s16)gCurrentOp.immediate;
	CHECK_R0();
}

void CALL_TYPE R4300_DADDIU( CALL_ARGS ) 			// Doubleword ADD Immediate Unsigned
{	
	// Reserved Instruction exception

	//rt = rs + immediate
	gGPR[gCurrentOp.rt]._s64 = gGPR[gCurrentOp.rs]._s64 + (s32)(s16)gCurrentOp.immediate;
	CHECK_R0();
}


void CALL_TYPE R4300_ADDI( CALL_ARGS ) 
{
	// Generates overflow exception

	//rt = rs + immediate
	gGPR[gCurrentOp.rt]._s64 = (s64)(s32)(gGPR[gCurrentOp.rs]._s32[0] + (s32)(s16)gCurrentOp.immediate);

	CHECK_R0();
}

void CALL_TYPE R4300_ADDIU( CALL_ARGS ) 		// Add Immediate Unsigned
{
	//rt = rs + immediate
	gGPR[gCurrentOp.rt]._s64 = (s64)(s32)(gGPR[gCurrentOp.rs]._s32[0] + (s32)(s16)gCurrentOp.immediate);
	CHECK_R0();

}

void CALL_TYPE R4300_SLTI( CALL_ARGS ) 			// Set on Less Than Immediate
{	
	// Cast to s32s to ensure sign is taken into account
	if (gGPR[gCurrentOp.rs]._s64 < (s64)(s32)(s16)gCurrentOp.immediate)
	{
		gGPR[gCurrentOp.rt]._u64 = 1;
		CHECK_R0();
	}
	else
	{
		gGPR[gCurrentOp.rt]._u64 = 0;
	}
}

void CALL_TYPE R4300_SLTIU( CALL_ARGS ) 		// Set on Less Than Immediate Unsigned 	
{	
	// Cast to s32s to ensure sign is taken into account
	if (gGPR[gCurrentOp.rs]._u64 < (u64)(s64)(s32)(s16)gCurrentOp.immediate)
	{
		gGPR[gCurrentOp.rt]._u64 = 1;
		CHECK_R0();
	}
	else
	{
		gGPR[gCurrentOp.rt]._u64 = 0;
	}
}


void CALL_TYPE R4300_ANDI( CALL_ARGS ) 				// AND Immediate
{
	//rt = rs & immediate
	gGPR[gCurrentOp.rt]._u64 = gGPR[gCurrentOp.rs]._u64 & (u64)(u16)gCurrentOp.immediate;
	CHECK_R0();
}


void CALL_TYPE R4300_ORI( CALL_ARGS ) 				// OR Immediate
{				
	//rt = rs | immediate
	gGPR[gCurrentOp.rt]._u64 = gGPR[gCurrentOp.rs]._u64 | (u64)(u16)gCurrentOp.immediate;
	CHECK_R0();
}

void CALL_TYPE R4300_XORI( CALL_ARGS ) 				// XOR Immediate
{
	//rt = rs ^ immediate
	gGPR[gCurrentOp.rt]._u64 = gGPR[gCurrentOp.rs]._u64 ^ (u64)(u16)gCurrentOp.immediate;
	CHECK_R0();
}

void CALL_TYPE R4300_LUI( CALL_ARGS ) 				// Load Upper Immediate
{
	gGPR[gCurrentOp.rt]._s64 = (s64)(s32)((s32)(s16)gCurrentOp.immediate<<16);
	CHECK_R0();
}

void CALL_TYPE R4300_BEQL( CALL_ARGS ) 			// Branch on Equal Likely
{
	//branch if rs == rt
	if ( gGPR[gCurrentOp.rs]._u64 == gGPR[gCurrentOp.rt]._u64 )
	{
		s16 offset = (s16)gCurrentOp.immediate;

		SpeedHack(offset == -1, gCurrentOp);

		g_dwNewPC = g_dwPC + ((s32)offset<<2) + 4;
		CPU_Delay();
	}
	else
	{
		// Don't execute subsequent instruction
		INCREMENT_PC();
	}
}

void CALL_TYPE R4300_BNEL( CALL_ARGS ) 			// Branch on Not Equal Likely
{
	//branch if rs <> rt
	if ( gGPR[gCurrentOp.rs]._u64 != gGPR[gCurrentOp.rt]._u64 )
	{
		s16 offset = (s16)gCurrentOp.immediate;

		SpeedHack(offset == -1, gCurrentOp);

		g_dwNewPC = g_dwPC + ((s32)offset<<2) + 4;
		CPU_Delay();	
	}
	else
	{
		// Don't execute subsequent instruction
		INCREMENT_PC();
	}
}

void CALL_TYPE R4300_BLEZL( CALL_ARGS ) 		// Branch on Less than or Equal to Zero Likely
{
	//branch if rs <= 0
	if ( gGPR[gCurrentOp.rs]._s64 <= 0 )
	{
		g_dwNewPC = g_dwPC + ((s32)(s16)gCurrentOp.immediate<<2) + 4;
		CPU_Delay();
	}
	else
	{
		// Don't execute subsequent instruction
		INCREMENT_PC();
	}
}

void CALL_TYPE R4300_BGTZL( CALL_ARGS ) 		// Branch on Greater than Zero Likely
{
	//branch if rs > 0
	if ( gGPR[gCurrentOp.rs]._s64 > 0 )
	{
		g_dwNewPC = g_dwPC + ((s32)(s16)gCurrentOp.immediate<<2) + 4;
		CPU_Delay();
	}
	else
	{
		// Don't execute subsequent instruction
		INCREMENT_PC();
	}
}





void CALL_TYPE R4300_LB( CALL_ARGS ) 			// Load Byte
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );

	gGPR[gCurrentOp.rt]._s64 = (s64)(s8)Read8Bits(address);
	CHECK_R0();
}

void CALL_TYPE R4300_LBU( CALL_ARGS ) 			// Load Byte Unsigned -- Zero extend byte...
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate);
	
	gGPR[gCurrentOp.rt]._s64 = (s64)(u8)Read8Bits(address);
	CHECK_R0();
}


void CALL_TYPE R4300_LH( CALL_ARGS ) 		// Load Halfword
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	gGPR[gCurrentOp.rt]._s64 = (s64)(s16)Read16Bits(address);
	CHECK_R0();
}

void CALL_TYPE R4300_LHU( CALL_ARGS )			// Load Halfword Unsigned -- Zero extend word
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );

	gGPR[gCurrentOp.rt]._u64 = (u64)(u16)Read16Bits(address);
	CHECK_R0();
}


void CALL_TYPE R4300_LWL( CALL_ARGS ) 			// Load Word Left
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate ); 
	u32 nMemory = Read32Bits(address & ~0x3);

	u32 nReg = gGPR[gCurrentOp.rt]._u32[0];

	switch (address % 4)
	{
        case 0: nReg = nMemory; break;
        case 1: nReg = ((nReg & 0x000000FF) | (nMemory << 8));  break;
        case 2: nReg = ((nReg & 0x0000FFFF) | (nMemory << 16)); break;
        case 3: nReg = ((nReg & 0x00FFFFFF) | (nMemory << 24)); break;
    }

	gGPR[gCurrentOp.rt]._s64 = (s64)(s32)nReg;
	CHECK_R0();
}

// Starcraft - not tested!
void CALL_TYPE R4300_LDL( CALL_ARGS )
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate ); 
	u64 nMemory = Read64Bits(address & ~0x7);		

	u64 nReg = gGPR[gCurrentOp.rt]._u64;

	switch (address % 8)
	{
        case 0: nReg = nMemory; break;
        case 1: nReg = ((nReg & 0x00000000000000FFLL) | (nMemory << 8));  break;
        case 2: nReg = ((nReg & 0x000000000000FFFFLL) | (nMemory << 16)); break;
        case 3: nReg = ((nReg & 0x0000000000FFFFFFLL) | (nMemory << 24)); break;
        case 4: nReg = ((nReg & 0x00000000FFFFFFFFLL) | (nMemory << 32)); break;
        case 5: nReg = ((nReg & 0x000000FFFFFFFFFFLL) | (nMemory << 40)); break;
        case 6: nReg = ((nReg & 0x0000FFFFFFFFFFFFLL) | (nMemory << 48)); break;
        case 7: nReg = ((nReg & 0x00FFFFFFFFFFFFFFLL) | (nMemory << 56)); break;
   }

	gGPR[gCurrentOp.rt]._u64 = nReg;
	CHECK_R0();
}


void CALL_TYPE R4300_LWR( CALL_ARGS ) 			// Load Word Right
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate ); 
	u32 nMemory = Read32Bits(address & ~0x3);
	
	u32 nReg = gGPR[gCurrentOp.rt]._u32[0];

	
	switch (address % 4)
	{
        case 0: nReg = (nReg & 0xFFFFFF00) | (nMemory >> 24); break;
        case 1: nReg = (nReg & 0xFFFF0000) | (nMemory >> 16); break;
        case 2: nReg = (nReg & 0xFF000000) | (nMemory >>  8); break;
        case 3: nReg = nMemory; break;
    }

	gGPR[gCurrentOp.rt]._s64 = (s64)(s32)nReg;
	CHECK_R0();
}

// Starcraft - not tested!
void CALL_TYPE R4300_LDR( CALL_ARGS )
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate ); 
	u64 nMemory = Read64Bits(address & ~0x7);
	
	u64 nReg = gGPR[gCurrentOp.rt]._u64;

	switch (address % 8)
	{
        case 0: nReg = (nReg & 0xFFFFFFFFFFFFFF00LL) | (nMemory >> 56); break;
        case 1: nReg = (nReg & 0xFFFFFFFFFFFF0000LL) | (nMemory >> 48); break;
        case 2: nReg = (nReg & 0xFFFFFFFFFF000000LL) | (nMemory >> 40); break;
        case 3: nReg = (nReg & 0xFFFFFFFF00000000LL) | (nMemory >> 32); break;
        case 4: nReg = (nReg & 0xFFFFFF0000000000LL) | (nMemory >> 24); break;
        case 5: nReg = (nReg & 0xFFFF000000000000LL) | (nMemory >> 16); break;
        case 6: nReg = (nReg & 0xFF00000000000000LL) | (nMemory >>  8); break;
        case 7: nReg = nMemory; break;
    }

	gGPR[gCurrentOp.rt]._u64 = nReg;
	CHECK_R0();
}


void CALL_TYPE R4300_LW( CALL_ARGS ) 			// Load Word
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	gGPR[gCurrentOp.rt]._s64 = (s64)(s32)Read32Bits(address);

	CHECK_R0();
}



void CALL_TYPE R4300_LWU( CALL_ARGS ) 			// Load Word Unsigned
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	gGPR[gCurrentOp.rt]._u64 = (u64)(u32)Read32Bits(address);

	CHECK_R0();
}

void CALL_TYPE R4300_SW( CALL_ARGS ) 			// Store Word
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );

	Write32Bits(address, gGPR[gCurrentOp.rt]._u32[0]);
}

void CALL_TYPE R4300_SH( CALL_ARGS ) 			// Store Halfword
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	Write16Bits(address, gGPR[gCurrentOp.rt]._u16[0]);
}

void CALL_TYPE R4300_SB( CALL_ARGS ) 			// Store Byte
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	Write8Bits(address, gGPR[gCurrentOp.rt]._u8[0]);
}



void CALL_TYPE R4300_SWL( CALL_ARGS ) 			// Store Word Left
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
  
	u32 dwMemory = Read32Bits(address & ~0x3);
	u32 dwReg = gGPR[gCurrentOp.rt]._u32[0];
	u32 dwNew = 0;

	switch (address % 4)
	{
	case 0:	dwNew = dwReg; break;			// Aligned
	case 1:	dwNew = (dwMemory & 0xFF000000) | (dwReg >> 8 ); break;
	case 2:	dwNew = (dwMemory & 0xFFFF0000) | (dwReg >> 16); break;
	case 3:	dwNew = (dwMemory & 0xFFFFFF00) | (dwReg >> 24); break;
	}
	Write32Bits(address & ~0x3, dwNew);
}


void CALL_TYPE R4300_SWR( CALL_ARGS ) 			// Store Word Right 
{	
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
  
	u32 dwMemory = Read32Bits(address & ~0x3);
	u32 dwReg = gGPR[gCurrentOp.rt]._u32[0];
	u32 dwNew = 0;

	switch (address % 4)
	{
	case 0:	dwNew = (dwMemory & 0x00FFFFFF) | (dwReg << 24); break;
	case 1:	dwNew = (dwMemory & 0x0000FFFF) | (dwReg << 16); break;
	case 2:	dwNew = (dwMemory & 0x000000FF) | (dwReg << 8); break;
	case 3:	dwNew = dwReg; break;			// Aligned
	}
	Write32Bits(address & ~0x3, dwNew);

}


void CALL_TYPE R4300_SDL( CALL_ARGS )//CYRUS64
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate);
	
	u64 nMemory = Read64Bits(address & ~0x7);
	u64 nReg = gGPR[gCurrentOp.rt]._u64;
	u64 nNew = 0;
	
	switch (address % 8)
	{
	case 0:	nNew = nReg; break;			// Aligned
	case 1:	nNew = (nMemory & 0xFF00000000000000LL) | (nReg >> 8); break;
	case 2:	nNew = (nMemory & 0xFFFF000000000000LL) | (nReg >> 16); break;
	case 3:	nNew = (nMemory & 0xFFFFFF0000000000LL) | (nReg >> 24); break;
	case 4:	nNew = (nMemory & 0xFFFFFFFF00000000LL) | (nReg >> 32); break;
	case 5:	nNew = (nMemory & 0xFFFFFFFFFF000000LL) | (nReg >> 40); break;
	case 6:	nNew = (nMemory & 0xFFFFFFFFFFFF0000LL) | (nReg >> 48); break;
	default:nNew = (nMemory & 0xFFFFFFFFFFFFFF00LL) | (nReg >> 56); break;
	}
	Write64Bits(address & ~0x7, nNew);
}

void CALL_TYPE R4300_SDR( CALL_ARGS )//CYRUS64
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate);
	
	u64 nMemory = Read64Bits(address & ~0x7);
	u64 nReg = gGPR[gCurrentOp.rt]._u64;
	u64 nNew = 0;
	
	//DBGConsole_Msg(0,"Address:		0x%08x", dwAddress );
	
	switch (address % 8)
	{
	case 0:	nNew = (nMemory & 0x00FFFFFFFFFFFFFFLL) | (nReg << 56); break;
	case 1:	nNew = (nMemory & 0x0000FFFFFFFFFFFFLL) | (nReg << 48); break;
	case 2:	nNew = (nMemory & 0x000000FFFFFFFFFFLL) | (nReg << 40); break;
	case 3:	nNew = (nMemory & 0x00000000FFFFFFFFLL) | (nReg << 32); break;
	case 4:	nNew = (nMemory & 0x0000000000FFFFFFLL) | (nReg << 24); break;
	case 5:	nNew = (nMemory & 0x000000000000FFFFLL) | (nReg << 16); break;
	case 6:	nNew = (nMemory & 0x00000000000000FFLL) | (nReg << 8); break;
	default:nNew = nReg; break;			// Aligned
	}
	Write64Bits(address & ~0x7, nNew);
	
	
}



void CALL_TYPE R4300_CACHE( CALL_ARGS )
{
	return;

	u32 cache_op  = gCurrentOp.rt;

	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );

	// Do Nothing
	u32 dwCache = cache_op & 0x3;
	u32 dwAction = (cache_op >> 2) & 0x7;

	static const char * szCaches[] =
	{
		"I", "D", "SI", "SD"
	};	

	DBGConsole_Msg(0, "CACHE %s/%d, 0x%08x", szCaches[dwCache], dwAction, address);
}

void CALL_TYPE R4300_LWC1( CALL_ARGS ) 				// Load Word to Copro 1 (FPU)
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	StoreFPR_Word( gCurrentOp.ft, Read32Bits(address) );
}


void CALL_TYPE R4300_LDC1( CALL_ARGS )				// Load Doubleword to Copro 1 (FPU)
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	StoreFPR_Long( gCurrentOp.ft, Read64Bits(address));
}


void CALL_TYPE R4300_LD( CALL_ARGS ) 				// Load Doubleword
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	gGPR[gCurrentOp.rt]._u64 = Read64Bits(address);
	CHECK_R0();
}


void CALL_TYPE R4300_SWC1( CALL_ARGS ) 			// Store Word From Copro 1
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	//Write32Bits(address, (u32)g_qwCPR[1][dwFT]);
	Write32Bits(address, LoadFPR_Word(gCurrentOp.ft));
}

void CALL_TYPE R4300_SDC1( CALL_ARGS )		// Store Doubleword From Copro 1
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );

	Write64Bits(address, LoadFPR_Long(gCurrentOp.ft));
}


void CALL_TYPE R4300_SD( CALL_ARGS )			// Store Doubleword
{
	u32 address = (u32)( gGPR[gCurrentOp.base]._s32[0] + (s32)(s16)gCurrentOp.immediate );
	Write64Bits(address, gGPR[gCurrentOp.rt]._u64);
}



/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

void CALL_TYPE R4300_Special_Unk( CALL_ARGS ) { WARN_NOEXIST("R4300_Special_Unk"); }
void CALL_TYPE R4300_Special_SLL( CALL_ARGS ) 		// Shift word Left Logical
{
	// NOP!
	if ( gCurrentOp._u32 == 0 ) return;

	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( (gGPR[ gCurrentOp.rt ]._u32[0] << gCurrentOp.sa) & 0xFFFFFFFF );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_SRL( CALL_ARGS ) 		// Shift word Right Logical
{
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( gGPR[ gCurrentOp.rt ]._u32[0] >> gCurrentOp.sa );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_SRA( CALL_ARGS ) 		// Shift word Right Arithmetic
{	
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( gGPR[ gCurrentOp.rt ]._s32[0] >> gCurrentOp.sa );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_SLLV( CALL_ARGS ) 		// Shift word Left Logical Variable
{
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( (gGPR[ gCurrentOp.rt ]._u32[0] << ( gGPR[ gCurrentOp.rs ]._u32[0] & 0x1F ) ) & 0xFFFFFFFF );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_SRLV( CALL_ARGS ) 		// Shift word Right Logical Variable
{
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( gGPR[ gCurrentOp.rt ]._u32[0] >> ( gGPR[ gCurrentOp.rs ]._u32[0] & 0x1F ) );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_SRAV( CALL_ARGS ) 		// Shift word Right Arithmetic Variable
{
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( gGPR[ gCurrentOp.rt ]._s32[0] >> ( gGPR[ gCurrentOp.rs ]._u32[0] & 0x1F ) );
	CHECK_R0();
}



void CALL_TYPE R4300_Special_JR_CheckBootAddress( CALL_ARGS ) 			// Jump Register
{	
	g_dwNewPC = gGPR[ gCurrentOp.rs ]._u32[0];

	// This is a hack to detect when the rom's boot code
	// has finished with the CRC check, and is *just* about to 
	// start executing the game code. We apply patches here,
	// otherwise the CRC would fail (which is bad :-)
	if (g_dwNewPC == SwapEndian(g_ROM.rh.BootAddress))
	{
		Write32Bits(0x80000318, gRamSize);
		//Write32Bits(0x800003F0, gRamSize);
		DBGConsole_Msg(0, "[RJumping to Game Boot Address]");

#ifdef DAEDALUS_RELEASE_BUILD
#ifdef DAEDALUS_PSP
		// Shut the console up when we get this far in public builds
		CDebugConsole::Get()->EnableConsole( false );
#endif
#endif

		// Disable this on the PSP for the time being (faster booting...)
#ifndef DAEDALUS_PSP
		Patch_ApplyPatches();
#endif

//		CPU_Halt( "Boot" );

		// Restore normal op handling - this reduces the 
		// overhead of this check for the rest of the
		// execution of the rom.
		R4300SpecialInstruction[OP_JR] = R4300_Special_JR;
	}

	CPU_Delay();
	
}

void CALL_TYPE R4300_Special_JR( CALL_ARGS ) 			// Jump Register
{	
	g_dwNewPC = gGPR[ gCurrentOp.rs ]._u32[0];

	CPU_Delay();
}


void CALL_TYPE R4300_Special_JALR( CALL_ARGS ) 		// Jump and Link register
{	
	// Jump And Link
	g_dwNewPC = gGPR[ gCurrentOp.rs ]._u32[0];
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)(g_dwPC + 8);		// Store return address
	CHECK_R0();
	
	CPU_Delay();

}


void CALL_TYPE R4300_Special_SYSCALL( CALL_ARGS )
{
	R4300_Exception_Syscall();
}

void CALL_TYPE R4300_Special_BREAK( CALL_ARGS ) 	// BREAK
{
	DPF( DEBUG_INTR, DSPrintf( "BREAK Called. PC: 0x%08x. COUNT: 0x%08x", g_dwPC, g_qwCPR[0][C0_COUNT]._u32[0] ) );
	R4300_Exception_Break();
}

void CALL_TYPE R4300_Special_SYNC( CALL_ARGS )
{
	// Just ignore
}

void CALL_TYPE R4300_Special_MFHI( CALL_ARGS ) 			// Move From MultHI
{
	gGPR[ gCurrentOp.rd ]._u64 = g_qwMultHI;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_MTHI( CALL_ARGS ) 			// Move To MultHI
{
	g_qwMultHI = gGPR[ gCurrentOp.rs ]._u64;
}

void CALL_TYPE R4300_Special_MFLO( CALL_ARGS ) 			// Move From MultLO
{
	gGPR[ gCurrentOp.rd ]._u64 = g_qwMultLO;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_MTLO( CALL_ARGS ) 			// Move To MultLO
{
	g_qwMultLO = gGPR[ gCurrentOp.rs ]._u64;
}

// BEGIN MODIFIED BY Lkb - 8/jun/2001 - changed 0x1f to 0x3f because the value to be shifted is 64-bit long
void CALL_TYPE R4300_Special_DSLLV( CALL_ARGS ) 
{
	// Reserved Instruction exception
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 << ( gGPR[ gCurrentOp.rs ]._u32[0] & 0x3F );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DSRLV( CALL_ARGS )
{
	// Reserved Instruction exception
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 >> ( gGPR[ gCurrentOp.rs ]._u32[0] & 0x3F );
	CHECK_R0();
}

// Aeroguage uses!
void CALL_TYPE R4300_Special_DSRAV( CALL_ARGS )
{
	// Reserved Instruction exception
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._s64 >> ( gGPR[ gCurrentOp.rs ]._u32[0] & 0x3F );
	CHECK_R0();

}
// END MODIFIED BY Lkb - 8/jun/2001



void CALL_TYPE R4300_Special_MULT( CALL_ARGS ) 			// MULTiply Signed
{
	s64 dwResult = (s64)gGPR[ gCurrentOp.rs ]._s32[0] * (s64)gGPR[ gCurrentOp.rt ]._s32[0];
	g_qwMultLO = (s64)(s32)(dwResult);
	g_qwMultHI = (s64)(s32)(dwResult >> 32);

}

void CALL_TYPE R4300_Special_MULTU( CALL_ARGS ) 		// MULTiply Unsigned
{
	u64 dwResult = (u64)gGPR[ gCurrentOp.rs ]._u32[0] * (u64)gGPR[ gCurrentOp.rt ]._u32[0];
	g_qwMultLO = (s64)(s32)(dwResult);
	g_qwMultHI = (s64)(s32)(dwResult >> 32);
}

void CALL_TYPE R4300_Special_DIV( CALL_ARGS ) 			//DIVide
{
	s32 nDividend = gGPR[ gCurrentOp.rs ]._s32[0];
	s32 nDivisor  = gGPR[ gCurrentOp.rt ]._s32[0];

	if (nDivisor)
	{
		g_qwMultLO = (s64)(s32)(nDividend / nDivisor);
		g_qwMultHI = (s64)(s32)(nDividend % nDivisor);
	}
}

void CALL_TYPE R4300_Special_DIVU( CALL_ARGS ) 			// DIVide Unsigned
{
	u32 dwDividend = gGPR[ gCurrentOp.rs ]._u32[0];
	u32 dwDivisor  = gGPR[ gCurrentOp.rt ]._u32[0];

	if (dwDivisor) {
		g_qwMultLO = (s64)(s32)(dwDividend / dwDivisor);
		g_qwMultHI = (s64)(s32)(dwDividend % dwDivisor);
	}
}

void CALL_TYPE R4300_Special_DMULT( CALL_ARGS ) 		// Double Multiply
{
	// Reserved Instruction exception
	g_qwMultLO = gGPR[ gCurrentOp.rs ]._s64 * gGPR[ gCurrentOp.rt ]._s64;
	g_qwMultHI = 0;
}

void CALL_TYPE R4300_Special_DMULTU( CALL_ARGS ) 			// Double Multiply Unsigned
{
	// Reserved Instruction exception
	g_qwMultLO = gGPR[ gCurrentOp.rs ]._u64 * gGPR[ gCurrentOp.rt ]._u64;
	g_qwMultHI = 0;
}

void CALL_TYPE R4300_Special_DDIV( CALL_ARGS ) 				// Double Divide
{
	s64 qwDividend = gGPR[ gCurrentOp.rs ]._s64;
	s64 qwDivisor = gGPR[ gCurrentOp.rt ]._s64;

	// Reserved Instruction exception
	if (qwDivisor)
	{
		g_qwMultLO = qwDividend / qwDivisor;
		g_qwMultHI = qwDividend % qwDivisor;
	}
}

void CALL_TYPE R4300_Special_DDIVU( CALL_ARGS ) 			// Double Divide Unsigned
{	
	u64 qwDividend = gGPR[ gCurrentOp.rs ]._u64;
	u64 qwDivisor = gGPR[ gCurrentOp.rt ]._u64;

	// Reserved Instruction exception
	if (qwDivisor)
	{
		g_qwMultLO = qwDividend / qwDivisor;
		g_qwMultHI = qwDividend % qwDivisor;
	}
}

void CALL_TYPE R4300_Special_ADD( CALL_ARGS ) 			// ADD signed - may throw exception
{
	// Can generate overflow exception
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( gGPR[ gCurrentOp.rs ]._s32[0] + gGPR[ gCurrentOp.rt ]._s32[0] );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_ADDU( CALL_ARGS ) 			// ADD Unsigned - doesn't throw exception
{
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( gGPR[ gCurrentOp.rs ]._s32[0] + gGPR[ gCurrentOp.rt ]._s32[0] );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_SUB( CALL_ARGS ) 			// SUB Signed - may throw exception
{
	// Can generate overflow exception
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( gGPR[ gCurrentOp.rs ]._s32[0] - gGPR[ gCurrentOp.rt ]._s32[0] );
	CHECK_R0();
}


void CALL_TYPE R4300_Special_SUBU( CALL_ARGS ) 			// SUB Unsigned - doesn't throw exception
{
	gGPR[ gCurrentOp.rd ]._s64 = (s64)(s32)( gGPR[ gCurrentOp.rs ]._s32[0] - gGPR[ gCurrentOp.rt ]._s32[0] );
	CHECK_R0();
}


void CALL_TYPE R4300_Special_AND( CALL_ARGS ) 				// logical AND
{
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rs ]._u64 & gGPR[ gCurrentOp.rt ]._u64;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_OR( CALL_ARGS ) 				// logical OR
{
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rs ]._u64 | gGPR[ gCurrentOp.rt ]._u64;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_XOR( CALL_ARGS ) 				// logical XOR
{

	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rs ]._u64 ^ gGPR[ gCurrentOp.rt ]._u64;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_NOR( CALL_ARGS ) 				// logical Not OR
{
	gGPR[ gCurrentOp.rd ]._u64 = ~( gGPR[ gCurrentOp.rs ]._u64 | gGPR[ gCurrentOp.rt ]._u64 );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_SLT( CALL_ARGS ) 				// Set on Less Than
{	
	// Cast to s32s to ensure sign is taken into account
	if ( gGPR[ gCurrentOp.rs ]._s64 < gGPR[ gCurrentOp.rt ]._s64 )
	{
		gGPR[ gCurrentOp.rd ]._u64 = 1;
		CHECK_R0();
	}
	else
	{
		gGPR[ gCurrentOp.rd ]._u64 = 0;
	}
}

void CALL_TYPE R4300_Special_SLTU( CALL_ARGS ) 				// Set on Less Than Unsigned
{
	// Treated as unsigned....
	if ( gGPR[ gCurrentOp.rs ]._u64 < gGPR[ gCurrentOp.rt ]._u64 )
	{
		gGPR[ gCurrentOp.rd ]._u64 = 1;
		CHECK_R0();
	}
	else
	{
		gGPR[ gCurrentOp.rd ]._u64 = 0;
	}

}


void CALL_TYPE R4300_Special_DADD( CALL_ARGS )//CYRUS64 
{  
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 + gGPR[ gCurrentOp.rs ]._u64;	
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DADDU( CALL_ARGS )//CYRUS64
{
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 + gGPR[ gCurrentOp.rs ]._u64;
	CHECK_R0();
}



void CALL_TYPE R4300_Special_DSUB( CALL_ARGS )
{
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 - gGPR[ gCurrentOp.rs ]._u64;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DSUBU( CALL_ARGS )
{
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 - gGPR[ gCurrentOp.rs ]._u64;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DSLL( CALL_ARGS ) 
{
	// Reserved Instruction exception
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 << gCurrentOp.sa;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DSRL( CALL_ARGS ) 
{
	// Reserved Instruction exception
    gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 >> gCurrentOp.sa;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DSRA( CALL_ARGS )
{
	// Reserved Instruction exception
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._s64 >> gCurrentOp.sa;
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DSLL32( CALL_ARGS ) 			// Double Shift Left Logical 32
{	
	// Reserved Instruction exception
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 << ( 32 + gCurrentOp.sa );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DSRL32( CALL_ARGS ) 			// Double Shift Right Logical 32
{
	// Reserved Instruction exception
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._u64 >> ( 32 + gCurrentOp.sa );
	CHECK_R0();
}

void CALL_TYPE R4300_Special_DSRA32( CALL_ARGS ) 			// Double Shift Right Arithmetic 32
{
	// Reserved Instruction exception
	gGPR[ gCurrentOp.rd ]._u64 = gGPR[ gCurrentOp.rt ]._s64 >> ( 32 + gCurrentOp.sa );
	CHECK_R0();
}


/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////




void CALL_TYPE R4300_RegImm_Unk( CALL_ARGS ) {  WARN_NOEXIST("R4300_RegImm_Unk"); }


void CALL_TYPE R4300_RegImm_BLTZ( CALL_ARGS ) 			// Branch on Less than Zero
{
	//branch if rs < 0
	if ( gGPR[ gCurrentOp.rs ]._s64 < 0 )
	{
		s16 offset = (s16)gCurrentOp.immediate;

		SpeedHack(offset == -1, gCurrentOp );

		g_dwNewPC = g_dwPC + ((s32)offset<<2) + 4;
		CPU_Delay();

	} 
}

void CALL_TYPE R4300_RegImm_BLTZL( CALL_ARGS ) 			// Branch on Less than Zero Likely
{
	//branch if rs < 0
	if ( gGPR[ gCurrentOp.rs ]._s64 < 0 )
	{
		g_dwNewPC = g_dwPC + ((s32)(s16)gCurrentOp.immediate<<2) + 4;
		CPU_Delay();
	}
	else
	{
		// Don't execute subsequent instruction
		INCREMENT_PC();
	}

}
void CALL_TYPE R4300_RegImm_BLTZAL( CALL_ARGS ) 		// Branch on Less than Zero And Link
{
	//branch if rs >= 0
	// Store the return address even if branch not taken
	gGPR[REG_ra]._s64 = (s64)(s32)(g_dwPC + 8);		// Store return address	
	if ( gGPR[ gCurrentOp.rs ]._s64 < 0 )
	{
		g_dwNewPC = g_dwPC + ((s32)(s16)gCurrentOp.immediate<<2) + 4;
		CPU_Delay();
	} 
}

void CALL_TYPE R4300_RegImm_BGEZ( CALL_ARGS ) 			// Branch on Greater than or Equal to Zero
{
	//branch if rs >= 0
	if ( gGPR[ gCurrentOp.rs ]._s64 >= 0 )
	{
		s16 offset = (s16)gCurrentOp.immediate;

		SpeedHack( offset == -1, gCurrentOp );

		g_dwNewPC = g_dwPC + ((s32)offset<<2) + 4;
		CPU_Delay();
	} 
}

void CALL_TYPE R4300_RegImm_BGEZL( CALL_ARGS ) 			// Branch on Greater than or Equal to Zero Likely
{
	//branch if rs >= 0
	if ( gGPR[ gCurrentOp.rs ]._s64 >= 0 )
	{
		g_dwNewPC = g_dwPC + ((s32)(s16)gCurrentOp.immediate<<2) + 4;
		CPU_Delay();
	}
	else
	{
		// Don't execute subsequent instruction
		INCREMENT_PC();
	}
}

void CALL_TYPE R4300_RegImm_BGEZAL( CALL_ARGS ) 		// Branch on Greater than or Equal to Zero And Link
{
	//branch if rs >= 0
	// This always happens, even if branch not taken
	gGPR[REG_ra]._s64 = (s64)(s32)(g_dwPC + 8);		// Store return address	
	if ( gGPR[ gCurrentOp.rs ]._s64 >= 0 )
	{
		g_dwNewPC = g_dwPC + ((s32)(s16)gCurrentOp.immediate<<2) + 4;
		CPU_Delay();
	} 
}



/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

void CALL_TYPE R4300_Cop0_Unk( CALL_ARGS ) { WARN_NOEXIST("R4300_Cop0_Unk"); }
void CALL_TYPE R4300_TLB_Unk( CALL_ARGS )  { WARN_NOEXIST("R4300_TLB_Unk"); }

void CALL_TYPE R4300_Cop0_TLB( CALL_ARGS ) {  R4300TLBInstruction[ gCurrentOp.funct ]( ); }


void CALL_TYPE R4300_Cop0_MFC0( CALL_ARGS )
{
#ifdef _DEBUG
	if ( gCurrentOp.fs == C0_CAUSE )
	{
		bool	mi_interrupt_set( (Memory_MI_GetRegister(MI_INTR_MASK_REG) & Memory_MI_GetRegister(MI_INTR_REG)) != 0 );
		bool	cause_int_3_set( (g_qwCPR[0][C0_CAUSE]._u32[ 0 ] & CAUSE_IP3) != 0 );

		DAEDALUS_ASSERT( mi_interrupt_set == cause_int_3_set, "CAUSE_IP3 inconsistant with MI_INTR_REG" );
	}
#endif

	// Copy from FS to RT
	if ( gCurrentOp.fs == C0_RAND )
	{
		u32 dwWired = g_qwCPR[0][C0_WIRED]._u32[0] & 0x1F;

		// Select a value between dwWired and 31
		// We should use TLB least-recently used here too?
		gGPR[ gCurrentOp.rt ]._s64 = (rand()%(32-dwWired)) + dwWired;
	}
	else
	{
		gGPR[ gCurrentOp.rt ]._s64 = (s64)g_qwCPR[0][ gCurrentOp.fs ]._s32[0];
	}

	CHECK_R0();
}

// Move Word To CopReg
void CALL_TYPE R4300_Cop0_MTC0( CALL_ARGS )
{
	// Copy from RT to FS
	u64 new_value = gGPR[ gCurrentOp.rt ]._u64;

	switch ( gCurrentOp.fs )
	{
		case C0_CONTEXT:
			DBGConsole_Msg(0, "Setting context register to 0x%08x", (u32)new_value);
			g_qwCPR[0][ gCurrentOp.fs ]._u64 = new_value;
			break;

		case C0_WIRED:
			// Set to top limit on write to wired
			g_qwCPR[0][C0_RAND]._u64 = 32-1;
			DBGConsole_Msg(0, "Setting Wired register to 0x%08x", (u32)new_value);
			g_qwCPR[0][ gCurrentOp.fs ]._u64 = new_value;
			break;
		case C0_RAND:
		case C0_BADVADDR:
		case C0_PRID:
		case C0_CACHE_ERR:			// Furthermore, this reg must return 0 on reads.

			// All these registers are read only - make sure that software doesn't write to them
			DBGConsole_Msg(0, "MTC0. Software attempted to write to read only reg %s: 0x%08x", Cop0RegNames[ gCurrentOp.fs ], (u32)new_value);
			break;

		case C0_CAUSE:
			// Only IP1:0 (Interrupt Pending) bits are software writeable.
			// On writes, set all others to 0. Is this correct?
			//  Other bits are CE (copro error) BD (branch delay), the other
			// Interrupt pendings and EscCode.
			if ( (new_value&~(CAUSE_SW1|CAUSE_SW2)) != (g_qwCPR[0][C0_CAUSE]._u64&~(CAUSE_SW1|CAUSE_SW2))  )
			{
				DBGConsole_Msg( 0, "[MWas previously clobbering CAUSE REGISTER" );
			}

			DPF( DEBUG_REGS, DSPrintf( "CAUSE set to 0x%08x (was: 0x%08x)", (u32)new_value, gGPR[ gCurrentOp.rt ]._u32[0] ) );
			g_qwCPR[0][C0_CAUSE]._u64 &=             ~(CAUSE_SW1|CAUSE_SW2);
			g_qwCPR[0][C0_CAUSE]._u64 |= (new_value & (CAUSE_SW1|CAUSE_SW2));
			break;
		case C0_SR:
			// Software can enable/disable interrupts here. We check if Interrupt Enable is
			//  set, and if there are any pending interrupts. If there are, then we set the
			//  CHECK_POSTPONED_INTERRUPTS flag to make sure we check for interrupts that have
			//  occurred since we disabled interrupts

			R4300_SetSR((u32)new_value);
			break;

		case C0_COUNT:
			{
				// See comments below for COMPARE.
				// When this register is set, we need to check whether the next timed interrupt will
				//  be due to vertical blank or COMPARE
				g_qwCPR[0][C0_COUNT]._u64 = new_value;

				DBGConsole_Msg(0, "Count set - setting int");
				// Add an event for this compare:
				/*if ((u32)g_qwCPR[0][C0_COMPARE] > (u32)g_qwCPR[0][C0_COUNT])
				{
					CPU_AddEvent((u32)g_qwCPR[0][C0_COMPARE] - (u32)g_qwCPR[0][C0_COUNT],
								CPU_EVENT_COMPARE);
				}*/

				break;
			}
		case C0_COMPARE:
			{
				// When the value of COUNT equals the value of COMPARE, IP7 of the CAUSE register is set 
				// (Timer interrupt). Writing to this register clears the timer interrupt pending flag.
				CPU_SetCompare(new_value);
			}
			break;


		// Need to check CONFIG register writes - not all fields are writable.
		// This also sets Endianness mode.

		// WatchHi/WatchLo are used to create a Watch Trap. This may not need implementing, but we should
		// Probably provide a warning on writes, just so that we know
		case C0_WATCHLO:	
			DBGConsole_Msg( 0, "[MWROTE TO WATCHLO REGISTER!" );
			g_qwCPR[0][ gCurrentOp.fs ]._u64 = new_value;
			break;
		case C0_WATCHHI:
			DBGConsole_Msg( 0, "[MWROTE TO WATCHHI REGISTER!" );
			g_qwCPR[0][ gCurrentOp.fs ]._u64 = new_value;
			break;

		default:
			// No specific handling needs for writes to these registers.
			g_qwCPR[0][ gCurrentOp.fs ]._u64 = new_value;
			break;
	}
}


void CALL_TYPE R4300_TLB_TLBR( CALL_ARGS ) 				// TLB Read
{
	u32 index = g_qwCPR[0][C0_INX]._u32[0] & 0x1F;

	g_qwCPR[0][C0_PAGEMASK]._u64 = g_TLBs[index].mask;
	g_qwCPR[0][C0_ENTRYHI ]._u64 = g_TLBs[index].hi   & (~g_TLBs[index].pagemask);
	g_qwCPR[0][C0_ENTRYLO0]._u64 = g_TLBs[index].pfne | g_TLBs[index].g;
	g_qwCPR[0][C0_ENTRYLO1]._u64 = g_TLBs[index].pfno | g_TLBs[index].g;

	DPF( DEBUG_TLB, DSPrintf( "TLBR: INDEX: 0x%04x. PAGEMASK: 0x%08x.", index, g_qwCPR[0][C0_PAGEMASK]._u32[0] ) );
	DPF( DEBUG_TLB, DSPrintf( "      ENTRYHI: 0x%08x. ENTRYLO1: 0x%08x. ENTRYLO0: 0x%08x", g_qwCPR[0][C0_ENTRYHI]._u32[0], g_qwCPR[0][C0_ENTRYLO1]._u32[0], g_qwCPR[0][C0_ENTRYLO0]._u32[0] ) );
}


void CALL_TYPE R4300_TLB_TLBWI( CALL_ARGS )			// TLB Write Index
{
	u32 i = g_qwCPR[0][C0_INX]._u32[0] & 0x1F;

	DPF( DEBUG_TLB, DSPrintf( "TLBWI: INDEX: 0x%04x. PAGEMASK: 0x%08x.", i, g_qwCPR[0][C0_PAGEMASK]._u32[0] ) );
	DPF( DEBUG_TLB, DSPrintf( "       ENTRYHI: 0x%08x. ENTRYLO1: 0x%08x. ENTRYLO0: 0x%08x", g_qwCPR[0][C0_ENTRYHI]._u32[0], g_qwCPR[0][C0_ENTRYLO1]._u32[0], g_qwCPR[0][C0_ENTRYLO0]._u32[0] ) );

	// From the R4300i Instruction manual:
	// The G bit of the TLB is written with the logical AND of the G bits of the EntryLo0 and EntryLo1 regs
	// The TLB entry is loaded with the contents of the EntryHi and EntryLo regs.

	// TLB[INDEX] <- PageMask || (EntryHi AND NOT PageMask) || EntryLo1 || EntryLo0

	g_TLBs[i].pagemask	= g_qwCPR[0][C0_PAGEMASK]._u32[0];
	g_TLBs[i].hi		= g_qwCPR[0][C0_ENTRYHI ]._u32[0];
	g_TLBs[i].pfno		= g_qwCPR[0][C0_ENTRYLO1]._u32[0];		// Clear Global bit?
	g_TLBs[i].pfne		= g_qwCPR[0][C0_ENTRYLO0]._u32[0];		// Clear Global bit?

	g_TLBs[i].g = g_qwCPR[0][C0_ENTRYLO1]._u32[0] &
		          g_qwCPR[0][C0_ENTRYLO0]._u32[0] &
				  TLBLO_G;

	// Build the masks:
	g_TLBs[i].mask     =  g_TLBs[i].pagemask | (~TLBHI_VPN2MASK);
	g_TLBs[i].mask2    =  g_TLBs[i].mask>>1;
	g_TLBs[i].vpnmask  = ~g_TLBs[i].mask;
	g_TLBs[i].vpn2mask =  g_TLBs[i].vpnmask>>1;

	g_TLBs[i].addrcheck= g_TLBs[i].hi & g_TLBs[i].vpnmask;

	g_TLBs[i].pfnehi = ((g_TLBs[i].pfne<<TLBLO_PFNSHIFT) & g_TLBs[i].vpn2mask);
	g_TLBs[i].pfnohi = ((g_TLBs[i].pfno<<TLBLO_PFNSHIFT) & g_TLBs[i].vpn2mask);

	switch (g_TLBs[i].pagemask)
	{
		case TLBPGMASK_4K:	// 4k 
			DPF(DEBUG_TLB, "       4k Pagesize");
            g_TLBs[i].checkbit = 0x00001000;   // bit 12
            break;
        case TLBPGMASK_16K: //  16k pagesize 
			DPF(DEBUG_TLB, "       16k Pagesize");
            g_TLBs[i].checkbit = 0x00004000;   // bit 14
            break;
        case TLBPGMASK_64K: //  64k pagesize
			DPF(DEBUG_TLB, "       64k Pagesize");
            g_TLBs[i].checkbit = 0x00010000;   // bit 16
            break;
        case 0x0007e000: // 256k pagesize
			DPF(DEBUG_TLB, "       256k Pagesize");
            g_TLBs[i].checkbit = 0x00040000;   // bit 18
            break;
        case 0x001fe000: //   1M pagesize
			DPF(DEBUG_TLB, "       1M Pagesize");
            g_TLBs[i].checkbit = 0x00100000;   // bit 20
            break;
        case 0x007fe000: //   4M pagesize
			DPF(DEBUG_TLB, "       4M Pagesize");
            g_TLBs[i].checkbit = 0x00400000;   // bit 22
            break;
        case 0x01ffe000: //  16M pagesize
			DPF(DEBUG_TLB, "       16M Pagesize");
            g_TLBs[i].checkbit = 0x01000000;   // bit 24
            break;
        default: // should not happen! 
			DPF(DEBUG_TLB, "       Unknown Pagesize");
            g_TLBs[i].checkbit = 0;
            break;
	}
}


void CALL_TYPE R4300_TLB_TLBWR( CALL_ARGS )
{
	u32 i;
	u32 iLeastRecent( 0 );			// Was uninitialised - not sure if this mattered
	u32 dwLeastRecent;
	
	// Select a value for index between dwWired and 31
	u32 dwWired = g_qwCPR[0][C0_WIRED]._u32[0] & 0x1F;
	//i = (rand()%(32-dwWired)) + dwWired;

	// Don't choose random value - choose the least recently used page to 
	// replace
	dwLeastRecent = ~0;
	for (i = dwWired; i < 32; i++)
	{
		if (g_TLBs[i].lastaccessed < dwLeastRecent)
		{
			iLeastRecent = i;
			dwLeastRecent = g_TLBs[i].lastaccessed;
		}
	}

	i = iLeastRecent;

	//DBGConsole_Msg(0, "TLBWR: Using %d for C0_RAND", i);

	DPF( DEBUG_TLB, DSPrintf( "TLBWR: INDEX: 0x%04x. PAGEMASK: 0x%08x.", i, g_qwCPR[0][C0_PAGEMASK]._u32[0] ) );
	DPF( DEBUG_TLB, DSPrintf( "       ENTRYHI: 0x%08x. ENTRYLO1: 0x%08x. ENTRYLO0: 0x%08x", g_qwCPR[0][C0_ENTRYHI]._u32[0], g_qwCPR[0][C0_ENTRYLO1]._u32[0], g_qwCPR[0][C0_ENTRYLO0]._u32[0] ) );

	// From the R4300i Instruction manual:
	// The G bit of the TLB is written with the logical AND of the G bits of the EntryLo0 and EntryLo1 regs
	// The TLB entry is loaded with the contents of the EntryHi and EntryLo regs.

	// TLB[INDEX] <- PageMask || (EntryHi AND NOT PageMask) || EntryLo1 || EntryLo0

	g_TLBs[i].pagemask	= g_qwCPR[0][C0_PAGEMASK]._u32[0];
	g_TLBs[i].hi		= g_qwCPR[0][C0_ENTRYHI ]._u32[0];
	g_TLBs[i].pfno		= g_qwCPR[0][C0_ENTRYLO1]._u32[0];		// Clear Global bit?
	g_TLBs[i].pfne		= g_qwCPR[0][C0_ENTRYLO0]._u32[0];		// Clear Global bit?

	g_TLBs[i].g = g_qwCPR[0][C0_ENTRYLO1]._u32[0] &
		          g_qwCPR[0][C0_ENTRYLO0]._u32[0] &
				  TLBLO_G;

	// Build the masks:
	g_TLBs[i].mask      =  g_TLBs[i].pagemask | (~TLBHI_VPN2MASK); // 0x1fff
	g_TLBs[i].mask2     =  g_TLBs[i].mask>>1;
	g_TLBs[i].vpnmask   = ~g_TLBs[i].mask;
	g_TLBs[i].vpn2mask  =  g_TLBs[i].vpnmask>>1;

	g_TLBs[i].addrcheck = g_TLBs[i].hi & g_TLBs[i].vpnmask;

	g_TLBs[i].pfnehi = ((g_TLBs[i].pfne<<TLBLO_PFNSHIFT) & g_TLBs[i].vpn2mask);
	g_TLBs[i].pfnohi = ((g_TLBs[i].pfno<<TLBLO_PFNSHIFT) & g_TLBs[i].vpn2mask);


	switch (g_TLBs[i].pagemask)
	{
		case TLBPGMASK_4K:	// 4k 
			DPF(DEBUG_TLB, "       4k Pagesize");
            g_TLBs[i].checkbit = 0x00001000;   // bit 12
            break;
        case TLBPGMASK_16K: //  16k pagesize 
			DPF(DEBUG_TLB, "       16k Pagesize");
            g_TLBs[i].checkbit = 0x00004000;   // bit 14
            break;
        case TLBPGMASK_64K: //  64k pagesize
			DPF(DEBUG_TLB, "       64k Pagesize");
            g_TLBs[i].checkbit = 0x00010000;   // bit 16
            break;
        case 0x0007e000: // 256k pagesize
			DPF(DEBUG_TLB, "       256k Pagesize");
            g_TLBs[i].checkbit = 0x00040000;   // bit 18
            break;
        case 0x001fe000: //   1M pagesize
			DPF(DEBUG_TLB, "       1M Pagesize");
            g_TLBs[i].checkbit = 0x00100000;   // bit 20
            break;
        case 0x007fe000: //   4M pagesize
			DPF(DEBUG_TLB, "       4M Pagesize");
            g_TLBs[i].checkbit = 0x00400000;   // bit 22
            break;
        case 0x01ffe000: //  16M pagesize
			DPF(DEBUG_TLB, "       16M Pagesize");
            g_TLBs[i].checkbit = 0x01000000;   // bit 24
            break;
        default: // should not happen! 
			DPF(DEBUG_TLB, "       Unknown Pagesize");
            g_TLBs[i].checkbit = 0;
            break;
	}
}


void CALL_TYPE R4300_TLB_TLBP( CALL_ARGS ) 				// TLB Probe
{
	bool bFound = false;

	u32 dwEntryHi = g_qwCPR[0][C0_ENTRYHI]._u32[0];

	DPF( DEBUG_TLB, DSPrintf( "TLBP: ENTRYHI: 0x%08x", dwEntryHi ) );
	//DBGConsole_Msg(0, "TLBP: ENTRYHI: 0x%08x", dwEntryHi);

    for( u32 i = 0; i < 32; i++ )
	{
		if( ((g_TLBs[i].hi & TLBHI_VPN2MASK) ==
		     (dwEntryHi    & TLBHI_VPN2MASK)) &&
			(
				 (g_TLBs[i].g) ||
				((g_TLBs[i].hi & TLBHI_PIDMASK) ==
				 (dwEntryHi    & TLBHI_PIDMASK))
			) ) {
				DPF( DEBUG_TLB, DSPrintf( "   Found matching TLB Entry - 0x%04x", i ) );
				g_qwCPR[0][C0_INX]._u64 = i;
				bFound = true;
				break;
            }       
    }

	if (!bFound)
	{
		//DBGConsole_Msg(0, "   No matching TLB Entry Found for 0x%08x", dwEntryHi);
		DPF( DEBUG_TLB, DSPrintf( "   No matching TLB Entry Found for 0x%08x", dwEntryHi ) );
		g_qwCPR[0][C0_INX]._u64 = TLBINX_PROBE;

		//CPUHalt();
	}
	else
	{
		//DBGConsole_Msg(0, "   Matching TLB Entry Found for 0x%08x", dwEntryHi);
		//CPUHalt();
	}

}

void CALL_TYPE R4300_TLB_ERET( CALL_ARGS )
{
	if( g_qwCPR[0][C0_SR]._u64 & SR_ERL )
	{
		// Returning from an error trap
		DPF(DEBUG_INTR, "ERET: Returning from error trap");
		CPU_SetPC( g_qwCPR[0][C0_ERROR_EPC]._u32[0] );
		g_qwCPR[0][C0_SR]._u64 &= ~SR_ERL;
	}
	else
	{
		DPF(DEBUG_INTR, "ERET: Returning from interrupt/exception");
		// Returning from an exception
		CPU_SetPC( g_qwCPR[0][C0_EPC]._u32[0] );
		g_qwCPR[0][C0_SR]._u64 &= ~SR_EXL;
	}
	// Point to previous instruction (as we increment the pointer immediately afterwards
	DECREMENT_PC();

	// Ensure we don't execute this in the delay slot
	g_nDelay = NO_DELAY;
	//CPU_AddJob(CPU_NO_DELAY);

	g_bLLbit = 0;
}

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////


void CALL_TYPE R4300_Cop1_Unk( CALL_ARGS )     { WARN_NOEXIST("R4300_Cop1_Unk"); }
void CALL_TYPE R4300_Cop1_BCInstr( CALL_ARGS ) { R4300Cop1BC1Instruction[R4300_BC( gCurrentOp._u32 )]( ); }
void CALL_TYPE R4300_Cop1_SInstr( CALL_ARGS )  { R4300Cop1SInstruction[ gCurrentOp.funct ]( ); }
void CALL_TYPE R4300_Cop1_DInstr( CALL_ARGS )  { R4300Cop1DInstruction[ gCurrentOp.funct ]( ); }

void CALL_TYPE R4300_Cop1_LInstr( CALL_ARGS )
{
	switch ( gCurrentOp.funct )
	{
		case OP_CVT_S:
			R4300_Cop1_L_CVT_S( );
			return;
		case OP_CVT_D:
			R4300_Cop1_L_CVT_D( );
			return;
	}
}


void CALL_TYPE R4300_Cop1_WInstr( CALL_ARGS )
{
	switch ( gCurrentOp.funct )
	{
		case OP_CVT_S:
			R4300_Cop1_W_CVT_S( );
			return;
		case OP_CVT_D:
			R4300_Cop1_W_CVT_D( );
			return;
	}
	WARN_NOEXIST("R4300_Cop1_WInstr_Unk");
}



void CALL_TYPE R4300_Cop1_MTC1( CALL_ARGS )
{
	// Manual says top bits undefined after load
	StoreFPR_Word( gCurrentOp.fs,  gGPR[ gCurrentOp.rt ]._s32[0] );
}

void CALL_TYPE R4300_Cop1_DMTC1( CALL_ARGS )
{
	// Manual says top bits undefined after load
	StoreFPR_Long( gCurrentOp.fs, gGPR[ gCurrentOp.rt ]._u64 );
}


void CALL_TYPE R4300_Cop1_MFC1( CALL_ARGS )
{
	// MFC1 in the manual says this is a sign-extended result
	gGPR[ gCurrentOp.rt ]._s64 = (s64)(s32)LoadFPR_Word( gCurrentOp.fs );
	CHECK_R0();

}

void CALL_TYPE R4300_Cop1_DMFC1( CALL_ARGS )
{
	// MFC1 in the manual says this is a sign-extended result
	gGPR[ gCurrentOp.rt ]._u64 = LoadFPR_Long( gCurrentOp.fs );
	CHECK_R0();
}


void CALL_TYPE R4300_Cop1_CFC1( CALL_ARGS ) 		// move Control word From Copro 1
{
	// Only defined for reg 0 or 31
	if ( gCurrentOp.fs == 0 || gCurrentOp.fs == 31 )
	{
		gGPR[ gCurrentOp.rt ]._s64 = (s64)g_qwCCR[1][ gCurrentOp.fs ]._s32[0];
		CHECK_R0();

		//if ( gCurrentOp.fs == 0 )
		//{
		//	DBGConsole_Msg( 0, "Reading FCR00 - %08x at %08x", gGPR[ gCurrentOp.rt ]._u32[ 0 ], g_dwPC  );
		//}
		//else
		//{
		//	DBGConsole_Msg( 0, "Reading FCR32 - %08x at %08x", gGPR[ gCurrentOp.rt ]._u32[ 0 ], g_dwPC  );
		//}
	}
	else
	{
	}
}

void CALL_TYPE R4300_Cop1_CTC1( CALL_ARGS ) 		// move Control word To Copro 1
{
	// Only defined for reg 0 or 31
	// TODO - Maybe an exception was raised?
	if ( gCurrentOp.fs == 0 )
	{
		g_qwCCR[1][ gCurrentOp.fs ]._u64 = gGPR[ gCurrentOp.rt ]._u64;

	}
	else if ( gCurrentOp.fs == 31 )
	{
		g_qwCCR[1][ gCurrentOp.fs ] = gGPR[ gCurrentOp.rt ];

		u32		fpcr( g_qwCCR[1][ gCurrentOp.fs ]._u32[ 0 ] );

		/*
		if ( fpcr & FPCSR_FS )
		{
			DBGConsole_Msg( 0, "FCR31: Flush denormalised results to 0, no exception" );
		}
		else
		{
			DBGConsole_Msg( 0, "FCR31: denormalised results cause exception" );
		}

		u32		fp_cause(  fpcr & ( FPCSR_CE|FPCSR_CV|FPCSR_CZ|FPCSR_CO|FPCSR_CU|FPCSR_CI ) );
		u32		fp_enable( fpcr & (	         FPCSR_EV|FPCSR_EZ|FPCSR_EO|FPCSR_EU|FPCSR_EI ) );
		u32		fp_flags(  fpcr & (          FPCSR_FV|FPCSR_FZ|FPCSR_FO|FPCSR_FU|FPCSR_FI ) );
	
		if ( fp_cause || fp_enable || fp_flags )
		{
			DBGConsole_Msg( 0, "FPCR: Cause %08x, Enable %08x, Flags %08x",
				(fp_cause >> 12), (fp_enable>>7), (fp_flags>>2) );
		}
		*/

		switch ( fpcr & FPCSR_RM_MASK )
		{
		case FPCSR_RM_RN:		gRoundingMode = RM_NEAR;	break;
		case FPCSR_RM_RZ:		gRoundingMode = RM_CHOP;	break;
		case FPCSR_RM_RP:		gRoundingMode = RM_UP;		break;
		case FPCSR_RM_RM:		gRoundingMode = RM_DOWN;	break;
		default:				NODEFAULT;
		}

		SET_ROUND_MODE( gRoundingMode );
	}
	else
	{

	}

	// Now generate lots of exceptions :-)
}




void CALL_TYPE R4300_BC1_BC1F( CALL_ARGS )		// Branch on FPU False
{
	if ( !(g_qwCCR[1][31]._u64 & FPCSR_C) )
	{
		g_dwNewPC = g_dwPC + (s32)(s16)gCurrentOp.immediate*4 + 4;
		CPU_Delay();

	} 
}

void CALL_TYPE R4300_BC1_BC1T( CALL_ARGS )	// Branch on FPU True
{
	if ( g_qwCCR[1][31]._u64 & FPCSR_C )
	{
		g_dwNewPC = g_dwPC + (s32)(s16)gCurrentOp.immediate*4 + 4;
		CPU_Delay();
	} 
}

void CALL_TYPE R4300_BC1_BC1FL( CALL_ARGS )	// Branch on FPU False Likely

{
	if ( !(g_qwCCR[1][31]._u64 & FPCSR_C) )
	{
		g_dwNewPC = g_dwPC + (s32)(s16)gCurrentOp.immediate*4 + 4;
		CPU_Delay();
	}
	else
	{
		// Don't execute subsequent instruction
		INCREMENT_PC();
	}
}

void CALL_TYPE R4300_BC1_BC1TL( CALL_ARGS )		// Branch on FPU True Likely
{
	if ( g_qwCCR[1][31]._u64 & FPCSR_C )
	{
		g_dwNewPC = g_dwPC + (s32)(s16)gCurrentOp.immediate*4 + 4;
		CPU_Delay();
	}
	else
	{
		// Don't execute subsequent instruction
		INCREMENT_PC();
	}
}


/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// WORD FP Instrs /////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////


void CALL_TYPE R4300_Cop1_W_CVT_S( CALL_ARGS )
{	
	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	s32 nTemp = LoadFPR_Word( gCurrentOp.fs );

	StoreFPR_Single( gCurrentOp.fd, s32_to_f32( nTemp ) );
}

void CALL_TYPE R4300_Cop1_W_CVT_D( CALL_ARGS )
{
	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	s32 nTemp = LoadFPR_Word( gCurrentOp.fs );

	// Convert using current rounding mode?

	StoreFPR_Double( gCurrentOp.fd, s32_to_f64( nTemp ) );
}

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// LONG FP Instrs /////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////


void CALL_TYPE R4300_Cop1_L_CVT_S( CALL_ARGS )
{	
	s64 nTemp = LoadFPR_Long( gCurrentOp.fs );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, s64_to_f32( nTemp ));
}

void CALL_TYPE R4300_Cop1_L_CVT_D( CALL_ARGS )
{
	s64 nTemp = LoadFPR_Long( gCurrentOp.fs );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd, s64_to_f64( nTemp ) );
}


/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////// Single FP Instrs //////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////


void CALL_TYPE R4300_Cop1_S_Unk( CALL_ARGS ) { WARN_NOEXIST("R4300_Cop1_S_Unk"); }


void CALL_TYPE R4300_Cop1_S_ADD( CALL_ARGS )
{
	// fd = fs+ft
	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	f32 fY = LoadFPR_Single( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, fX + fY );
}

void CALL_TYPE R4300_Cop1_S_SUB( CALL_ARGS )
{
	// fd = fs-ft
	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	f32 fY = LoadFPR_Single( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, fX - fY );
}

void CALL_TYPE R4300_Cop1_S_MUL( CALL_ARGS )
{
	// fd = fs*ft
	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	f32 fY = LoadFPR_Single( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, fX * fY );
}

void CALL_TYPE R4300_Cop1_S_DIV( CALL_ARGS )
{
	// fd = fs/ft
	f32 fDividend = LoadFPR_Single( gCurrentOp.fs );
	f32 fDivisor  = LoadFPR_Single( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	if ( fDivisor == 0 )
	{
		if ( g_qwCCR[1][ 31 ]._u32[ 0 ] & FPCSR_EZ )
		{
			// Exception
			DBGConsole_Msg( 0, "[MShould trigger FPU exception for /0 here" );
		}
		else
		{
			//DBGConsole_Msg( 0, "Float divide by zero, setting flag" );
			g_qwCCR[1][ 31 ]._u32[ 0 ] |= FPCSR_FZ;
		}
	}

	StoreFPR_Single( gCurrentOp.fd, fDividend / fDivisor );
}

void CALL_TYPE R4300_Cop1_S_SQRT( CALL_ARGS )
{
	// fd = sqrt(fs)
	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, sqrtf(fX) );
}


void CALL_TYPE R4300_Cop1_S_NEG( CALL_ARGS )
{
	// fd = -(fs)
	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, -fX );
}

void CALL_TYPE R4300_Cop1_S_MOV( CALL_ARGS )
{
	// fd = fs
	f32 fValue = LoadFPR_Single( gCurrentOp.fs );

	// Just copy bits directly?
	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, fValue );
}

void CALL_TYPE R4300_Cop1_S_ABS( CALL_ARGS )
{
	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, fabsf(fX) );
}


void CALL_TYPE R4300_Cop1_S_TRUNC_W( CALL_ARGS )
{
	SET_ROUND_MODE( RM_CHOP );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f32_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_S_TRUNC_L( CALL_ARGS )
{
	SET_ROUND_MODE( RM_CHOP );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f32_to_s64( fX ) );
}


void CALL_TYPE R4300_Cop1_S_ROUND_W( CALL_ARGS )
{
	SET_ROUND_MODE( RM_NEAR );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f32_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_S_ROUND_L( CALL_ARGS )
{
	SET_ROUND_MODE( RM_NEAR );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f32_to_s64( fX ) );
}


void CALL_TYPE R4300_Cop1_S_CEIL_W( CALL_ARGS )
{
	SET_ROUND_MODE( RM_UP );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f32_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_S_CEIL_L( CALL_ARGS )
{
	SET_ROUND_MODE( RM_UP );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f32_to_s64( fX ) );
}

void CALL_TYPE R4300_Cop1_S_FLOOR_W( CALL_ARGS )
{
	SET_ROUND_MODE( RM_DOWN );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f32_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_S_FLOOR_L( CALL_ARGS )
{
	SET_ROUND_MODE( RM_DOWN );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f32_to_s64( fX ) );
}


// Convert single to long - this is used by WarGods
void CALL_TYPE R4300_Cop1_S_CVT_L( CALL_ARGS )
{
	SET_ROUND_MODE( gRoundingMode );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f32_to_s64( fX ) );
}

// Convert single to word...
void CALL_TYPE R4300_Cop1_S_CVT_W( CALL_ARGS )
{
	SET_ROUND_MODE( gRoundingMode );

	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	s32	sX = f32_to_s32( fX );

	StoreFPR_Word( gCurrentOp.fd, sX );
}

// Convert single to f64...
void CALL_TYPE R4300_Cop1_S_CVT_D( CALL_ARGS )
{
	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	f32 fX = LoadFPR_Single( gCurrentOp.fs );

	StoreFPR_Double( gCurrentOp.fd, f32_to_f64( fX ) );
}


void CALL_TYPE R4300_Cop1_S_EQ( CALL_ARGS ) 				// Compare for Equality
{
	// fs == ft?
	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	f32 fY = LoadFPR_Single( gCurrentOp.ft );

	if ( fX == fY )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;
}

void CALL_TYPE R4300_Cop1_S_LT( CALL_ARGS ) 				// Compare for Equality
{
	// fs < ft?
	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	f32 fY = LoadFPR_Single( gCurrentOp.ft );

	if ( fX < fY )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;
}



void CALL_TYPE R4300_Cop1_C_S_Generic( CALL_ARGS )
{
	// fs < ft?
	bool less;
	bool equal;
	bool unordered;
	bool cond;
	bool cond0 = (gCurrentOp._u32   ) & 0x1;
	bool cond1 = (gCurrentOp._u32>>1) & 0x1;
	bool cond2 = (gCurrentOp._u32>>2) & 0x1;
	bool cond3 = (gCurrentOp._u32>>3) & 0x1;

	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	f32 fY = LoadFPR_Single( gCurrentOp.ft );

	if (_isnan(fX) || _isnan(fY))
	{
		//DBGConsole_Msg(0, "is nan");
		less = false;
		equal = false;
		unordered = true;

		// exception
		if (cond3)
		{
			// Exception
			DBGConsole_Msg( 0, "[MShould throw fp nan exception?" );
		}
	}
	else
	{
		less  = (fX < fY);
		equal = (fX == fY);
		unordered = false;	
	}

	cond = ((cond0 && unordered) ||
			(cond1 && equal) ||
			(cond2 && less));

	if ( cond )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;

}

// Same as above, but unordered?
// cond3 = 1
// cond2 = 1
// cond1 = 0
// cond0 = 1
void CALL_TYPE R4300_Cop1_S_NGE( CALL_ARGS )
{
	// fs < ft?
	bool less;
	bool unordered;
	bool cond;

	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	f32 fY = LoadFPR_Single( gCurrentOp.ft );

	if (_isnan(fX) || _isnan(fY))
	{
		less = false;
		unordered = true;

		// exception
	}
	else
	{
		less = ( fX < fY );
		unordered = false;
	}

	cond = (unordered || less);

	if ( cond )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;

}

void CALL_TYPE R4300_Cop1_S_LE( CALL_ARGS ) 				// Compare for Equality
{
	// fs <= ft?
	f32 fX = LoadFPR_Single( gCurrentOp.fs );
	f32 fY = LoadFPR_Single( gCurrentOp.ft );

	// SRInliner didn't like two rets (I think). 
	// VC6++ with SP4 generates 2 rets with the commented code
	if ( fX <= fY )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;

}


void CALL_TYPE R4300_Cop1_S_F( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_UN( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_UEQ( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_OLT( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_ULT( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_OLE( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_ULE( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_SF( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_NGLE( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_SEQ( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_NGL( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }
void CALL_TYPE R4300_Cop1_S_NGT( CALL_ARGS ) { R4300_Cop1_C_S_Generic( ); }




/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////


void CALL_TYPE R4300_Cop1_D_Unk( CALL_ARGS )		{ WARN_NOEXIST("R4300_Cop1_D_Unk"); }


void CALL_TYPE R4300_Cop1_D_ABS( CALL_ARGS )
{
	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd, fabs(fX) );
}

void CALL_TYPE R4300_Cop1_D_ADD( CALL_ARGS )
{
	// fd = fs+ft
	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	f64 fY = LoadFPR_Double( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd, fX + fY );
}

void CALL_TYPE R4300_Cop1_D_SUB( CALL_ARGS )
{
	// fd = fs-ft
	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	f64 fY = LoadFPR_Double( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd, fX - fY );
}

void CALL_TYPE R4300_Cop1_D_MUL( CALL_ARGS )
{
	// fd = fs*ft
	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	f64 fY = LoadFPR_Double( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd, fX * fY );
}

void CALL_TYPE R4300_Cop1_D_DIV( CALL_ARGS )
{
	// fd = fs/ft
	f64 fDividend = LoadFPR_Double( gCurrentOp.fs );
	f64 fDivisor = LoadFPR_Double( gCurrentOp.ft );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd,  fDividend / fDivisor );
}

void CALL_TYPE R4300_Cop1_D_SQRT( CALL_ARGS )
{
	// fd = sqrt(fs)
	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd, sqrt(fX) );
}


void CALL_TYPE R4300_Cop1_D_NEG( CALL_ARGS )
{
	// fd = -(fs)
	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd, -fX );
}

void CALL_TYPE R4300_Cop1_D_MOV( CALL_ARGS )
{
	// fd = fs
	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	
	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Double( gCurrentOp.fd, fX );

	// TODO - Just copy bits - don't interpret!
}

void CALL_TYPE R4300_Cop1_D_TRUNC_W( CALL_ARGS )
{
	SET_ROUND_MODE( RM_CHOP );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f64_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_D_TRUNC_L( CALL_ARGS )
{
	SET_ROUND_MODE( RM_CHOP );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f64_to_s64( fX ) );
}


void CALL_TYPE R4300_Cop1_D_ROUND_W( CALL_ARGS )
{
	SET_ROUND_MODE( RM_NEAR );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f64_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_D_ROUND_L( CALL_ARGS )
{
	SET_ROUND_MODE( RM_NEAR );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f64_to_s64( fX ) );
}


void CALL_TYPE R4300_Cop1_D_CEIL_W( CALL_ARGS )
{
	SET_ROUND_MODE( RM_UP );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f64_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_D_CEIL_L( CALL_ARGS )
{
	SET_ROUND_MODE( RM_UP );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f64_to_s64( fX ) );
}

void CALL_TYPE R4300_Cop1_D_FLOOR_W( CALL_ARGS )
{
	SET_ROUND_MODE( RM_DOWN );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f64_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_D_FLOOR_L( CALL_ARGS )
{
	SET_ROUND_MODE( RM_DOWN );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f64_to_s64( fX ) );
}

void CALL_TYPE R4300_Cop1_D_CVT_S( CALL_ARGS )
{	
	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	
	SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	StoreFPR_Single( gCurrentOp.fd, f64_to_f32( fX ) );
}

// Convert f64 to word...
void CALL_TYPE R4300_Cop1_D_CVT_W( CALL_ARGS )
{	
	SET_ROUND_MODE( gRoundingMode );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Word( gCurrentOp.fd, f64_to_s32( fX ) );
}

void CALL_TYPE R4300_Cop1_D_CVT_L( CALL_ARGS )
{
	SET_ROUND_MODE( gRoundingMode );

	f64 fX = LoadFPR_Double( gCurrentOp.fs );

	StoreFPR_Long( gCurrentOp.fd, f64_to_s64( fX ) );
}

void CALL_TYPE R4300_Cop1_C_D_Generic( CALL_ARGS )
{
	// fs < ft?
	bool less;
	bool equal;
	bool unordered;
	bool cond;
	bool cond0 = (gCurrentOp._u32   ) & 0x1;
	bool cond1 = (gCurrentOp._u32>>1) & 0x1;
	bool cond2 = (gCurrentOp._u32>>2) & 0x1;
	bool cond3 = (gCurrentOp._u32>>3) & 0x1;

	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	f64 fY = LoadFPR_Double( gCurrentOp.ft );

	if (_isnan(fX) || _isnan(fY))
	{
		//DBGConsole_Msg(0, "is nan");
		less = false;
		equal = false;
		unordered = true;

		// exception
		if (cond3)
		{
			// Exception
			DBGConsole_Msg( 0, "[MShould throw fp nan exception?" );
		}
	}
	else
	{
		less  = (fX < fY);
		equal = (fX == fY);
		unordered = false;	
	}

	cond = ((cond0 && unordered) ||
			(cond1 && equal) ||
			(cond2 && less));

	if ( cond )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;

}


void CALL_TYPE R4300_Cop1_D_EQ( CALL_ARGS )				// Compare for Equality
{
	// fs == ft?
	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	f64 fY = LoadFPR_Double( gCurrentOp.ft );

	if( fX == fY )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;
}

void CALL_TYPE R4300_Cop1_D_LE( CALL_ARGS )				// Compare for Less Than or Equal
{

	// fs <= ft?
	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	f64 fY = LoadFPR_Double( gCurrentOp.ft );

	if( fX <= fY )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;

}


void CALL_TYPE R4300_Cop1_D_LT( CALL_ARGS )
{
	// fs < ft?
	f64 fX = LoadFPR_Double( gCurrentOp.fs );
	f64 fY = LoadFPR_Double( gCurrentOp.ft );

	if( fX < fY )
		g_qwCCR[1][31]._u64 |= FPCSR_C;
	else
		g_qwCCR[1][31]._u64 &= ~FPCSR_C;

}


void CALL_TYPE R4300_Cop1_D_F( CALL_ARGS )		{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_UN( CALL_ARGS )		{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_UEQ( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_OLT( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_ULT( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_OLE( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_ULE( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_SF( CALL_ARGS )		{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_NGLE( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_SEQ( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_NGL( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_NGE( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
void CALL_TYPE R4300_Cop1_D_NGT( CALL_ARGS )	{ R4300_Cop1_C_D_Generic( ); }
