/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __REGISTERS_H__
#define __REGISTERS_H__

#define R4300_BC(x) ((x>>16)&0x3)					// For BC1 Stuff

/*
    CPU: Instructions encoded by opcode field.
    31---------26---------------------------------------------------0
    |  opcode   |                                                   |
    ------6----------------------------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 | *1    | *2    | J     | JAL   | BEQ   | BNE   | BLEZ  | BGTZ  |
001 | ADDI  | ADDIU | SLTI  | SLTIU | ANDI  | ORI   | XORI  | LUI   |
010 | *3    | *4    |  ---  |  ---  | BEQL  | BNEL  | BLEZL | BGTZL |
011 | DADDI |DADDIU | LDL   | LDR   |  xxx  |  xxx  |  xxx  |  xxx  | top 4 bits == 7
100 | LB    | LH    | LWL   | LW    | LBU   | LHU   | LWR   | LWU   |
101 | SB    | SH    | SWL   | SW    | SDL   | SDR   | SWR   | CACHE |
110 | LL    | LWC1  |  ---  |  ---  | LLD   | LDC1  | LDC2  | LD    |
111 | SC    | SWC1  |  xxx  |  ---  | SCD   | SDC1  | SDC2  | SD    |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
     *1 = SPECIAL, see SPECIAL list    *2 = REGIMM, see REGIMM list
     *3 = COP0                         *4 = COP1
*/

#define OP_IS_A_HACK(x) (((x) >> 28) == 0x7)


enum OpCodeValue {
	OP_SPECOP		= 0,
	OP_REGIMM		= 1,
	OP_J			= 2,
	OP_JAL			= 3,
	OP_BEQ			= 4,
	OP_BNE			= 5,
	OP_BLEZ		= 6,
	OP_BGTZ		= 7,
	OP_ADDI		= 8,
	OP_ADDIU		= 9,
	OP_SLTI		= 10,
	OP_SLTIU		= 11,
	OP_ANDI		= 12,
	OP_ORI			= 13,
	OP_XORI		= 14,
	OP_LUI			= 15,
	OP_COPRO0		= 16,
	OP_COPRO1		= 17,
/*	OP_SRHACK1		= 18,
	OP_SRHACK2		= 19,*/
	OP_BEQL		= 20,
	OP_BNEL		= 21,
	OP_BLEZL		= 22,
	OP_BGTZL		= 23,
	OP_DADDI		= 24,
	OP_DADDIU		= 25,
	OP_LDL			= 26,
	OP_LDR			= 27,
	OP_PATCH		= 28,
	OP_SRHACK_UNOPT = 29,
	OP_SRHACK_OPT   = 30,
	OP_SRHACK_NOOPT = 31,
/*	OP_UNK4		= 29,
	OP_UNK5		= 30,
	OP_UNK6		= 31,*/
	OP_LB			= 32,
	OP_LH			= 33,
	OP_LWL			= 34,
	OP_LW 			= 35,
	OP_LBU			= 36,
	OP_LHU			= 37,
	OP_LWR			= 38,
	OP_LWU			= 39,
	OP_SB			= 40,
	OP_SH			= 41,
	OP_SWL			= 42,
	OP_SW			= 43,
	OP_SDL			= 44,
	OP_SDR			= 45,
	OP_SWR			= 46,
	OP_CACHE		= 47,
	OP_LL			= 48,
	OP_LWC1			= 49,
	OP_UNK7			= 50,
	OP_UNK8			= 51,
	OP_LLD			= 52,
	OP_LDC1			= 53,
	OP_LDC2 		= 54,
	OP_LD			= 55,
	OP_SC 			= 56,
	OP_SWC1 		= 57,
	OP_DBG_BKPT		= 58,
	OP_UNK10 		= 59,
	OP_SCD 			= 60,
	OP_SDC1 		= 61,
	OP_SDC2 		= 62,
	OP_SD			= 63
};

enum SpecialOpCodeValue {
	OP_SLL					= 0,
	OP_UNK11				= 1,
	OP_SRL					= 2,
	OP_SRA					= 3,
	OP_SLLV				= 4,
	OP_UNK12				= 5,
	OP_SRLV				= 6,
	OP_SRAV				= 7,
	OP_JR					= 8,
	OP_JALR				= 9,
	OP_UNK13				= 10,
	OP_UNK14				= 11,
	OP_SYSCALL				= 12,
	OP_BREAK				= 13,
	OP_UNK15				= 14,
	OP_SYNC				= 15,
	OP_MFHI				= 16,
	OP_MTHI				= 17,
	OP_MFLO				= 18,
	OP_MTLO				= 19,
	OP_DSLLV				= 20,
	OP_UNK16				= 21,
	OP_DSRLV				= 22,
	OP_DSRAV				= 23,
	OP_MULT				= 24,
	OP_MULTU				= 25,
	OP_DIV					= 26,
	OP_DIVU				= 27,
	OP_DMULT				= 28,
	OP_DMULTU				= 29,
	OP_DDIV				= 30,
	OP_DDIVU				= 31,
	OP_ADD					= 32,
	OP_ADDU				= 33,
	OP_SUB					= 34,
	OP_SUBU				= 35,
	OP_AND					= 36,
	OP_OR					= 37,
	OP_XOR					= 38,
	OP_NOR					= 39,
	OP_UNK17				= 40,
	OP_UNK18				= 41,
	OP_SLT					= 42,
	OP_SLTU				= 43,
	OP_DADD				= 44,
	OP_DADDU				= 45,
	OP_DSUB				= 46,
	OP_DSUBU				= 47,
	OP_TGE					= 48,
	OP_TGEU				= 49,
	OP_TLT					= 50,
	OP_TLTU				= 51,
	OP_TEQ					= 52,
	OP_UNK19				= 53,
	OP_TNE					= 54,
	OP_UNK20				= 55,
	OP_DSLL				= 56,
	OP_UNK21				= 57,
	OP_DSRL				= 58,
	OP_DSRA				= 59,
	OP_DSLL32				= 60,
	OP_UNK22				= 61,
	OP_DSRL32				= 62,
	OP_DSRA32				= 63
};

enum RegImmOpCodeValue {
	OP_BLTZ				= 0,
	OP_BGEZ				= 1,
	OP_BLTZL				= 2,
	OP_BGEZL				= 3,
	OP_TGEI				= 8,
	OP_TGEIU				= 9,
	OP_TLTI				= 10,
	OP_TLTIU				= 11,
	OP_TEQI				= 12,
	OP_TNEI				= 14,
	OP_BLTZAL				= 16,
	OP_BGEZAL				= 17,
	OP_BLTZALL				= 18,
	OP_BGEZALL				= 19
};

enum Cop0OpCodeValue {
	OP_MFC0	= 0,
	OP_MTC0	= 4,
	OP_TLB		= 16
};

enum TLBOpCodeValue {
	OP_TLBR = 1,
	OP_TLBWI = 2,
	OP_TLBWR = 6,
	OP_TLBBP = 8,
	OP_ERET = 24
};

enum Cop1OpCodeValue {
	OP_MFC1	= 0,
	OP_DMFC1	= 1,
	OP_CFC1	= 2,
	OP_MTC1	= 4,
	OP_DMTC1	= 5,
	OP_CTC1	= 6,

	BCInstr = 8,
	OP_SInstr = 16,
	OP_DInstr = 17,
	OP_WInstr = 20,
	OP_LInstr = 21

};
#define	OP_CVT_S 32
#define	OP_CVT_D 33


extern const char *Cop1WOpCodeNames[64];
extern const char *Cop1LOpCodeNames[64];


enum Cop1BC1OpCodeValue {
	OP_BC1F = 0,
	OP_BC1T = 1,
	OP_BC1FL = 2,
	OP_BC1TL = 3
};


// Register names

/* Shared by RSP */
extern const char *RegNames[32];
extern const char *Cop0RegNames[32];
extern const char *ShortCop0RegNames[32];

#endif
