/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "DBGConsole.h"

#include <pspdebug.h>

class IDebugConsole : public CDebugConsole
{
	public:
		IDebugConsole();
		virtual ~IDebugConsole();

		void			EnableConsole( bool bEnable );
		bool			IsVisible() const;

		void			UpdateDisplay();

		void			DisplayProfile() ;

		void __cdecl	Msg(u32 type, const TCHAR * szFormat, ...);

		void			MsgOverwriteStart();
		void __cdecl	MsgOverwrite(u32 type, const TCHAR * szFormat, ...);
		void			MsgOverwriteEnd();
		void __cdecl	Stats( StatType stat, const TCHAR * szFormat, ...)			{}

	static u32			GetStringHighlight(char c);
	static void			DisplayString( const char * p_string, u32 default_colour );

private:
			bool		mEnabled;
			u32			mCurrentY;

	static	char		mFormattingBuffer[ 2048 ];

};


//*****************************************************************************
//
//*****************************************************************************
char	IDebugConsole::mFormattingBuffer[ 2048 ];

//*****************************************************************************
//
//*****************************************************************************
template<> bool	CSingleton< CDebugConsole >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new IDebugConsole();
	
	return true;
}


//*****************************************************************************
//
//*****************************************************************************
u32 IDebugConsole::GetStringHighlight(char c)
{
	const u8 NORMAL = 200;
	const u8 BRIGHT = 255;

	switch(c)
	{
		case 'r': return MAKE_COLOUR_U32( NORMAL, 0, 0 );
		case 'g': return MAKE_COLOUR_U32( 0, NORMAL, 0 );
		case 'b': return MAKE_COLOUR_U32( 0, 0, NORMAL );
		case 'c': return MAKE_COLOUR_U32( 0, NORMAL, NORMAL );
		case 'm': return MAKE_COLOUR_U32( NORMAL, 0, NORMAL );
		case 'y': return MAKE_COLOUR_U32( NORMAL, NORMAL, 0 );
		case 'w': return MAKE_COLOUR_U32( NORMAL, NORMAL, NORMAL );
		case 'R': return MAKE_COLOUR_U32( BRIGHT, 0, 0 );
		case 'G': return MAKE_COLOUR_U32( 0, BRIGHT, 0 );
		case 'B': return MAKE_COLOUR_U32( 0, 0, BRIGHT );
		case 'C': return MAKE_COLOUR_U32( 0, BRIGHT, BRIGHT );
		case 'M': return MAKE_COLOUR_U32( BRIGHT, 0, BRIGHT );
		case 'Y': return MAKE_COLOUR_U32( BRIGHT, BRIGHT, 0 );
		case 'W': return MAKE_COLOUR_U32( BRIGHT, BRIGHT, BRIGHT );
		default: return 0;		// Colour of 0 indicates invalid highlight char
	}
}

//*****************************************************************************
//
//*****************************************************************************
void IDebugConsole::DisplayString( const char * p_string, u32 default_colour )
{
	pspDebugScreenSetTextColor( default_colour );

	char		current_string[ 1024 ];
	u32			out_idx( 0 );
	u32 string_len( strlen(p_string) );
	for ( u32 in_idx = 0; in_idx < string_len; ++in_idx )
	{
		if (p_string[in_idx] == '[')
		{
			u32 highlight = GetStringHighlight(p_string[in_idx+1]);
			if(highlight != 0)
			{
				// Flush the current string and update the colour
				current_string[out_idx] = 0;
				pspDebugScreenPrintf( "%s", current_string );
				out_idx = 0;
				pspDebugScreenSetTextColor( highlight );
			}
			else
			{
				switch (p_string[in_idx+1])
				{
				case '[':
				case ']':
					current_string[out_idx] = p_string[in_idx+1];
					out_idx++;
					break;
				}
			}

			// Skip colour character
			in_idx++;
		}
		else if (p_string[in_idx] == ']')
		{
			// Flush the current string and update the colour
			current_string[out_idx] = 0;
			pspDebugScreenPrintf( "%s", current_string );
			out_idx = 0;
			pspDebugScreenSetTextColor( default_colour );
		}
		else
		{
			current_string[out_idx] = p_string[in_idx];
			out_idx++;
		}
	}

	// Flush the current string and restore the colour
	current_string[out_idx] = 0;
	pspDebugScreenPrintf( "%s", current_string );

	pspDebugScreenSetTextColor( MAKE_COLOUR_U32( 255,255,255 ) );
}

//*****************************************************************************
//
//*****************************************************************************
IDebugConsole::IDebugConsole() 
:	mEnabled( true )
,	mCurrentY( 0 )
{
	pspDebugScreenSetXY(0, 0);
	/*FILE * fh( fopen( g_szDaedalusLogFile, "w" ) );
	if(fh)
	{
		fclose(fh);
	}*/
}

//*****************************************************************************
//
//*****************************************************************************
IDebugConsole::~IDebugConsole()
{
}

//*****************************************************************************
//
//*****************************************************************************
void	IDebugConsole::EnableConsole( bool bEnable )
{
	if( !mEnabled && bEnable )
	{
		pspDebugScreenSetXY(0, 0);
	}
	mEnabled = bEnable;
}

//*****************************************************************************
//
//*****************************************************************************
bool	IDebugConsole::IsVisible() const
{
	return mEnabled;
}

//*****************************************************************************
//
//*****************************************************************************
void	IDebugConsole::UpdateDisplay()
{
}

//*****************************************************************************
//
//*****************************************************************************
void	IDebugConsole::DisplayProfile() 
{
}

//*****************************************************************************
//
//*****************************************************************************
void __cdecl	IDebugConsole::Msg(u32 type, const TCHAR * szFormat, ...)
{
	if( !mEnabled )
	{
		return;
	}

	va_list			va;

	// Format the output
	va_start( va, szFormat );
	// Don't use wvsprintf as it doesn't handle floats!
	vsprintf( mFormattingBuffer, szFormat, va );
	va_end( va );

	Daedalus_ForceDebugScreenToFront();

	DisplayString( mFormattingBuffer, GetStringHighlight( 'w' ) );
	pspDebugScreenPrintf( "\n" );

	/*FILE * fh( fopen( g_szDaedalusLogFile, "a" ) );
	if(fh)
	{
		fprintf( fh, "%s\n", mFormattingBuffer );
		fclose(fh);
	}*/
}

//*****************************************************************************
//
//*****************************************************************************
void	IDebugConsole::MsgOverwriteStart()
{
	mCurrentY = pspDebugScreenGetY();
}

//*****************************************************************************
//
//*****************************************************************************
void __cdecl	IDebugConsole::MsgOverwrite(u32 type, const TCHAR * szFormat, ...)
{
	if( !mEnabled )
	{
		return;
	}

	va_list			va;

	// Format the output
	va_start( va, szFormat );
	// Don't use wvsprintf as it doesn't handle floats!
	vsprintf( mFormattingBuffer, szFormat, va );
	va_end( va );

	Daedalus_ForceDebugScreenToFront();
	pspDebugScreenSetXY(0, mCurrentY);

	DisplayString( mFormattingBuffer, GetStringHighlight( 'w' ) );
	pspDebugScreenPrintf( "\n" );

}

//*****************************************************************************
//
//*****************************************************************************
void	IDebugConsole::MsgOverwriteEnd()
{
}

