/*
Copyright (C) 2001,2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#if _MSC_VER > 1000
#pragma once
#endif

#ifndef ASSEMBLYBUFFER_H_
#define ASSEMBLYBUFFER_H_

class CAssemblyBuffer
{
	public:
		CAssemblyBuffer()
		:	mpCodeBuffer(NULL)
		,	mCurrentPos( 0 )
		{
		}

		inline void	PadTo16Bytes()
		{
			mCurrentPos = ((mCurrentPos - 1) & 0xfffffff0) + 0x10; // align to 16-byte boundary
		}

		inline void EmitBYTE(u8 byte)
		{
			mpCodeBuffer[mCurrentPos] = byte;
			mCurrentPos++;
		}

		inline void EmitWORD(u16 word)
		{
			*(u16 *)(&mpCodeBuffer[mCurrentPos]) = word;
			mCurrentPos += 2;
		}

		inline void EmitDWORD(u32 dword)
		{
			*(u32 *)(&mpCodeBuffer[mCurrentPos]) = dword;
			mCurrentPos += 4;
		}

		void EmitData( const void * pdata, u32 count )
		{
			memcpy( &mpCodeBuffer[ mCurrentPos ], pdata, count );
			mCurrentPos += count;
		}

		u8 *				GetStartAddress() const										{ return &mpCodeBuffer[0]; }
		u8 *				GetCurrentAddress() const									{ return &mpCodeBuffer[mCurrentPos]; }
		u32					GetSize() const												{ return mCurrentPos; }

		void				SetBuffer( u8 * pbuffer )
		{
			mpCodeBuffer = pbuffer;
			mCurrentPos = 0;
		}

protected:
	u8 *					mpCodeBuffer;
	u32						mCurrentPos;		// Current writing position in buffer
};

#endif // ASSEMBLYBUFFER_H_
