/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "AssemblyWriterPSP.h"

#include "Core\Registers.h"

#include "OSHLE\ultra_R4300.h"

#define MR_R0 REG_r0

//*****************************************************************************
//
//*****************************************************************************
void	CAssemblyWriterPSP::LoadConstant( EMipsReg reg, s32 value )
{
	OpCode				op_code;

	if ( value >= SHRT_MIN && value <= SHRT_MAX )
	{
		// ORI
		op_code.op = OP_ADDIU;
		op_code.rt = reg;
		op_code.rs = MR_R0;
		op_code.immediate = u16( value );
		AppendOp( op_code );
	}
	else
	{
		// It's too large - we need to load in two parts
		LUI( reg, u16(value >> 16) );

		op_code.op = OP_ORI;
		op_code.rt = reg;
		op_code.rs = reg;
		op_code.immediate = u16( value );

		//
		// No need to do this if low bits are clear!
		//
		if ( op_code.immediate != 0 )
		{
			AppendOp( op_code );
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CAssemblyWriterPSP::LoadRegister( EMipsReg reg_dst, OpCodeValue load_op, EMipsReg reg_base, s16 offset )
{
	OpCode		op_code;
	op_code._u32 = 0;

	op_code.op = load_op;
	op_code.rt = reg_dst;
	op_code.base = reg_base;
	op_code.immediate = offset;

	AppendOp( op_code );
}

//*****************************************************************************
//
//*****************************************************************************
void	CAssemblyWriterPSP::StoreRegister( EMipsReg reg_src, OpCodeValue store_op, EMipsReg reg_base, s16 offset )
{
	OpCode		op_code;
	op_code._u32 = 0;

	op_code.op = store_op;
	op_code.rt = reg_src;
	op_code.base = reg_base;
	op_code.immediate = offset;

	AppendOp( op_code );
}

//*****************************************************************************
//
//*****************************************************************************
void	CAssemblyWriterPSP::SW( EMipsReg reg_src, EMipsReg reg_base, s16 offset )
{
	StoreRegister( reg_src, OP_SW, reg_base, offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CAssemblyWriterPSP::LUI( EMipsReg reg, u16 value )
{
	OpCode				op_code;
	op_code.op = OP_LUI;
	op_code.rt = reg;
	op_code.rs = MR_R0;
	op_code.immediate = value;
	AppendOp( op_code );
}

//*****************************************************************************
//
//*****************************************************************************
void	CAssemblyWriterPSP::JAL( const void * p_target )
{
	const u8 *		p_target_location( reinterpret_cast< const u8 * >( p_target ) );

	OpCode	op_code;
	op_code._u32 = 0;
	op_code.op = OP_JAL;
	op_code.target = (p_target_location - mpAssemblyBuffer->GetCurrentAddress()) >> 2;
	AppendOp( op_code );

	// N.B. If we wanted to make this relocatable we'd need to store the offset here
	//u32		location( AppendOp( op_code ) );
	//mpCodeBuffer->AppendJumpFixup( location, label );

	// Stuff a nop in the delay slot
	op_code._u32 = 0;
	AppendOp( op_code );
}

