/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "CodeBufferManager.h"

#include "DaedMathUtil.h"

#include "Debug\DBGConsole.h"

#include "CodeGeneratorX86.h"
#include "CodeGeneratorPSP.h"

class CCodeBufferManagerPSP : public CCodeBufferManager
{
public:
	CCodeBufferManagerPSP()
		:	mpBuffer( NULL )
		,	mBufferPtr( 0 )
		,	mBufferSize( 0 )
	{
	}

	virtual bool				Initialise();
	virtual void				Finalise();

	virtual CCodeGenerator *	StartNewBlock();
	virtual void				FinaliseCurrentBlock( CCodeGenerator * p_code_gen );

private:
	
	u8	*						mpBuffer;
	u32							mBufferPtr;
	u32							mBufferSize;

private:
	CAssemblyBuffer				mAssemblyBuffer;
};

//*****************************************************************************
//
//*****************************************************************************
CCodeBufferManager *	CCodeBufferManager::Create()
{
	return new CCodeBufferManagerPSP;
}

//*****************************************************************************
//
//*****************************************************************************
bool	CCodeBufferManagerPSP::Initialise()
{
	mpBuffer = NULL;

	mBufferPtr = 0;
	mBufferSize = 0;

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeBufferManagerPSP::Finalise()
{
	if (mpBuffer != NULL)
	{
		DAEDALUS_ASSERTMSG( "How come we have a buffer on the PSP?" );
		mpBuffer = NULL;
	}
}

//*****************************************************************************
//
//*****************************************************************************
CCodeGenerator * CCodeBufferManagerPSP::StartNewBlock()
{
	// Round up to 16 byte boundry
	mBufferPtr = AlignPow2( mBufferPtr, 16 );

	// This is a bit of a hack. We assume that no single entry will generate more than 
	// 32k of storage. If there appear to be problems with this assumption, this
	// value can be enlarged
	if (mBufferPtr + 32768 > mBufferSize)
	{
		// Increase by 1MB
		LPVOID pNewAddress;

		mBufferSize += 1024 * 1024;
		pNewAddress = NULL;
		if (pNewAddress == 0)
		{
			DBGConsole_Msg(0, "SR Buffer allocation failed"); // maybe this should be an abort?
		}
		else
		{
			DBGConsole_Msg(0, "Allocated %dMB of storage for dynarec buffer", mBufferSize / (1024*1024));
		}

	}

	mAssemblyBuffer.SetBuffer( mpBuffer + mBufferPtr );
	return new CCodeGeneratorPSP( &mAssemblyBuffer );
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeBufferManagerPSP::FinaliseCurrentBlock( CCodeGenerator * p_code_gen )
{
	p_code_gen->Finalise();

	mBufferPtr += mAssemblyBuffer.GetSize();
}
