/*
Copyright (C) 2001,2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#if _MSC_VER > 1000
#pragma once
#endif

#ifndef CODEGENERATOR_H_
#define CODEGENERATOR_H_

struct	SRegisterStats;
class	CDynarecCode;

#include "Core\R4300.h"

// XXXX EMipsReg isn't just the *target* register type...
#include "DynarecTargetPSP.h"


// XXXX For GenerateCompare_S/D
#define FLAG_SWAP			0x100
#define FLAG_C_LT		(0x41)					// jne-   le
#define FLAG_C_LE		(FLAG_SWAP|0x41)		// je -   ! gt
#define FLAG_C_EQ		(FLAG_SWAP|0x40)		// je -   ! eq


class CCodeGenerator
{
	public:
									CCodeGenerator() {}
		virtual						~CCodeGenerator() {}

		virtual void				Finalise() = 0;

		virtual	void				AnalyseStats( const SRegisterStats & stats ) = 0;

		virtual void *				GetEntryPoint() const = 0;
		virtual u32					GetCompiledCodeSize() const = 0;

		virtual	void				GeneratePrologue() = 0;
		virtual	void				GenerateEpilogue( u32 pc ) = 0;

		virtual	void				SetVar( u32 * p_var, u32 value ) = 0;

		virtual	void				CleanupForExit() = 0;

		virtual void				PreEmitCachedRegCheck( const SRegisterStats & stats ) = 0;
		virtual void				PostEmitCachedRegCheck( const SRegisterStats & stats, bool optimised ) = 0;

		virtual void				GenerateBlockEndBranch( CDynarecCode * p_target, u32 num_ops ) = 0;
		virtual	void				GenerateBlockEnd() = 0;

		virtual	void				GenerateGenericR4300( bool optimise, OpCode op_code, CPU_Instruction pF ) = 0;

		virtual	void				GenerateMFHI( EMipsReg rd ) = 0;
		virtual	void				GenerateMFLO( EMipsReg rd ) = 0;

		virtual	void				GenerateMULTU( EMipsReg rs, EMipsReg rt ) = 0;
		virtual	void				GenerateADDU( EMipsReg rd, EMipsReg rs, EMipsReg rt ) = 0;
		virtual	void				GenerateSUBU( EMipsReg rd, EMipsReg rs, EMipsReg rt ) = 0;
		virtual	void				GenerateAND( EMipsReg rd, EMipsReg rs, EMipsReg rt ) = 0;
		virtual	void				GenerateOR( EMipsReg rd, EMipsReg rs, EMipsReg rt ) = 0;
		virtual	void				GenerateXOR( EMipsReg rd, EMipsReg rs, EMipsReg rt ) = 0;
		virtual	void				GenerateNOR( EMipsReg rd, EMipsReg rs, EMipsReg rt ) = 0;
		virtual	void				GenerateSLT( EMipsReg rd, EMipsReg rs, EMipsReg rt ) = 0;
		virtual	void				GenerateSLTU( EMipsReg rd, EMipsReg rs, EMipsReg rt ) = 0;

		virtual void				GenerateADDIU( EMipsReg rt, EMipsReg rs, s32 data ) = 0;
		virtual void				GenerateANDI( EMipsReg rt, EMipsReg rs, u32 data ) = 0;
		virtual void				GenerateORI( EMipsReg rt, EMipsReg rs, u32 data ) = 0;
		virtual void				GenerateXORI( EMipsReg rt, EMipsReg rs, u32 data ) = 0;
		virtual void				GenerateLUI( EMipsReg rt, s32 data ) = 0;
		virtual void				GenerateSLTI( EMipsReg rt, EMipsReg rs, s32 data ) = 0;
		virtual void				GenerateSLTIU( EMipsReg rt, EMipsReg rs, s32 data ) = 0;
		virtual void				GenerateSLL( EMipsReg rd, EMipsReg rt, u32 sa ) = 0;
		virtual void				GenerateSRL( EMipsReg rd, EMipsReg rt, u32 sa ) = 0;
		virtual void				GenerateSRA( EMipsReg rd, EMipsReg rt, u32 sa ) = 0;

		virtual void				GenerateLB( EMipsReg rt, EMipsReg base, s32 offset ) = 0;
		virtual void				GenerateLBU( EMipsReg rt, EMipsReg base, s32 offset ) = 0;
		virtual void				GenerateLH( EMipsReg rt, EMipsReg base, s32 offset ) = 0;
		virtual void				GenerateLHU( EMipsReg rt, EMipsReg base, s32 offset ) = 0;
		virtual void				GenerateLW( EMipsReg rt, EMipsReg base, s32 offset ) = 0;
		virtual	void				GenerateLWC1( u32 ft, EMipsReg base, s32 offset ) = 0;

		virtual	void				GenerateSB( EMipsReg rt, EMipsReg base, s32 offset ) = 0;
		virtual	void				GenerateSH( EMipsReg rt, EMipsReg base, s32 offset ) = 0;
		virtual	void				GenerateSW( EMipsReg rt, EMipsReg base, s32 offset ) = 0;
		virtual	void				GenerateSWC1( u32 ft, EMipsReg base, s32 offset ) = 0;

		virtual	void				GenerateMFC1( EMipsReg rt, u32 fs ) = 0;
		virtual	void				GenerateMTC1( u32 fs, EMipsReg rt ) = 0;
		virtual	void				GenerateCFC1( EMipsReg rt, u32 fs ) = 0;
		virtual	void				GenerateCTC1( u32 fs, EMipsReg rt ) = 0;

		virtual	void				GenerateBEQ( EMipsReg rs, EMipsReg rt, u32 target ) = 0;
		virtual	void				GenerateBNE( EMipsReg rs, EMipsReg rt, u32 target ) = 0;
		virtual	void				GenerateBLEZ( EMipsReg rs, u32 target ) = 0;
		virtual	void				GenerateBGTZ( EMipsReg rs, u32 target ) = 0;

		virtual	void				GenerateADD_S( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateSUB_S( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateMUL_S( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateDIV_S( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateSQRT_S( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateMOV_S( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateNEG_S( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCompare_S( u32 fs, u32 ft, u16 flag ) = 0;

		virtual	void				GenerateADD_D( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateSUB_D( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateMUL_D( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateDIV_D( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateSQRT_D( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateMOV_D( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateNEG_D( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCompare_D( u32 fs, u32 ft, u16 flag ) = 0;

		virtual	void				GenerateCVT_S_W( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCVT_S_D( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCVT_D_W( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCVT_D_S( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCVT_W_S( u32 fd, u32 fs ) = 0;
};

#endif // CODEGENERATOR_H_
