/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "CodeGeneratorPSP.h"

#include "Core\R4300_Regs.h"

#include "OSHLE\ultra_R4300.h"

#define FUNCTION_NAME	"?"		// __FUNCTION__ not working??

//*****************************************************************************
//
//*****************************************************************************
CCodeGeneratorPSP::CCodeGeneratorPSP( CAssemblyBuffer * p_buffer )
:	CCodeGenerator( )
,	CAssemblyWriterPSP( p_buffer )
{
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::Finalise()
{
	SetAssemblyBuffer( NULL );
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::AnalyseStats( const SRegisterStats & stats )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void *	CCodeGeneratorPSP::GetEntryPoint() const
{
	return GetAssemblyBuffer()->GetStartAddress();
}

//*****************************************************************************
//
//*****************************************************************************
u32	CCodeGeneratorPSP::GetCompiledCodeSize() const
{
	return GetAssemblyBuffer()->GetSize();
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::GeneratePrologue()
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
// EPILOGUE - set up g_pPCMemBase before returning
//*****************************************************************************
void CCodeGeneratorPSP::GenerateEpilogue( u32 pc )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetVar( u32 * p_var, u32 value )
{
	u32		address( reinterpret_cast< u32 >( p_var ) );
	u16		hi_bits( (u16)( address >> 16 ) );
	u16		lo_bits( (u16)( address ) );

	//
	//	Move up
	//
	if(lo_bits >= 0x8000)
	{
		hi_bits++;
	}

	s32		long_offset( address - ((s32)hi_bits<<16) );

	DAEDALUS_ASSERT( long_offset >= SHRT_MIN && long_offset <= SHRT_MAX, "OFfset is out of range!" );

	s16		offset( (s16)long_offset );

	DAEDALUS_ASSERT( ((s32)hi_bits<<16) + offset == (s32)address, "Incorrect address calculation" );

	LoadConstant( REG_t0, value );
	LUI( REG_v0, hi_bits );
	SW( REG_v0, lo_bits, REG_t0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::PreEmitCachedRegCheck( const SRegisterStats & stats )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::PostEmitCachedRegCheck( const SRegisterStats & stats, bool optimised)
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}


//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::CleanupForExit()
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBlockEnd()
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBlockEndBranch( CDynarecCode * p_target, u32 num_ops )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

/* A typical function
0000161c <_Z18R4300_Special_ADDUv>:
    161c:	3c030000 	lui			v1,0x0
    1620:	24630000 	addiu		v1,v1,0
    1624:	8c640d1c 	lw			a0,3356(v1)
    1628:	7c8522c0 	ext			a1,a0,0xb,0x5
    162c:	000528c0 	sll			a1,a1,0x3
    1630:	00a32821 	addu		a1,a1,v1
    1634:	7c822540 	ext			v0,a0,0x15,0x5
    1638:	000210c0 	sll			v0,v0,0x3
    163c:	00431021 	addu		v0,v0,v1
    1640:	7c842400 	ext			a0,a0,0x10,0x5
    1644:	000420c0 	sll			a0,a0,0x3
    1648:	00832021 	addu		a0,a0,v1
    164c:	8c420800 	lw			v0,2048(v0)
    1650:	8c840800 	lw			a0,2048(a0)
    1654:	00441021 	addu		v0,v0,a0
    1658:	00021fc3 	sra			v1,v0,0x1f
    165c:	aca20800 	sw			v0,2048(a1)
    1660:	03e00008 	jr			ra
    1664:	aca30804 	sw			v1,2052(a1)
*/

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateGenericR4300( bool optimise, OpCode op_code, CPU_Instruction pF )
{
	// XXXX optimise flag is ignored for now

	// Flush all fp registers before a generic call
	if (optimise)
	{
		//FlushMipsRegs();		// Valid flag updated after op "executed"
	}

	SetVar( &gCurrentOp._u32, op_code._u32 );
	JAL( reinterpret_cast< const void * >( pF ) );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFHI( EMipsReg rd )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFLO( EMipsReg rd )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMULTU( EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADDU( EMipsReg rd, EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUBU( EMipsReg rd, EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateAND( EMipsReg rd, EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateOR( EMipsReg rd, EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateXOR( EMipsReg rd, EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNOR( EMipsReg rd, EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLT( EMipsReg rd, EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTU( EMipsReg rd, EMipsReg rs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADDIU( EMipsReg rt, EMipsReg rs, s32 data )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateANDI( EMipsReg rt, EMipsReg rs, u32 data )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateORI( EMipsReg rt, EMipsReg rs, u32 data )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateXORI( EMipsReg rt, EMipsReg rs, u32 data )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLUI( EMipsReg rt, s32 data )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTI( EMipsReg rt, EMipsReg rs, s32 data )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTIU( EMipsReg rt, EMipsReg rs, s32 data )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLL( EMipsReg rd, EMipsReg rt, u32 sa )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRL( EMipsReg rd, EMipsReg rt, u32 sa )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRA( EMipsReg rd, EMipsReg rt, u32 sa )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLB( EMipsReg rt, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLBU( EMipsReg rt, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLH( EMipsReg rt, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLHU( EMipsReg rt, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLW( EMipsReg rt, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLWC1( u32 ft, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSB( EMipsReg rt, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSH( EMipsReg rt, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSW( EMipsReg rt, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSWC1( u32 ft, EMipsReg base, s32 offset )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFC1( EMipsReg rt, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMTC1( u32 fs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCFC1( EMipsReg rt, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCTC1( u32 fs, EMipsReg rt )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBEQ( EMipsReg rs, EMipsReg rt, u32 target )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBNE( EMipsReg rs, EMipsReg rt, u32 target )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBLEZ( EMipsReg rs, u32 target )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBGTZ( EMipsReg rs, u32 target )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADD_S( u32 fd, u32 fs, u32 ft )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUB_S( u32 fd, u32 fs, u32 ft )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMUL_S( u32 fd, u32 fs, u32 ft )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateDIV_S( u32 fd, u32 fs, u32 ft )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSQRT_S( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMOV_S( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNEG_S( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCompare_S( u32 fs, u32 ft, u16 flag )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADD_D( u32 fd, u32 fs, u32 ft )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUB_D( u32 fd, u32 fs, u32 ft )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMUL_D( u32 fd, u32 fs, u32 ft )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateDIV_D( u32 fd, u32 fs, u32 ft )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSQRT_D( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMOV_D( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNEG_D( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCompare_D( u32 fs, u32 ft, u16 flag )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_D_S( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_W_S( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_D_W( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_S_W( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_S_D( u32 fd, u32 fs )
{
	DAEDALUS_ERROR( "Shouldn't be here: " FUNCTION_NAME );
}

