/*
Copyright (C) 2001,2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#if _MSC_VER > 1000
#pragma once
#endif

#ifndef CODEGENERATORX86_H_
#define CODEGENERATORX86_H_

#include "CodeGenerator.h"
#include "AssemblyWriterX86.h"
#include "DynarecTargetX86.h"

class CCodeGeneratorX86 : public CCodeGenerator, public CAssemblyWriterX86
{
	public:
		CCodeGeneratorX86( CAssemblyBuffer * p_primary, CAssemblyBuffer * p_secondary );

		virtual void				Finalise();

		virtual	void				AnalyseStats( const SRegisterStats & stats );

		virtual void *				GetEntryPoint() const;
		virtual u32					GetCompiledCodeSize() const;

		virtual	void				GeneratePrologue();
		virtual	void				GenerateEpilogue( u32 pc );

		virtual	void				SetVar( u32 * p_var, u32 value );

		virtual	void				CleanupForExit();

		virtual void				PreEmitCachedRegCheck( const SRegisterStats & stats );
		virtual void				PostEmitCachedRegCheck( const SRegisterStats & stats, bool optimised );

		virtual void				GenerateBlockEndBranch( CDynarecCode * p_target, u32 num_ops );
		virtual	void				GenerateBlockEnd();

		virtual	void				GenerateGenericR4300( bool optimise, OpCode op_code, CPU_Instruction pF );

		virtual	void				GenerateMFHI( EMipsReg rd );
		virtual	void				GenerateMFLO( EMipsReg rd );

		virtual	void				GenerateMULTU( EMipsReg rs, EMipsReg rt );
		virtual	void				GenerateADDU( EMipsReg rd, EMipsReg rs, EMipsReg rt );
		virtual	void				GenerateSUBU( EMipsReg rd, EMipsReg rs, EMipsReg rt );
		virtual	void				GenerateAND( EMipsReg rd, EMipsReg rs, EMipsReg rt );
		virtual	void				GenerateOR( EMipsReg rd, EMipsReg rs, EMipsReg rt );
		virtual	void				GenerateXOR( EMipsReg rd, EMipsReg rs, EMipsReg rt );
		virtual	void				GenerateNOR( EMipsReg rd, EMipsReg rs, EMipsReg rt );
		virtual	void				GenerateSLT( EMipsReg rd, EMipsReg rs, EMipsReg rt );
		virtual	void				GenerateSLTU( EMipsReg rd, EMipsReg rs, EMipsReg rt );

		virtual void				GenerateADDIU( EMipsReg rt, EMipsReg rs, s32 data );
		virtual void				GenerateANDI( EMipsReg rt, EMipsReg rs, u32 data );
		virtual void				GenerateORI( EMipsReg rt, EMipsReg rs, u32 data );
		virtual void				GenerateXORI( EMipsReg rt, EMipsReg rs, u32 data );
		virtual void				GenerateLUI( EMipsReg rt, s32 data );
		virtual void				GenerateSLTI( EMipsReg rt, EMipsReg rs, s32 data );
		virtual void				GenerateSLTIU( EMipsReg rt, EMipsReg rs, s32 data );
		virtual void				GenerateSLL( EMipsReg rd, EMipsReg rt, u32 sa );
		virtual void				GenerateSRL( EMipsReg rd, EMipsReg rt, u32 sa );
		virtual void				GenerateSRA( EMipsReg rd, EMipsReg rt, u32 sa );

		virtual void				GenerateLB( EMipsReg rt, EMipsReg base, s32 offset );
		virtual void				GenerateLBU( EMipsReg rt, EMipsReg base, s32 offset );
		virtual void				GenerateLH( EMipsReg rt, EMipsReg base, s32 offset );
		virtual void				GenerateLHU( EMipsReg rt, EMipsReg base, s32 offset );
		virtual void				GenerateLW( EMipsReg rt, EMipsReg base, s32 offset );
		virtual	void				GenerateLWC1( u32 ft, EMipsReg base, s32 offset );

		virtual	void				GenerateSB( EMipsReg rt, EMipsReg base, s32 offset );
		virtual	void				GenerateSH( EMipsReg rt, EMipsReg base, s32 offset );
		virtual	void				GenerateSW( EMipsReg rt, EMipsReg base, s32 offset );
		virtual	void				GenerateSWC1( u32 ft, EMipsReg base, s32 offset );

		virtual	void				GenerateMFC1( EMipsReg rt, u32 fs );
		virtual	void				GenerateMTC1( u32 fs, EMipsReg rt );
		virtual	void				GenerateCFC1( EMipsReg rt, u32 fs );
		virtual	void				GenerateCTC1( u32 fs, EMipsReg rt );

		virtual	void				GenerateBEQ( EMipsReg rs, EMipsReg rt, u32 target );
		virtual	void				GenerateBNE( EMipsReg rs, EMipsReg rt, u32 target );
		virtual	void				GenerateBLEZ( EMipsReg rs, u32 target );
		virtual	void				GenerateBGTZ( EMipsReg rs, u32 target );

		virtual	void				GenerateADD_S( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateSUB_S( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateMUL_S( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateDIV_S( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateSQRT_S( u32 fd, u32 fs );
		virtual	void				GenerateMOV_S( u32 fd, u32 fs );
		virtual	void				GenerateNEG_S( u32 fd, u32 fs );
		virtual	void				GenerateCompare_S( u32 fs, u32 ft, u16 flag );

		virtual	void				GenerateADD_D( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateSUB_D( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateMUL_D( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateDIV_D( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateSQRT_D( u32 fd, u32 fs );
		virtual	void				GenerateMOV_D( u32 fd, u32 fs );
		virtual	void				GenerateNEG_D( u32 fd, u32 fs );
		virtual	void				GenerateCompare_D( u32 fs, u32 ft, u16 flag );

		virtual	void				GenerateCVT_S_W( u32 fd, u32 fs );
		virtual	void				GenerateCVT_S_D( u32 fd, u32 fs );
		virtual	void				GenerateCVT_D_W( u32 fd, u32 fs );
		virtual	void				GenerateCVT_D_S( u32 fd, u32 fs );
		virtual	void				GenerateCVT_W_S( u32 fd, u32 fs );


	private:
				EIntelReg			GetMIPSLoRegisterNoLoad(u32 mreg, EIntelReg scratch_to_use);
				EIntelReg			GetMIPSLoRegisterAndLoad(u32 mreg, EIntelReg scratch_to_use);

				enum EUpdateRegOptions
				{
					URO_HI_LEAVE,			// Leave hi bits alone
					URO_HI_SIGN_EXTEND,		// Sign extend from src
					URO_HI_CLEAR,			// Clear hi bits
					URO_HI_SET,				// Set hit bits
				};

				void				UpdateMIPSReg( u32 mreg, EIntelReg src, EUpdateRegOptions options, EIntelReg scratch );

				void				LoadMIPSLo(EIntelReg ireg, u32 mreg);
				void				LoadMIPSHi(EIntelReg ireg, u32 mreg);
				void				StoreMIPSLo(u32 mreg, EIntelReg ireg);
				void				StoreMIPSHi(u32 mreg, EIntelReg ireg);
				void				SetMIPSLo(u32 mreg, u32 data);
				void				SetMIPSHi(u32 mreg, u32 data);

				void				FlushMipsRegs();
				void				RestoreIntelRegs();

	// XXXX
	public:
				void				GenerateReadWriteAddress(u32 base, u32 offset, u32 xor_val, bool write = false, u32 rt = ~0, bool baseimm = false, bool base_already_loaded = false, EIntelReg destreg = EAX_CODE);

	public:
				bool				mSpCachedInESI;		// Is sp cached in ESI?
				u32					mSetSpPostUpdate;	// Set Sp base counter after this update

				CAssemblyBuffer *	mpPrimary;
				CAssemblyBuffer *	mpSecondary;
};

#endif // CODEGENERATORX86_H_
