#define SR_MATH_OPTIMISE_FLAG		pCode->dwOptimiseLevel < 1
//#define SR_MATH_OPTIMISE_FLAG		1

#define TEST_DISABLE_SR_MATH		//return FALSE;

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_MFHI(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	const u32 dwRD = op_code.rd;

	pCode->Stat_D(dwRD);
	
	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_MFHI);
	}
	else
	{
		pCode->dwNumOptimised++;

		pCode->mpCodeGenerator->GenerateMFHI( dwRD );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_MFLO(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	const u32 dwRD = op_code.rd;

	pCode->Stat_D(dwRD);
	
	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_MFLO);
	}
	else
	{
		pCode->dwNumOptimised++;

		pCode->mpCodeGenerator->GenerateMFLO( dwRD );
	}

	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_MULTU(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;

	pCode->Stat_S_S(dwRS, dwRT);
	
	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_MULTU);
	}
	else
	{
		pCode->dwNumOptimised++;

		pCode->mpCodeGenerator->GenerateMULTU( dwRS, dwRT );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_ADD(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	// Ignore exceptions - just use unsigned routine
	return SR_Emit_Special_ADDU(pCode, op_code, pdwFlags);
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_ADDU(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH

	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S_S(dwRD, dwRS, dwRT);

	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_ADDU);
	}
	else
	{
		pCode->dwNumOptimised++;

		if (dwRD == 0) return TRUE;	// Don't clear r0

		pCode->mpCodeGenerator->GenerateADDU( dwRD, dwRS, dwRT );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_SUB(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	// Ignore exceptions - just use unsigned routine
	return SR_Emit_Special_SUBU(pCode,op_code, pdwFlags);
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_SUBU(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH

	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S_S(dwRD, dwRS, dwRT);

	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_SUBU);
	}
	else
	{
		pCode->dwNumOptimised++;

		if (dwRD == 0) return TRUE;	// Don't clear r0

		pCode->mpCodeGenerator->GenerateSUBU( dwRD, dwRS, dwRT );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_AND(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH

	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S_S(dwRD, dwRS, dwRT);

	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_AND);
	}
	else
	{
		pCode->dwNumOptimised++;

		// Don't write to r0
		if (dwRD == 0) return TRUE;	// Don't clear r0

		pCode->mpCodeGenerator->GenerateAND( dwRD, dwRS, dwRT );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_OR(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S_S(dwRD, dwRS, dwRT);


	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_OR);
	}
	else
	{

		pCode->dwNumOptimised++;

		// Don't write to r0
		if (dwRD == 0) return TRUE;	// Don't clear r0

		pCode->mpCodeGenerator->GenerateOR( dwRD, dwRS, dwRT );
	}
	return TRUE;

}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_XOR(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S_S(dwRD, dwRS, dwRT);


	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_XOR);
	}
	else
	{
		pCode->dwNumOptimised++;

		// Don't write to r0
		if (dwRD == 0) return TRUE;	// Don't clear r0

		pCode->mpCodeGenerator->GenerateXOR( dwRD, dwRS, dwRT );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_NOR(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S_S(dwRD, dwRS, dwRT);
	
	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_NOR);
	}
	else
	{
		pCode->dwNumOptimised++;

		// Don't write to r0
		if (dwRD == 0) return TRUE;	// Don't clear r0

		pCode->mpCodeGenerator->GenerateNOR( dwRD, dwRS, dwRT );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_SLT(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH

	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S_S(dwRD, dwRS, dwRT);

	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_SLT);
	}
	else
	{
		pCode->dwNumOptimised++;

		if (dwRD == 0) return TRUE;		// Don't set r0

		pCode->mpCodeGenerator->GenerateSLT( dwRD, dwRS, dwRT );
	}
	
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_SLTU(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH
	const u32 dwRS = op_code.rs;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S_S(dwRD, dwRS, dwRT);

	if (SR_MATH_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_SLTU);
	}
	else
	{
		pCode->dwNumOptimised++;

		if (dwRD == 0) return TRUE;		// Don't set r0

		pCode->mpCodeGenerator->GenerateSLTU( dwRD, dwRS, dwRT );
	}

	return TRUE;
}
