#define SR_MATH_IMM_OPTIMISE_FLAG		pCode->dwOptimiseLevel < 1
//#define SR_MATH_IMM_OPTIMISE_FLAG		1

#define TEST_DISABLE_SR_MATH_IMM		//return FALSE;

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_ADDI(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM
	return SR_Emit_ADDIU(pCode, op_code, pdwFlags);
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_ADDIU(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM
	const u32 dwRT = op_code.rt;
	const u32 dwRS = op_code.rs;

	pCode->Stat_D_S(dwRT, dwRS);
			
	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_ADDIU);
	}
	else
	{
		s32 data = (s32)(s16)op_code.immediate;

		pCode->dwNumOptimised++;

		if (dwRT == REG_r0) return TRUE;

		pCode->mpCodeGenerator->GenerateADDIU( dwRT, dwRS, data );
	}

	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_ANDI(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM

	const u32 dwRT = op_code.rt;
	const u32 dwRS = op_code.rs;

	pCode->Stat_D_S(dwRT, dwRS);

	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_ANDI);
	}
	else
	{
		u32 data = (u32)(u16)op_code.immediate;

		pCode->dwNumOptimised++;

		if (dwRT == REG_r0) return TRUE;			// Skip attempts to write to r0

		pCode->mpCodeGenerator->GenerateANDI( dwRT, dwRS, data );
	}
	
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_ORI(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM

	const u32 dwRT = op_code.rt;
	const u32 dwRS = op_code.rs;

	pCode->Stat_D_S(dwRT, dwRS);

	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_ORI);
	}
	else
	{
		u32 data = (u32)(u16)op_code.immediate;

		pCode->dwNumOptimised++;

		if (dwRT == REG_r0) return TRUE;			// Skip attempts to write to r0

		pCode->mpCodeGenerator->GenerateORI( dwRT, dwRS, data );
	}

	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_XORI(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM
	const u32 dwRT = op_code.rt;
	const u32 dwRS = op_code.rs;

	pCode->Stat_D_S(dwRT, dwRS);

	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_XORI);
	}
	else
	{
		u32 data = (u32)(u16)op_code.immediate;

		pCode->dwNumOptimised++;

		if (dwRT == 0) return TRUE;

		pCode->mpCodeGenerator->GenerateXORI( dwRT, dwRS, data );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_LUI(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM

	const u32 dwRT = op_code.rt;

	pCode->Stat_D(dwRT);

	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_LUI);
	}
	else
	{
		pCode->dwNumOptimised++;
		
		s32 data = ((s32)(s16)op_code.immediate)<<16;

		if (dwRT == 0) return TRUE;		// Don't set reg0

		pCode->mpCodeGenerator->GenerateLUI( dwRT, data );
	}

	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_SLTI(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM

	const u32 dwRT = op_code.rt;
	const u32 dwRS = op_code.rs;

	pCode->Stat_D_S(dwRT, dwRS);

	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_SLTI);
	}
	else
	{
		s32 data = (s32)(s16)op_code.immediate;

		pCode->dwNumOptimised++;

		if (dwRT == 0) return TRUE;		// Don't set r0

		pCode->mpCodeGenerator->GenerateSLTI( dwRT, dwRS, data );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_SLTIU(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM
	const u32 dwRT = op_code.rt;
	const u32 dwRS = op_code.rs;

	pCode->Stat_D_S(dwRT, dwRS);

	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_SLTIU);
	}
	else
	{
		s32 data = (s32)(s16)op_code.immediate;

		pCode->dwNumOptimised++;

		if (dwRT == 0) return TRUE;		// Don't set r0

		pCode->mpCodeGenerator->GenerateSLTIU( dwRT, dwRS, data );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_SLL(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM

	// NOP!
	if (op_code._u32 == 0) 
	{
		pCode->dwNumOptimised++;
		return TRUE;
	}
	
	const u32 dwSA = op_code.sa;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S(dwRD, dwRT);

	// Skip NOPS

	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_SLL);
	}
	else
	{
		pCode->dwNumOptimised++;

		if (dwRD == 0) return TRUE;

		pCode->mpCodeGenerator->GenerateSLL( dwRD, dwRT, dwSA );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_SRL(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM
	const u32 dwSA = op_code.sa;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S(dwRD, dwRT);
	
	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_SRL);
	}
	else
	{
		pCode->dwNumOptimised++;

		if (dwRD == 0) return TRUE;	// Don't set reg0

		pCode->mpCodeGenerator->GenerateSRL( dwRD, dwRT, dwSA );
	}
	return TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
BOOL SR_Emit_Special_SRA(CDynarecCode *pCode, OpCode op_code, DWORD * pdwFlags)
{
TEST_DISABLE_SR_MATH_IMM
	const u32 dwSA = op_code.sa;
	const u32 dwRT = op_code.rt;
	const u32 dwRD = op_code.rd;

	pCode->Stat_D_S(dwRD, dwRT);

	if (SR_MATH_IMM_OPTIMISE_FLAG)
	{
		SR_Emit_Generic_R4300(pCode, op_code, R4300_Special_SRA);
	}
	else
	{
		pCode->dwNumOptimised++;

		if (dwRD == 0) return TRUE;	// Don't set reg0

		pCode->mpCodeGenerator->GenerateSRA( dwRD, dwRT, dwSA );
	}
	return TRUE;
}
