#include "StdAfx.h"
#include "InputManager.h"

#include <pspctrl.h>


//*****************************************************************************
//
//*****************************************************************************

class IInputManager : public CInputManager
{
	public:
		virtual ~IInputManager() {}

		virtual bool Initialise()
		{
			return true;
		}

		virtual void Finalise()
		{

		}

		virtual bool GetState( OSContPad pPad[4] );
};

//*****************************************************************************
//
//*****************************************************************************
bool IInputManager::GetState( OSContPad pPad[4] )
{
	// Clear the initial state
	for(u32 cont = 0; cont < 4; cont++)
	{
		pPad[cont].button = 0;
		pPad[cont].stick_x = 0;
		pPad[cont].stick_y = 0;
	}

	SceCtrlData pad;

	sceCtrlReadBufferPositive(&pad, 1);

	static const s32	N64_ANALOGUE_STICK_RANGE( 80 );
	static const s32	PSP_ANALOGUE_STICK_RANGE( 128 );

	//
	//	'Normalise' from 0..255 -> -128..+127
	//
	s32					normalised_x( pad.Lx - PSP_ANALOGUE_STICK_RANGE );
	s32					normalised_y( pad.Ly - PSP_ANALOGUE_STICK_RANGE );

	//
	//	Now scale from -128..+127 -> -80..+80 (y is inverted from PSP coords)
	//
	pPad[0].stick_x =  (normalised_x * N64_ANALOGUE_STICK_RANGE ) / PSP_ANALOGUE_STICK_RANGE;
	pPad[0].stick_y = -(normalised_y * N64_ANALOGUE_STICK_RANGE) / PSP_ANALOGUE_STICK_RANGE;

	if(pad.Buttons & PSP_CTRL_START)	{ pPad[0].button |= START_BUTTON; }
	if(pad.Buttons & PSP_CTRL_UP)		{ pPad[0].button |= U_JPAD; }
	if(pad.Buttons & PSP_CTRL_DOWN)		{ pPad[0].button |= D_JPAD; }
	if(pad.Buttons & PSP_CTRL_LEFT)		{ pPad[0].button |= L_JPAD; }
	if(pad.Buttons & PSP_CTRL_RIGHT)	{ pPad[0].button |= R_JPAD; }

	if(pad.Buttons & PSP_CTRL_CROSS)	{ pPad[0].button |= A_BUTTON; }
	if(pad.Buttons & PSP_CTRL_SQUARE)	{ pPad[0].button |= B_BUTTON; }

	if(pad.Buttons & PSP_CTRL_TRIANGLE)	{ pPad[0].button |= Z_TRIG; }
	if(pad.Buttons & PSP_CTRL_LTRIGGER)	{ pPad[0].button |= L_TRIG; }
	if(pad.Buttons & PSP_CTRL_RTRIGGER)	{ pPad[0].button |= R_TRIG; }

	//U_CBUTTONS
	//L_CBUTTONS
	//R_CBUTTONS
	//D_CBUTTONS

	// Stick

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
template<> bool	CSingleton< CInputManager >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new IInputManager();
	
	return true;
}

