/*
Copyright (C) 2001 StrmnNrmn

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	  
*/

#include "stdafx.h"
#include "TextureCache.h"
#include "ConvertImage.h"
#include "Core\Memory.h"

#include "RDP.h"

//*****************************************************************************
//
//*****************************************************************************
void ConvertTmemRGBA16( CTexture *pSurf, const TextureInfo & ti )
{
	DrawInfo dst;
	u32 x;
	u32 nFiddle;
		
	if (!pSurf->StartUpdate(&dst))
		return;

	u32 * pdwDst = (u32 *)(dst.lpSurface);

	u32 offset;
	u32 tile_width = gRDPTiles[ gTextureTile ].line * 8;		
	
	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = ( ti.TmemAddress << 3 ) + ( tile_width * y );

		if ((y&1) == 0)
			nFiddle = 0x2;
		else
			nFiddle = 0x2 | 0x4;					
		
		for (x = 0; x < ti.Width; x++)
		{
			u16 w = *(u16 *)&gTextureMemory[offset ^ nFiddle];
			
			pdwDst[x] = Convert555ToRGBA(w);
			
			// Increment word offset to point to the next pixel
			offset += 2;
		}

		pdwDst = (u32 *)((u8*)pdwDst + dst.lPitch);
	}

	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
//
//*****************************************************************************
void ConvertTmemRGBA32(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;

	u32 offset;
	u32 tile_width = gRDPTiles[ gTextureTile ].line * 8 * 2;		
	
	if (!pSurf->StartUpdate(&dst))
		return;
	
	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = ( ti.TmemAddress << 3 ) + ( tile_width * y );

		if ((y&1) == 0)
		{
			
			u8 *pD = (u8 *)dst.lpSurface + y * dst.lPitch;
			u8 *pS = (u8 *)&gTextureMemory[ offset ];
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				pD[0] = pS[1];	// Blue
				pD[1] = pS[2];	// Green
				pD[2] = pS[3];	// Red
				pD[3] = pS[0];	// Alpha
				pS+=4;
				pD+=4;
			}
		}
		else
		{
			
			u32 *pD = (u32 *)((u8 *)dst.lpSurface + y * dst.lPitch);
			u8 *pS = (u8 *)&gTextureMemory[ 0 ];

			for (u32 x = 0; x < ti.Width; x++)
			{
				*pD++ = MAKE_COLOUR_U32(pS[(offset+3)^0x4],
									  pS[(offset+2)^0x4],
									  pS[(offset+1)^0x4],
									  pS[(offset+0)^0x4]);
				
				offset += 4;
			}
		}
	}

	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
// E.g. Dear Mario text
// Copy, Score etc
//*****************************************************************************
void ConvertTmemIA4(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;

	if (!pSurf->StartUpdate(&dst))
		return;
	
	u32 offset;
	u32 tile_width = gRDPTiles[ gTextureTile ].line * 8;		
	
	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = ( ti.TmemAddress << 3 ) + ( tile_width * y );

		u8 *pD = (u8 *)dst.lpSurface + y * dst.lPitch;
		
		// For odd lines, swap words too
		if ((y&1) == 0)
			nFiddle = 0x3;
		else
			nFiddle = 0x7;
				
		// Do two pixels at a time
		for (u32 x = 0; x < ti.Width; x+=2)
		{
			u8 b = gTextureMemory[offset ^ nFiddle];
			
			// Even
			*pD++ = TwoToEight[(b & 0xc0) >> 6];
			*pD++ = TwoToEight[(b & 0xc0) >> 6];
			*pD++ = TwoToEight[(b & 0xc0) >> 6];
			*pD++ = TwoToEight[(b & 0x30) >> 4];	
			// Odd
			*pD++ = TwoToEight[(b & 0x0c) >> 2];
			*pD++ = TwoToEight[(b & 0x0c) >> 2];
			*pD++ = TwoToEight[(b & 0x0c) >> 2];
			*pD++ = TwoToEight[(b & 0x03)     ];
			
			offset++;
		}
		
	}

	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
// E.g Mario's head textures
//*****************************************************************************
void ConvertTmemIA8(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;

	if (!pSurf->StartUpdate(&dst))
		return;

	u32 offset = ti.TmemAddress<<3;
	
	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = (offset+7)&~7;

		// For odd lines, swap words too
		if ((y&1) == 0)
			nFiddle = 0x3;
		else
			nFiddle = 0x7;
		
		u8 *pD = (u8 *)dst.lpSurface + y * dst.lPitch;
		
		for (u32 x = 0; x < ti.Width; x++)
		{
			u8 b = gTextureMemory[offset ^ nFiddle];
			
			*pD++ = FourToEight[(b & 0xf0)>>4];
			*pD++ = FourToEight[(b & 0xf0)>>4];
			*pD++ = FourToEight[(b & 0xf0)>>4];
			*pD++ = FourToEight[(b & 0x0f)   ];
			
			offset++;
		}
	}		

	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
// E.g. camera's clouds, shadows
//*****************************************************************************
void ConvertTmemIA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	if (!pSurf->StartUpdate(&dst))
		return;
	
	u32 offset;
	u32 tile_width = gRDPTiles[ gTextureTile ].line * 8;		
	//u32 texture_width = (ti.Width+1)/2;
	
	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = ( ti.TmemAddress << 3 ) + ( tile_width * y );
		u8 *pD = (u8 *)dst.lpSurface + y * dst.lPitch;
		
		if ((y&1) == 0)
			nFiddle = 0x2;
		else
			nFiddle = 0x4 | 0x2;
				
		for (u32 x = 0; x < ti.Width; x++)
		{
			u16 w = *(u16 *)&gTextureMemory[offset^nFiddle];
			
			*pD++ = (u8)(w >> 8);
			*pD++ = (u8)(w >> 8);
			*pD++ = (u8)(w >> 8);
			*pD++ = (u8)(w & 0xFF);
			
			offset += 2;
		}
	}		
	
	pSurf->EndUpdate(&dst);
}



//*****************************************************************************
// Used by MarioKart
//*****************************************************************************
void ConvertTmemI4(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	if (!pSurf->StartUpdate(&dst))
		return;

	u32 offset;
	u32 tile_width = gRDPTiles[ gTextureTile ].line * 8;		
	//u32 texture_width = (ti.Width+1)/2;
	
	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = ( ti.TmemAddress << 3 ) + ( tile_width * y );
		
		u8 *pD = (u8 *)dst.lpSurface + y * dst.lPitch;
		
		// For odd lines, swap words too
		if ((y&1) == 0)
			nFiddle = 0x3;
		else
			nFiddle = 0x7;
		
		for (u32 x = 0; x < ti.Width; x+=2)
		{
			u8 b = gTextureMemory[offset ^ nFiddle];
		
			// Even
			*pD++ = FourToEight[(b & 0xF0)>>4];	// Other implementations seem to or in (b&0xF0)>>4
			*pD++ = FourToEight[(b & 0xF0)>>4]; // why?
			*pD++ = FourToEight[(b & 0xF0)>>4];
			*pD++ = FourToEight[(b & 0xF0)>>4];	
			// Odd
			*pD++ = FourToEight[(b & 0x0F)];
			*pD++ = FourToEight[(b & 0x0F)];
			*pD++ = FourToEight[(b & 0x0F)];
			*pD++ = FourToEight[(b & 0x0F)];
			
			offset++;
		}	
	}		

	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
// Used by MarioKart
//*****************************************************************************
void ConvertTmemI8(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	if (!pSurf->StartUpdate(&dst))
		return;
	
	u32 offset;
	u32 tile_width = gRDPTiles[ gTextureTile ].line * 8;		

	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = ( ti.TmemAddress << 3 ) + ( tile_width * y );
		u8 *pD = (u8 *)dst.lpSurface + y * dst.lPitch;

		if ((y&1) == 0)
			nFiddle = 0x3;
		else
			nFiddle = 0x7;
				
		for (u32 x = 0; x < ti.Width; x++)
		{
			//u8 b = pSrc[dwByteOffset ^ nFiddle];
			u8 b = gTextureMemory[offset ^ nFiddle];
			
			*pD++ = b;
			*pD++ = b;
			*pD++ = b;
			*pD++ = b;		// Alpha not 255?
			
			offset++;
		}
	}	

	pSurf->EndUpdate(&dst);
}


//*****************************************************************************
// Used by Starfox intro
//*****************************************************************************
void ConvertTmemCI4_RGBA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;

	u16 * pPal = (u16 *)&gTextureMemory[ ti.TmemPalAddress << 3 ];

	if (!pSurf->StartUpdate(&dst))
		return;

	u32 offset;
	u32 tile_width = gRDPTiles[ gTextureTile ].line * 8;		
	
	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = ( ti.TmemAddress << 3 ) + ( tile_width * y );

		if ((y&1) == 0)
			nFiddle = 0x3;
		else
			nFiddle = 0x7;
		
		u32 * pD = (u32 *)((u8 *)dst.lpSurface + y * dst.lPitch);
				
		for (u32 x = 0; x < ti.Width; x+=2)
		{
			u8 b = gTextureMemory[offset ^ nFiddle];
			
			u8 bhi = (b&0xf0)>>4;
			u8 blo = (b&0x0f);
			
			pD[0] = Convert555ToRGBA(pPal[bhi^0x1]);	// Remember palette is in different endian order!
			pD[1] = Convert555ToRGBA(pPal[blo^0x1]);	// Remember palette is in different endian order!
			pD+=2;
			
			offset++;
		}
	}	
	

	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
// Used by Starfox intro
//*****************************************************************************
void ConvertTmemCI4_IA16(CTexture *pSurf, const TextureInfo & ti)
{
return;
	DrawInfo dst;
	u32 nFiddle;
	
	u8 * pSrc = g_pu8RamBase + ti.Address;
	u16 * pPal = (u16 *)&gTextureMemory[ ti.TmemPalAddress << 3 ];

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Top)
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			
			u32 * pD = (u32 *)((u8 *)dst.lpSurface + y * dst.lPitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left / 2);
			
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ nFiddle];
				
				u8 bhi = (b&0xf0)>>4;
				u8 blo = (b&0x0f);
				
				pD[0] = ConvertIA16ToRGBA(pPal[bhi^0x1]);	// Remember palette is in different endian order!
				pD[1] = ConvertIA16ToRGBA(pPal[blo^0x1]);	// Remember palette is in different endian order!
				pD+=2;
				
				dwByteOffset++;
			}
		}	
		
	}
	else
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			u32 * pD = (u32 *)((u8 *)dst.lpSurface + y * dst.lPitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left / 2);
			
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ 0x3];
				
				u8 bhi = (b&0xf0)>>4;
				u8 blo = (b&0x0f);
				
				pD[0] = ConvertIA16ToRGBA(pPal[bhi^0x1]);	// Remember palette is in different endian order!
				pD[1] = ConvertIA16ToRGBA(pPal[blo^0x1]);	// Remember palette is in different endian order!
				pD+=2;
				
				dwByteOffset++;
			}
		}	
		
	}
	pSurf->EndUpdate(&dst);
}



//*****************************************************************************
// Used by MarioKart for Cars etc
//*****************************************************************************
void ConvertTmemCI8_RGBA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	u16 * pPal = (u16 *)&gTextureMemory[ ti.TmemPalAddress << 3 ];

	if (!pSurf->StartUpdate(&dst))
		return;
	
	u32 offset;
	u32 tile_width = gRDPTiles[ gTextureTile ].line * 8;		
	
	for (u32 y = 0; y < ti.Height; y++)
	{
		offset = ( ti.TmemAddress << 3 ) + ( tile_width * y );
		u32 *pD = (u32 *)((u8 *)dst.lpSurface + y * dst.lPitch);
		
		if ((y&1) == 0)
			nFiddle = 0x3;
		else
			nFiddle = 0x7;
			
		for (u32 x = 0; x < ti.Width; x++)
		{
			u8 b = gTextureMemory[offset ^ nFiddle];
			
			*pD++ = Convert555ToRGBA(pPal[b^0x1]);	// Remember palette is in different endian order!
			
			offset++;
		}
	}	
		
		
	pSurf->EndUpdate(&dst);
	
}



//*****************************************************************************
// Used by MarioKart for Cars etc
//*****************************************************************************
void ConvertTmemCI8_IA16(CTexture *pSurf, const TextureInfo & ti)
{
return;
	DrawInfo dst;
	u32 nFiddle;
	
	u8 * pSrc = g_pu8RamBase + ti.Address;
	u16 * pPal = (u16 *)&gTextureMemory[ ti.TmemPalAddress << 3 ];

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Top)
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			u32 *pD = (u32 *)((u8 *)dst.lpSurface + y * dst.lPitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				u8 b = pSrc[dwByteOffset ^ nFiddle];
				
				*pD++ = ConvertIA16ToRGBA(pPal[b^0x1]);	// Remember palette is in different endian order!
				
				dwByteOffset++;
			}
		}	
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			u32 *pD = (u32 *)((u8 *)dst.lpSurface + y * dst.lPitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				u8 b = pSrc[dwByteOffset ^ 0x3];
				
				*pD++ = ConvertIA16ToRGBA(pPal[b^0x1]);	// Remember palette is in different endian order!
				
				dwByteOffset++;
			}
		}
	}
	pSurf->EndUpdate(&dst);
}


