/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __TEXTURECACHE_H__
#define __TEXTURECACHE_H__

#include "Texture.h"
#include "DaedCritSect.h"
#include "DaedSingleton.h"

typedef struct _TextureInfo
{
	u32		TmemAddress;
	u32		TmemPalAddress;

	u32 Address;		// Corresponds to dwAddress in Tile
	u32 Format;		// e.g. RGBA, IA
	u32 Size;		// e.g. 16bpp
	s32 Left;
	s32 Top;
	u32 Width;
	u32 Height;
	u32 Pitch;
	u32 TLutFmt;

	BOOL Swapped;			// Are odd lines word swapped?

	// or bool operator==(,,) ?
	BOOL IsEqual(const _TextureInfo & ti) const
	{
		return ( memcmp(this, &ti, sizeof(_TextureInfo)) == 0 );
	}

} TextureInfo;

class CTextureCache : public CCritSect, public CSingleton< CTextureCache >
{
	public:
		virtual ~CTextureCache() {};

		virtual void PurgeOldTextures() = 0;
		virtual void DropTextures() = 0;

		virtual void SetDumpTextures( BOOL bDumpTextures ) = 0;
		virtual BOOL GetDumpTextures( ) const = 0;

		virtual void DisplayStats() = 0;

		virtual CTexture * GetTexture( TextureInfo * pti ) = 0;

		virtual bool InvalidateDeviceObjects() = 0;
};

#endif	// __TEXTURECACHE_H__
