/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "PspRomSelector.h"

#include <psptypes.h>
#include <pspkernel.h>
#include <pspdebug.h>
#include <pspctrl.h>
#include <pspdisplay.h>
#include <pspgu.h>
#include <pspgum.h>

#include <functional>

#include "PSPGraphics\FFXMaths.h"
#include "PSPGraphics\Vector2.h"
#include "PSPGraphics\Vector3.h"

#include "Core\ROM.h"
#include "Interface\IniFile.h"
#include "Interface\RomDB.h"

#include "PSPGraphics\GraphicsContext.h"
#include "PSPGraphics\DrawText.h"
#include "PSPGraphics\NativeTexture.h"

#define GL_TRUE                           1
#define GL_FALSE                          0

static const u32 COLOUR_RED = 0xff0000ff;
static const u32 COLOUR_GREEN = 0xff00ff00;
static const u32 COLOUR_BLUE = 0xffff0000;
static const u32 COLOUR_BLACK = 0xff000000;
static const u32 COLOUR_WHITE = 0xffffffff;
static const u32 COLOUR_TEXT_WHITE = 0xffeeeeee;
static const u32 DEBUG_SCREEN_WIDTH = 67;

// XXXX
TCHAR		g_szDaedalusRomsDir[MAX_PATH+1] = "Roms/";

static const char * const		BACKGROUND_FILENAME = "Resources/background.raw";
static const u32				BACKGROUND_WIDTH = 480;
static const u32				BACKGROUND_HEIGHT = 272;

static const u32				TEXT_AREA_TOP = 55;
static const u32				TEXT_AREA_LEFT = 40;
static const u32				TEXT_AREA_WIDTH = 400;
static const u32				TEXT_AREA_HEIGHT = 176;
static const u32				ROW_HEIGHT = 16;

bool gDoShadow = true;


struct BackgroundVtx
{
    v2	t0;
    v3	pos;
};

DAEDALUS_STATIC_ASSERT( sizeof(BackgroundVtx) == 20 );

#define BACKGROUND_VERTEX_FLAGS (GU_TEXTURE_32BITF|GU_VERTEX_32BITF )

// XXXX
extern float	GetElapsedTime();
extern void		ResetElapsedTime();
extern LPSTR	PathFindFileName(LPCSTR pszPath);


//*************************************************************************************
//
//*************************************************************************************
struct SRomInfo
{
	std::string		mFilename;

	RomID			mRomID;
	u32				mRomSize;
	ECicType		mCicType;

	RomSettings		mSettings;

	SRomInfo( const char * filename )
		:	mFilename( filename )
	{
		memset( &mSettings, 0, sizeof( mSettings ) );

		if ( ROM_GetRomID( filename, &mRomID, &mRomSize, &mCicType ) )
		{
			strcpy( mSettings.szGameName, PathFindFileName( filename ) );
			if ( !CIniFile::Get()->GetSettings( mRomID, &mSettings ) )
			{
				// Create new entry, add
				ROM_ResetSettings( &mSettings );
				strcpy( mSettings.szComment, "No Comment" );

				//
				// We want to get the "internal" name for this rom from the header
				// Failing that, use the filename
				//
				if ( !ROM_GetRomName( filename, mSettings.szGameName ) )
				{
					memset( mSettings.szGameName, 0, sizeof( mSettings.szGameName ) );
					strncpy( mSettings.szGameName, PathFindFileName( filename ), sizeof( mSettings.szGameName ) );
				}
				CIniFile::Get()->SetSettings( mRomID, mSettings );
			}
		}
		else
		{
			strcpy( mSettings.szGameName, "Can't get rom info" );
		}

	}
};


//*************************************************************************************
//
//*************************************************************************************
class IPspRomSelector : public CPspRomSelector
{
		typedef std::vector<SRomInfo*>	RomInfoList;
	public:

		IPspRomSelector();
		~IPspRomSelector();

		virtual void				SelectRom( char * p_filename );

	private:
				void				EnumerateRoms(RomInfoList & roms);
				void				RedrawRomsList();
				void				ChooseRom( char * p_filename );
				void				Render();

				void				RenderBackground();

	private:
		RomInfoList					mRomsList;
		u32							mCurrentSelection;
		CNativeTexture *			mpBackgroundTexture;
		s32							mCurrentScrollOffset;
		float						mSelectionAccumulator;
};

// XXXX If this is added as a member var seem to get random bus errors!!
SceIoDirent					gDirEntry;

// XXXX Need to reevaluate whether this safety buffer is needed anymore..
static u8					gBytes[128];

//*************************************************************************************
//
//*************************************************************************************
CPspRomSelector *	CPspRomSelector::Create()
{
	return new IPspRomSelector;
}

//*************************************************************************************
//
//*************************************************************************************
IPspRomSelector::IPspRomSelector()
:	mCurrentSelection( 0 )
,	mpBackgroundTexture( NULL )
,	mCurrentScrollOffset( 0 )
,	mSelectionAccumulator( 0 )
{
	mpBackgroundTexture = CNativeTexture::CreateFromRaw( BACKGROUND_FILENAME, BACKGROUND_WIDTH, BACKGROUND_HEIGHT );
}

//*************************************************************************************
//
//*************************************************************************************
IPspRomSelector::~IPspRomSelector()
{
	SAFE_RELEASE( mpBackgroundTexture );

	for(RomInfoList::iterator it = mRomsList.begin(); it != mRomsList.end(); ++it)
	{
		SRomInfo *	p_rominfo( *it );

		delete p_rominfo;
	}
	mRomsList.clear();
}

//*************************************************************************************
//
//*************************************************************************************
bool	CFileSystem_FindFirstFile( const char * path, SceUID & handle )
{
	handle = sceIoDopen( path );

	return ( handle >= 0 );
}

//*************************************************************************************
//
//*************************************************************************************
bool	CFileSystem_FindNextFile( SceIoDirent & dir_entry, SceUID handle )
{
	DAEDALUS_ASSERT( handle >= 0, "Cannot search with invalid directory handle" );

	const s32	error_code( sceIoDread( handle, &dir_entry ) );

	DAEDALUS_ASSERT( error_code >= 0, "Error reading directory entry" );

	if ( error_code > 0 )
	{
		return true;
	}

	return false;
}

//*************************************************************************************
//
//*************************************************************************************
bool	CFileSystem_FindCloseFile( SceUID handle )
{
	DAEDALUS_ASSERT( handle >= 0, "Trying to close an invalid directory handle" );

	return ( sceIoDclose( handle ) >= 0 );
}


//*************************************************************************************
//
//*************************************************************************************
static bool SortByGameName( const SRomInfo * a, const SRomInfo * b )
{
	return stricmp( a->mSettings.szGameName, b->mSettings.szGameName ) < 0;
}


//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::EnumerateRoms(RomInfoList & roms)
{
	roms.clear();

	SceUID				dir_handle;
	if(CFileSystem_FindFirstFile( g_szDaedalusRomsDir, dir_handle ))
	{
		//pspDebugScreenPrintf( "found first\n" );

		while(CFileSystem_FindNextFile( gDirEntry, dir_handle ))
		{
			//pspDebugScreenPrintf( "found next\n" );

			const char * rom_filename( gDirEntry.d_name );
			const char * last_period( strrchr( rom_filename, '.' ) );
			if(last_period != NULL)
			{
				if(stricmp(last_period, ".v64") == 0 ||
				   stricmp(last_period, ".z64") == 0 ||
				   stricmp(last_period, ".rom") == 0 ||
				   stricmp(last_period, ".zip") == 0)
				{
					//pspDebugScreenPrintf( "found a rom\n" );

					std::string		full_path;

					full_path = g_szDaedalusRomsDir;
					full_path += rom_filename;

					//pspDebugScreenPrintf( "Adding: %s\n", full_path.c_str() );

					SRomInfo *	p_rom_info = new SRomInfo( full_path.c_str() );

					roms.push_back( p_rom_info );
				}
			}
		}

		CFileSystem_FindCloseFile( dir_handle );
	}
}

//*************************************************************************************
//
//*************************************************************************************
void IPspRomSelector::RedrawRomsList()
{
	s32		x = TEXT_AREA_LEFT;
	s32		y = TEXT_AREA_TOP + mCurrentScrollOffset;

	sceGuEnable(GU_SCISSOR_TEST);
	sceGuScissor(TEXT_AREA_LEFT, TEXT_AREA_TOP, TEXT_AREA_LEFT+TEXT_AREA_WIDTH, TEXT_AREA_TOP+TEXT_AREA_HEIGHT);

	for(u32 i = 0; i < mRomsList.size(); ++i)
	{
		const char * p_gamename( mRomsList[ i ]->mSettings.szGameName );
		u32		colour;

		//
		// Check if this entry would be onscreen
		//
		if((y+ROW_HEIGHT) >= s32(TEXT_AREA_TOP) && y < s32(TEXT_AREA_TOP + TEXT_AREA_HEIGHT))
		{
			if(i == mCurrentSelection)
			{
				colour = COLOUR_RED;
			}
			else
			{
				colour = COLOUR_TEXT_WHITE;
			}

			static s32 COLOUR_SHADOW = 0x80000000;
			static s32 COLOUR_SHADOW2 = 0x50000000;

			CDrawText::Render( x+1, y+1, p_gamename, COLOUR_SHADOW );

			if(gDoShadow)
			{
				CDrawText::Render( x+1, y+0, p_gamename, COLOUR_SHADOW2 );
				CDrawText::Render( x+0, y+1, p_gamename, COLOUR_SHADOW2 );
				CDrawText::Render( x+2, y+2, p_gamename, COLOUR_SHADOW2 );
			}
			CDrawText::Render( x, y, p_gamename, colour );
		}
		y += 16;
	}

	// Restore scissoring
	sceGuScissor(0,0, 480,272);
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::RenderBackground()
{
	u32				num_verts( 2 );
	BackgroundVtx*	p_verts = (BackgroundVtx*)sceGuGetMemory(num_verts*sizeof(BackgroundVtx));

	sceGuDisable(GU_DEPTH_TEST);
	sceGuDepthMask( GL_TRUE );	// GL_TRUE to disable z-writes
	sceGuShadeModel( GU_FLAT );

	sceGuTexFilter(GU_NEAREST,GU_NEAREST);
	sceGuTexScale(1.0f,1.0f);
	sceGuTexOffset(0.0f,0.0f);

	if(mpBackgroundTexture != NULL)
	{
		mpBackgroundTexture->InstallTexture();
	}

	sceGuDisable(GU_ALPHA_TEST);
	sceGuDisable(GU_BLEND);
	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
	//sceGumOrtho(0, 480 +1, 272 +1, 0, 0.0f, -1.0f);

	v2		tex_uv0( (float)0, (float)0 );
	v2		tex_uv1( (float)BACKGROUND_WIDTH, (float)BACKGROUND_HEIGHT );

	p_verts[0].pos = v3( 0.0f, 0.0f, 0.0f );
	p_verts[0].t0 = v2( 0.0f, 0.0f );

	p_verts[1].pos = v3( (float)BACKGROUND_WIDTH, (float)BACKGROUND_HEIGHT, 0.0f );
	p_verts[1].t0 = v2( (float)BACKGROUND_WIDTH, (float)BACKGROUND_HEIGHT );

	sceGumDrawArray(GU_SPRITES,BACKGROUND_VERTEX_FLAGS|GU_TRANSFORM_2D,num_verts,NULL,p_verts);
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::Render()
{
	CGraphicsContext::Get()->BeginFrame();
	//CGraphicsContext::Get()->Clear( 0xff39A2F9, 0 );

	RenderBackground();

	RedrawRomsList();

	CGraphicsContext::Get()->EndFrame();

	CGraphicsContext::Get()->UpdateFrame( true );
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::ChooseRom( char * p_filename )
{
	SceCtrlData		pad;
	sceCtrlReadBufferPositive(&pad, 1);

	u32		oldButtons = pad.Buttons;

	static const float	SCROLL_RATE_PER_SECOND = 15.0f;		// 5 roms/second
	static const float	MAX_SCROLL_PER_FRAME = 2.0f;

	static const s32	STICK_DEADZONE = 20;

	ResetElapsedTime();

	// Simple rom chooser
	bool		rom_selected = false;
	while(!rom_selected)
	{
		Render();

		float		elapsed_time( GetElapsedTime() );

		sceCtrlReadBufferPositive(&pad, 1);

		if(pad.Buttons & PSP_CTRL_LTRIGGER)
		{
			mSelectionAccumulator += -SCROLL_RATE_PER_SECOND * elapsed_time;
		}
		if(pad.Buttons & PSP_CTRL_RTRIGGER)
		{
			mSelectionAccumulator += +SCROLL_RATE_PER_SECOND * elapsed_time;
		}

		s32		stick_y( pad.Ly - 128 );

		if(stick_y >= -STICK_DEADZONE && stick_y <= STICK_DEADZONE)
		{
			stick_y = 0;
		}

		float pad_y( float(stick_y) / 128.0f );

		mSelectionAccumulator += pad_y * SCROLL_RATE_PER_SECOND * elapsed_time; 

		if(oldButtons != pad.Buttons)
		{
			if(pad.Buttons & PSP_CTRL_UP)
			{
				if(mCurrentSelection > 0)
				{
					mCurrentSelection--;
				}
			}
			if(pad.Buttons & PSP_CTRL_DOWN)
			{
				if(mCurrentSelection < mRomsList.size() - 1)
				{
					mCurrentSelection++;
				}
			}

			if((pad.Buttons & PSP_CTRL_START) ||
			   (pad.Buttons & PSP_CTRL_CROSS))
			{
				rom_selected = true;
				if(mCurrentSelection < mRomsList.size())
				{
					strcpy( p_filename, mRomsList[ mCurrentSelection ]->mFilename.c_str() );
				}
			}
			oldButtons = pad.Buttons;
		}

		//
		//	Clamp the accumulator
		//
		if(mSelectionAccumulator > MAX_SCROLL_PER_FRAME)
		{
			mSelectionAccumulator = MAX_SCROLL_PER_FRAME;
		}
		else if(mSelectionAccumulator < -MAX_SCROLL_PER_FRAME)
		{
			mSelectionAccumulator = -MAX_SCROLL_PER_FRAME;
		}

		//
		//	Apply the selection accumulator
		//
		while(mSelectionAccumulator >= 1.0f)
		{
			if(mCurrentSelection < mRomsList.size() - 1)
			{
				mCurrentSelection++;
			}
			mSelectionAccumulator -= 1.0f;
		}
		while(mSelectionAccumulator <= -1.0f)
		{
			if(mCurrentSelection > 0)
			{
				mCurrentSelection--;
			}
			mSelectionAccumulator += 1.0f;
		}

		//
		//	Scroll to keep things in view
		//
		s32		current_selection_y = (mCurrentSelection * ROW_HEIGHT) + (ROW_HEIGHT/2) + mCurrentScrollOffset;

		s32		adjust_amount( (TEXT_AREA_HEIGHT/2) - current_selection_y );

		float d( 1.0f - powf(0.995f, elapsed_time * 1000.0f) );

		mCurrentScrollOffset += s32(float(adjust_amount) * d);
	}

	// Clear everything to black
	for( u32 i = 0; i < 2; ++i )
	{
		CGraphicsContext::Get()->BeginFrame();
		CGraphicsContext::Get()->Clear( true, true );
		CGraphicsContext::Get()->EndFrame();
		CGraphicsContext::Get()->UpdateFrame( true );
	}
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::SelectRom( char * p_filename )
{
	EnumerateRoms( mRomsList );

	sort( mRomsList.begin(), mRomsList.end(), SortByGameName );

	ChooseRom( p_filename );

	CRomDB::Get()->Reset();
}
