/*
Copyright (C) 20/02/2004 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "Synchroniser.h"

#include "Interface\MainWindow.h"

#include "Debug\DBGConsole.h"

#include "Core\CPU.h"

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser * CSynchroniser::mpSynchroniser( NULL );

//*****************************************************************************
//
//*****************************************************************************
class ISynchProducer : public CSynchroniser
{
	public:
		ISynchProducer( const CHAR * filename );
		~ISynchProducer();

		virtual ESynchResult	SynchData( const void * p_data, u32 length );
		virtual bool			IsOpen() const;

	private:
		FILE * const		mFile;
};

class ISynchConsumer : public CSynchroniser
{
	public:
		ISynchConsumer( const CHAR * filename );
		~ISynchConsumer();

		virtual ESynchResult	SynchData( const void * p_data, u32 length );
		virtual bool			IsOpen() const;

	private:
		static const u32		mMaxLength = 1024;
		u8						mBuffer[ mMaxLength ];
		FILE * const			mFile;
};

//*****************************************************************************
// Creation
//*****************************************************************************
CSynchroniser *	CSynchroniser::CreateProducer( const CHAR * filename )
{
	mpSynchroniser = new ISynchProducer( filename );
	return mpSynchroniser;
}

CSynchroniser *	CSynchroniser::CreateConsumer( const CHAR * filename )
{
	mpSynchroniser = new ISynchConsumer( filename );
	return mpSynchroniser;
}

void CSynchroniser::Destroy()
{
	SAFE_DELETE( mpSynchroniser );
}

//*****************************************************************************
//
//*****************************************************************************
void	CSynchroniser::HandleOutOfStorage()
{
	const CHAR *	p_message( "Out of storage for synchroniser" );
	CMainWindow::Get()->MessageBox( p_message );
	DBGConsole_Msg( 0, p_message );
	SAFE_DELETE( mpSynchroniser );
}

//*****************************************************************************
//
//*****************************************************************************
void	CSynchroniser::HandleOutOfInput()
{
	const CHAR *	p_message( "Out of input for synchroniser" );
	CMainWindow::Get()->MessageBox( p_message );
	DBGConsole_Msg( 0, p_message );	
	SAFE_DELETE( mpSynchroniser );
}

//*****************************************************************************
//
//*****************************************************************************
void	CSynchroniser::HandleOutOfSynch( const CHAR * msg )
{
	const CHAR *	p_message( DSPrintf( "Synchronisation Failed at 0x%08x: %s", g_dwPC, msg ) );
	CMainWindow::Get()->MessageBox( p_message );
	CPU_Halt( p_message );
	SAFE_DELETE( mpSynchroniser );
}

//*****************************************************************************
//
//*****************************************************************************
ISynchProducer::ISynchProducer( const CHAR * filename )
:	mFile( fopen( filename, "wb" ) )
{
}

//*****************************************************************************
//
//*****************************************************************************
ISynchProducer::~ISynchProducer()
{
	if ( mFile != NULL )
	{
		fclose( mFile );
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool	ISynchProducer::IsOpen() const
{
	return mFile != NULL;
}

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser::ESynchResult	ISynchProducer::SynchData( const void * p_data, u32 length )
{
	if ( mFile != NULL )
	{
		u32	written = fwrite( p_data, 1, length, mFile );

		return written == length ? SR_OK : SR_OUT_OF_STORAGE;
	}

	return SR_OUT_OF_STORAGE;
}


//*****************************************************************************
//
//*****************************************************************************
ISynchConsumer::ISynchConsumer( const CHAR * filename )
:	mFile( fopen( filename, "rb" ) )
{
}

//*****************************************************************************
//
//*****************************************************************************
ISynchConsumer::~ISynchConsumer()
{
	if ( mFile != NULL )
	{
		fclose( mFile );
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool	ISynchConsumer::IsOpen() const
{
	return mFile != NULL;
}

//*****************************************************************************
//
//*****************************************************************************
CSynchroniser::ESynchResult	ISynchConsumer::SynchData( const void * p_data, u32 length )
{
	if ( mFile != NULL )
	{
		if ( length > mMaxLength )
		{
			DAEDALUS_ASSERTMSG( "Trying to synch too many bytes" );
		}
		else
		{
			u32	read = fread( mBuffer, 1, length, mFile );

			if ( read == length )
			{
				return memcmp( mBuffer, p_data, length ) == 0 ? SR_OK : SR_OUT_OF_SYNCH;
			}
			else
			{
				return SR_OUT_OF_INPUT;
			}
		}
	}

	return SR_OUT_OF_INPUT;
}




