/*
 * PSP Software Development Kit - http://www.pspdev.org
 * -----------------------------------------------------------------------
 * Licensed under the BSD license, see LICENSE in PSPSDK root for details.
 *
 * Copyright (c) 2005 Jesper Svennevid
 */

#include "stdafx.h"

//namespace psp
//{
#include <psptypes.h>
#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include <pspgu.h>
#include <pspgum.h>
#include <pspctrl.h>
#include <pspsdk.h>
#include <psprtc.h>
#include <psppower.h>
#include <sys\stat.h>


#include "Debug\DBGConsole.h"
#include "Debug\Debug.h"
#include "Interface\RomDB.h"
#include "Interface\IniFile.h"
#include "Core\Memory.h"
#include "Core\PIF.h"
#include "DynaRec\DynaRec.h"
#include "Input\InputManager.h"
#include "Utility\Profiler.h"

#include "SysPSP\PspRomSelector.h"
#include "PSPGraphics\VideoMemoryManager.h"
#include "PSPGraphics\GraphicsContext.h"
#include "PSPGraphics\DrawText.h"
#include "PSPGraphics\DaedalusGraphics.h"
#include "PSPGraphics\TextureCache.h"

#include <stdarg.h>

TCHAR		g_szDaedalusExeDir[MAX_PATH+1] = "";
TCHAR		g_szDaedalusName[] = "Daedalus";
TCHAR		g_szDaedalusIniFile[MAX_PATH+1] = "daedalus.ini";
TCHAR		g_szDaedalusLogFile[MAX_PATH+1] = "daedalus.txt";
TCHAR		g_szDaedalusRomName[MAX_PATH+1] = "";

HINSTANCE	g_hInstance = NULL;

#define FAILED(hr)	(s32(hr) < 0)

static int  WaitForCPUThread(void);
static void DisplayRomsAndChoose();
static int DoMain();

#define BOOTLOADER_NAME "Daedalus"

#if defined( DAEDALUS_KERNEL_MODE )

PSP_MODULE_INFO( BOOTLOADER_NAME, 0x1000, 1, 1 );
PSP_MAIN_THREAD_ATTR( 0 );

#else

PSP_MODULE_INFO( BOOTLOADER_NAME, 0, 1, 1 );
PSP_MAIN_THREAD_ATTR( THREAD_ATTR_USER );

#endif


void * __cdecl operator new( size_t count )
{
	void * p_mem( malloc( count ) );
	if(p_mem == NULL)
	{
		char msg[ 1024 ];
		sprintf( msg, "Out of memory (operator new(%d)", count );
		DAEDALUS_ASSERTMSG( msg );
	}

	return p_mem;
}


void *__cdecl operator new[]( size_t count )
{
	void * p_mem( malloc( count ) );
	if(p_mem == NULL)
	{
		char msg[ 1024 ];
		sprintf( msg, "Out of memory (operator new[](%d)", count );
		DAEDALUS_ASSERTMSG( msg );
	}

	return p_mem;
}

void __cdecl operator delete[]( void * p_mem )
{
	if( p_mem != NULL )
	{
		free( p_mem );
	}
}

void __cdecl operator delete( void * p_mem )
{
	if( p_mem != NULL )
	{
		free( p_mem );
	}
}

// XXXX 
#define PSP_LINE_SIZE 512
#define PSP_PIXEL_FORMAT 3


void Daedalus_ForceDebugScreenToFront()
{
	//
	//	Urgh - make sure the debug screen is visible
	//
	void * p_vram_base = (void *) (0x40000000 | (u32) sceGeEdramGetAddr());
	sceDisplaySetFrameBuf((void *) p_vram_base, PSP_LINE_SIZE, PSP_PIXEL_FORMAT, PSP_DISPLAY_SETBUF_IMMEDIATE);
	//sceDisplaySetFrameBuf((void *) p_vram_base, PSP_LINE_SIZE, PSP_PIXEL_FORMAT, PSP_DISPLAY_SETBUF_NEXTFRAME);
}

#if defined( DAEDALUS_KERNEL_MODE )
void MyExceptionHandler(PspDebugRegBlock *regs)
{
	Daedalus_ForceDebugScreenToFront();
	pspDebugScreenPrintf("\n\n");
	pspDebugDumpException(regs);
}
/**
 * Function that is called from _init in kernelmode before the
 * main thread is started in usermode.
 */

__attribute__ ((constructor))
void loaderInit()
{
    pspKernelSetKernelPC();
	pspDebugScreenInit();
	pspDebugInstallKprintfHandler(NULL);
	pspDebugInstallErrorHandler(MyExceptionHandler);
	pspSdkInstallNoDeviceCheckPatch();
	pspSdkInstallNoPlainModuleCheckPatch();
}
#endif

/* Exit callback */
int exit_callback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();
	return 0;
}

/* Callback thread */
int CallbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}


extern "C"
{
int main(void)
{
	pspDebugScreenInit();

	pspDebugScreenPrintf( "Entering DoMain()\n" );
	DoMain();
	pspDebugScreenPrintf( "Exited DoMain()\n" );

	sceKernelExitGame();
	return 0;
}
}

int WaitForCPUThread(void)
{
	while ( true )
	{
		pspDebugScreenPrintf( "Sleeping!\n" );
		sceKernelSleepThread();
	}

	return 0;
}

extern "C" { void _DisableFPUExceptions(); } 

int DoMain()
{
	int nResult;
	BOOL bResult;

// Disable for profiling
//	srand(time(0));

	pspDebugScreenPrintf( "Disabling FPU Exceptions\n" );

	_DisableFPUExceptions();

	SetupCallbacks();

	pspDebugScreenPrintf( "Callbacks done\n" );

	if (scePowerSetClockFrequency(333, 333, 166) == 0)
	{
		pspDebugScreenPrintf( "CPU set to 333MHz\n" );
	}
	else
	{
		pspDebugScreenPrintf( "Could not set CPU to 333MHz\n" );
	}

	pspDebugScreenPrintf( "vram base address: %p\n", sceGeEdramGetAddr() );
	pspDebugScreenPrintf( "vram size: %d bytes\n", sceGeEdramGetSize() );

    //setup Pad
    sceCtrlSetSamplingCycle(0);
    sceCtrlSetSamplingMode(PSP_CTRL_MODE_ANALOG);
	
	//
	// Read configuration before doing anything major
	//
//	ReadConfiguration();

	pspDebugScreenPrintf( "In DoMain()\n" );

	//
	// In ReleaseWithLogging mode, open the logfile for writing
	//
	#ifdef DAEDALUS_LOG
	hr = Debug_InitLogging();
	if (FAILED(hr))
		return 1;
	#endif

	pspDebugScreenPrintf( "DebugConsole up\n" );

	if (!CDebugConsole::Create())
		return 1;

	DBGConsole_Msg( 0, "ROMDB create\n" );
	
	//
	// Create and load the RomDB
	//
	if (!CRomDB::Create())
		return 1;

	DBGConsole_Msg( 0, "ROMDB init\n" );

	char romdb_filename[ MAX_PATH + 1 ];
	PathCombine( romdb_filename, g_szDaedalusExeDir, "rom.db" );
	bResult = CRomDB::Get()->OpenDB( romdb_filename );
	if ( !bResult )
	{
		// That's ok - it probably didn't exist
	}

	//
	// Create and load the IniFile singleton
	//
	DBGConsole_Msg( 0, "IniFile create\n" );

	if (!CIniFile::Create())
		return 1;

	DBGConsole_Msg( 0, "IniFile init\n" );

	bResult = CIniFile::Get()->OpenIniFile( g_szDaedalusIniFile );
	if (!bResult)
	{
		//MessageBox(NULL, "Unable to read IniFile from disk", g_szDaedalusName, MB_OK);
		//return 1;
	}

	//
	// Init memory before Console (which tries to display
	// things that are refs to memory)
	//
	DBGConsole_Msg( 0, "Memory init\n" );

	if (!Memory_Init())
		return 1;

	//
	// Create the console if it's enabled. Don't care about failures
	//
	CDebugConsole::Get()->EnableConsole( g_DaedalusConfig.ShowDebug ? true : false );
//	DisplayDisclaimer();
//	DisplayConfig();

	//
	// Create the Profiler singleton
	//
	if (!CProfiler::Create())
		return 1;

	//
	// Create the input manager BEFORE the window is created and initialised
	//
	DBGConsole_Msg( 0, "InputManager init\n" );

	if (!CInputManager::Create())
		return 1;

	if(!CInputManager::Get()->Initialise())
		return 1;

	DBGConsole_Msg( 0, "Controller init\n" );

	if (!CController::Create())
		return 1;

	//
	// Create the main window singleton
	//
//	hr = CMainWindow::Create();
//	if (FAILED(hr))
//		return 1;
	if(!CVideoMemoryManager::Create())
		return false;

	if(!CGraphicsContext::Create())
		return false;

	CDrawText::Initialise();
	
	DisplayRomsAndChoose();

	//
	// Init the static recompiler
	//
	if (!DynaRec_Init(10))
//	if (!DynaRec_Init(30000))
		return 1;

	//
	// Create/Show the main window
	//
//	hr = CMainWindow::Get()->CreateWnd( 0 );
//	if (FAILED(hr))
//		return 1;

	DBGConsole_Msg( 0, "Loading ROM!\n" );

	ROM_LoadFile( g_szDaedalusRomName );

	DBGConsole_Msg( 0, "Running Main\n" );

	// Wrap the main processing in an exception handler
	// This should at least let us save the config/inifile
	// if something goes wrong.
	{
		nResult = WaitForCPUThread();
	}

	//
	// Write current config out to the registry
	//
//	WriteConfiguration();

	//	
	// Clear the recompiler
	//
	DynaRec_Fini();

	//
	// Unload any loaded rom and cleanup
	//
	ROM_Unload();

	//
	// Turn off the debug console
	// 
	CDebugConsole::Get()->EnableConsole( false );

	Memory_Fini();

	//
	// Destroy the singletons
	//
//	CMainWindow::Destroy();
	CController::Destroy();
	CInputManager::Destroy();
	CProfiler::Destroy();

	#ifdef DAEDALUS_LOG
	Debug_FinishLogging();
	#endif

	//
	// Destroy the inifile.
	// Will write ini file if changed
	//
	CIniFile::Destroy();

	//
	// Destroy / Commit the rom database
	//
	CRomDB::Get()->Commit();
	CRomDB::Destroy();

	CDebugConsole::Destroy();

	return nResult;
}



///////////////////////////////////////////////////////////////////////////////////////

const char gPathSeparator( '/' );

LPSTR   PathFindExtension(LPCSTR pszPath)
{
	return strrchr( pszPath, '.' );
}


BOOL	PathAppend(LPSTR pszPath, LPCSTR pMore)
{
	u32		len( strlen(pszPath) );

	if(len > 0)
	{
		if(pszPath[len-1] != gPathSeparator)
		{
			pszPath[len] = gPathSeparator;
			pszPath[len+1] = '\0';
			len++;
		}
	}
	strcat( pszPath, pMore );
	return TRUE;
}


BOOL	PathIsDirectory(LPCSTR pszPath)
{	
	struct stat		s;

	if(stat( pszPath, &s ) == 0)
	{
		if(s.st_mode & _IFDIR)
		{
			return TRUE;
		}
	}

	return FALSE;
}


LPSTR	PathRemoveBackslash(LPSTR pszPath)
{
	u32 len = strlen( pszPath );
	if ( len > 0 )
	{
		if ( pszPath[ len - 1 ] == gPathSeparator )
		{
			pszPath[ len - 1 ] = '\0';
		}
		return &pszPath[ len - 1 ];
	}
	return NULL;
}


BOOL	PathRemoveFileSpec(LPSTR pszPath)
{
	char * last_slash = strrchr( pszPath, gPathSeparator );
	if ( last_slash )
	{
		*last_slash = '\0';
		return TRUE;
	}
	return FALSE;
}


LPSTR	PathCombine(LPSTR szDest, LPCSTR lpszDir, LPCSTR lpszFile)
{
	strcpy( szDest, lpszDir );
	PathAppend( szDest, lpszFile );
	return szDest;
}

void	PathStripPath(LPSTR pszPath)
{
	// ?
}
void	PathRemoveExtension(LPSTR pszPath)
{
	char * ext = PathFindExtension( pszPath );
	if ( ext )
	{
		*ext = '\0';
	}
}
BOOL	PathAddExtension(LPSTR pszPath, LPCSTR pszExt)
{
	PathAppend( pszPath, pszExt );
	return TRUE;
}

BOOL	PathFileExists(LPCSTR pszPath)
{
	FILE * fh = fopen( pszPath, "rb" );
	if ( fh )
	{
		fclose( fh );
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

LPSTR	PathFindFileName(LPCSTR pszPath)
{
	char * last_slash = strrchr( pszPath, gPathSeparator );
	if ( last_slash )
	{
		return last_slash + 1;
	}
	else
	{
		return NULL;
	}
}

BOOL	StrTrim(LPSTR psz, LPCSTR pszTrimChars)
{
	u32 num_trims = strlen( pszTrimChars );
	char * pin = psz;
	char * pout = psz;
	bool found;
	while ( *pin )
	{
		char c = *pin;

		found = false;
		for ( u32 i = 0; i < num_trims; i++ )
		{
			if ( pszTrimChars[ i ] == c )
			{
				// Skip
				found = true;
				break;
			}

		}

		if ( found )
		{
			pin++;
		}
		else
		{
			// Copy
			*pout++ = *pin++;
		}
	}
	*pout = '\0';
	return TRUE;
}

BOOL	CreateDirectory(LPCSTR pszPath, void * )
{
	return sceIoMkdir( pszPath, 0777 ) == 0;
}


BOOL	MoveFile ( LPCSTR lpExistingFileName, LPCSTR lpNewFileName )
{

	return sceIoRename(lpExistingFileName, lpNewFileName) >= 0;

}

BOOL	DeleteFile ( LPCSTR lpFileName )
{
	return sceIoRemove( lpFileName ) >= 0;
}


FILE * const			gInvalidFileHandle = NULL;

FILE * Debug_CreateDumpFile( LPCTSTR szFileName )
{
	return fopen( szFileName, "w" );
}

void Debug_CloseDumpFile( FILE * fh )
{
	if(fh)
	{
		fclose(fh);
	}
}

void Debug_DumpLine (FILE * hFile, LPCTSTR szFormat, va_list va)
{
	if(hFile != NULL)
	{
		vfprintf( hFile, szFormat, va );
		fputs( "\n", hFile );
	}
}



///////////////////////////////////////////////////////////////////////////////////////



static bool gTimerInitialised( false );
static u64	gLastTime( 0 );
static f32	gFreqInv( 1.0f );


float	GetElapsedTime()
{
	u64			now;
	NTiming::GetPreciseTime( &now );

	if(!gTimerInitialised)
	{
		u64		freq;
		NTiming::GetPreciseFrequency( &freq );
		gLastTime = now;
		gFreqInv = 1.0f / f32( freq );
		gTimerInitialised = true;
	}

	u64			elapsed_ticks( now - gLastTime );
	gLastTime = now;

	return f32(elapsed_ticks) * gFreqInv;
}

void ResetElapsedTime()
{
	u64			now;
	NTiming::GetPreciseTime( &now );

	gLastTime = now;
}


bool	gDumpNextScreen = false;

bool	gMenuActive = false;




#ifdef DAEDALUS_KERNEL_MODE
bool	gProfileNextFrame = false;
bool	gWasProfilingFrame = false;
bool	gShowProfilerResults = false;
#endif

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
static bool	gShowTextureCacheUsage = false;
extern bool gSoftwareClipping;
extern bool gDebugDisplayList;
#include "PSPGraphics\DLParser.h"
#endif

void HandleEndOfFrame()
{
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if(gDebugDisplayList)
		return;
#endif

#ifdef DAEDALUS_KERNEL_MODE
	//
	//	Update the profiling stats
	//
	if(gWasProfilingFrame)
	{
		// Show results
		pspDebugProfilerDisable();
		gShowProfilerResults = true;
		gWasProfilingFrame = false;

		// Drop back into the menu
		gMenuActive = true;
	}
#endif

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if( gShowTextureCacheUsage )
	{
		pspDebugScreenSetXY(0, 0);
		CTextureCache::Get()->DisplayStats();
	}
#endif

	if( gDumpNextScreen )
	{
		CGraphicsContext::Get()->DumpScreenShot();
		gDumpNextScreen = false;
	}

	//
	//	Figure out how long the last frame took
	//
	float		elapsed_time( GetElapsedTime() );
	float		framerate( 0.0f );
	if(elapsed_time > 0)
	{
		framerate = 1.0f / elapsed_time;
	}

	//
	//	Enter the debug menu as soon as select is newly pressed
	//
    SceCtrlData pad;

	static u32 oldButtons = 0;

	sceCtrlReadBufferPositive(&pad, 1);
	if(oldButtons != pad.Buttons)
	{
		if(pad.Buttons & PSP_CTRL_SELECT)
		{
			gMenuActive = true;
		}
	}

	oldButtons = pad.Buttons;

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	bool	show_clipping_mode( false );
	bool	dump_displaylist( false );
	bool	debug_displaylist( false );
#endif

	if(gMenuActive)
	{
		bool menu_button_pressed( false );

		Daedalus_ForceDebugScreenToFront();

		// Remain paused until the Select button is pressed again
		while(!menu_button_pressed)
		{
			sceDisplayWaitVblankStart();
			pspDebugScreenSetXY(0, 0);
			pspDebugScreenPrintf( "Framerate: %.2f\n", framerate );

#ifdef DAEDALUS_KERNEL_MODE
			if(gShowProfilerResults)
			{
				pspDebugProfilerPrint();
			}
			if(gProfileNextFrame)
			{
				pspDebugScreenPrintf( "Will profile next frame on resume\n" );
			}
#endif
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
			if( gShowTextureCacheUsage )
			{
				CTextureCache::Get()->DisplayStats();
			}

			if(show_clipping_mode)
			{
				pspDebugScreenPrintf( "Software clipping %s\n", gSoftwareClipping ? "enabled" : "disabled" );
			}
			if(dump_displaylist)
			{
				pspDebugScreenPrintf( "Will dump next displaylist on resume\n" );
			}
			if(debug_displaylist)
			{
				pspDebugScreenPrintf( "Will debug next displaylist on resume" );
			}
#endif
			if(gDumpNextScreen)
			{
				pspDebugScreenPrintf( "Will grab screenshot of next frame on resume\n" );
			}

			sceCtrlReadBufferPositive(&pad, 1);
			if(oldButtons != pad.Buttons)
			{
				if(pad.Buttons & PSP_CTRL_SELECT)
				{
					menu_button_pressed = true;
				}
#ifdef DAEDALUS_KERNEL_MODE
				if(pad.Buttons & PSP_CTRL_CIRCLE)
				{
					gProfileNextFrame = !gProfileNextFrame;
				}
#endif
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
				if(pad.Buttons & PSP_CTRL_TRIANGLE)
				{
					gShowTextureCacheUsage = !gShowTextureCacheUsage;
				}
				if(pad.Buttons & PSP_CTRL_LTRIGGER)
				{
					gSoftwareClipping = !gSoftwareClipping;
					show_clipping_mode = true;
				}
				if(pad.Buttons & PSP_CTRL_CROSS)
				{
					dump_displaylist = !dump_displaylist;
				}
				if(pad.Buttons & PSP_CTRL_RTRIGGER)
				{
					debug_displaylist = true;
				}
#endif
				// Need to do this every frame
				if(pad.Buttons & PSP_CTRL_SQUARE)
				{
					gDumpNextScreen = !gDumpNextScreen;
				}
			}

			oldButtons = pad.Buttons;
		}

		gMenuActive = false;
	}

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if(dump_displaylist)
	{
		DLParser_DumpNextDisplayList();
	}
	if(debug_displaylist)
	{
		Graphics_ExecuteCommand( DAEDALUS_GFX_DEBUGDL, NULL );
	}
#endif


	//
	//	Reset the elapsed time to avoid glitches when we restart
	//
	ResetElapsedTime();

	//
	//	Check whether to profile the next frame
	//
#ifdef DAEDALUS_KERNEL_MODE
	gShowProfilerResults = false;
	if(gProfileNextFrame)
	{
		gProfileNextFrame = false;
		gWasProfilingFrame = true;

		pspDebugProfilerClear();
		pspDebugProfilerEnable();
	}
#endif
}

void DisplayRomsAndChoose()
{
	CPspRomSelector *	p_selector( CPspRomSelector::Create() );

	if(p_selector != NULL)
	{
		p_selector->SelectRom( g_szDaedalusRomName );

		delete p_selector;
	}
}

