// stdafx.h : include file for standard system include files,
//  or project specific include files that are used frequently, but
//      are changed infrequently
//

#ifndef PRECOMPILED_H_
#define PRECOMPILED_H_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//*****************************************************************************
//	These are our various compilation options
//*****************************************************************************

//#define DAEDALUS_ENABLE_SYNCHRONISATION
#define DAEDALUS_ENABLE_DYNAREC
#define DAEDALUS_ZLIB_SUPPORT				// Undefine this if you don't have a local install of zlib. If you undefine this, you will need to remove unzip.c from the project too.
#define DAEDALUS_USE_EXCEPTIONS
#define DAEDALUS_BREAKPOINTS_ENABLED

#ifdef DAEDALUS_PSP
#undef DAEDALUS_ENABLE_DYNAREC
#undef DAEDALUS_USE_EXCEPTIONS
#undef DAEDALUS_BREAKPOINTS_ENABLED
#endif

//*****************************************************************************
//
//*****************************************************************************
#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers


#ifdef _MSC_VER
// BEGIN ADDED BY Lkb - 8/jun/2001
#pragma warning(disable : 4786)	// identifier was truncated to '255' characters in the debug information ?
// END ADDED BY Lkb

#define __PRETTY_FUNCTION__ __FUNCTION__

#endif

#if !defined( DAEDALUS_XB ) && defined( DAEDALUS_RELEASE_BUILD )
#define DAEDALUS_ALIGN_REGISTERS
#endif

#ifdef _DEBUG

#define NODEFAULT		DAEDALUS_ASSERTMSG( "No default - we shouldn't be here" )

#else

#define NODEFAULT		__assume( 0 )

#endif

#ifdef DAEDALUS_PSP



typedef char TCHAR;
typedef char * LPTSTR;
typedef char * LPSTR;
typedef const char * LPCSTR;
typedef const char * LPCTSTR;
typedef void * LPVOID;
#define MAX_PATH 260
#define __cdecl        //No meaning on the PSP __attribute__((cdecl))
#define __fastcall     //No meaning on the PSP __attribute__((fastcall))
#define __stdcall      //No meaning on the PSP __attribute__((stdcall))
#define __forceinline	inline
#define __assume( x )
#define __declspec( x )

typedef long BOOL;
#define FALSE 0
#define TRUE 1

typedef int HRESULT;

typedef signed int INT_PTR;
typedef unsigned int UINT_PTR;
typedef unsigned int DWORD;
typedef char CHAR;

typedef unsigned char BYTE;
typedef short WORD;
typedef int* HWND;
typedef int* HINSTANCE;
typedef int* HANDLE;


LPSTR	PathCombine(LPSTR szDest, LPCSTR lpszDir, LPCSTR lpszFile);
LPSTR   PathFindExtension(LPCSTR pszPath);
BOOL	PathAppend(LPSTR pszPath, LPCSTR pMore);
BOOL	PathIsDirectory(LPCSTR pszPath);
LPSTR	PathRemoveBackslash(LPSTR pszPath);
BOOL	PathRemoveFileSpec(LPSTR pszPath);
BOOL	CreateDirectory(LPCSTR pszPath, void * );
BOOL	MoveFile ( LPCSTR lpExistingFileName, LPCSTR lpNewFileName );
BOOL	DeleteFile ( LPCSTR lpFileName );
void	PathStripPath(LPSTR pszPath);
void	PathRemoveExtension(LPSTR pszPath);
BOOL	PathAddExtension(LPSTR pszPath, LPCSTR pszExt);
BOOL	PathFileExists(LPCSTR pszPath);
BOOL	StrTrim(LPSTR psz, LPCSTR pszTrimChars);

BOOL Graphics_InitiateGFX ();
void Graphics_RomOpen();
HRESULT Graphics_ExecuteCommand ( const CHAR * pszCommand, void ** ppResult );
void	Graphics_RomClosed();
void Graphics_ProcessDList (void);

void Daedalus_ForceDebugScreenToFront();


#else
#include <atlbase.h>
extern CComModule _Module;
#include <atlcom.h>
#include <atlwin.h>
#include <windows.h>
#include <windowsx.h>
#include <winnt.h>			// For 32x32To64 multiplies
#include <commctrl.h>
#include <commdlg.h>
#include <shlwapi.h>		// PathCombine() etc
#include <shellapi.h>		// ShellExecute()
#include <shlobj.h>

#define DIRECTINPUT_VERSION         DIRECTINPUT_HEADER_VERSION
#include <dinput.h>
#include <oleauto.h> // BSTR
#include <crtdbg.h>
#endif

#include <stdio.h>
#include <math.h>			// For sqrt()
#include <stdlib.h>
#include <stddef.h>		// offsetof
#include <float.h>

#include <algorithm>
#include <vector>
#include <map>
// BEGIN ADDED BY Lkb - 8/jun/2001 - use stdlib iostream instead of deprecated ms iostream
#include <fstream>
#include <sstream>

// END ADDED BY Lkb - 8/jun/2001

#include <stdarg.h>


#include "Debug\DaedalusAssert.h"
#include "DaedalusTypes.h"

using namespace daedalus;

#ifdef DAEDALUS_PSP

inline u32 MAKE_COLOUR_U32( s32 r, s32 g, s32 b )
{
	return (0xff << 24) | ((b&255)<<16) | ((g&255)<<8) | (r&255);
}

inline u32 MAKE_COLOUR_U32( s32 r, s32 g, s32 b, s32 a )
{
	return ((a&255) << 24) | ((b&255)<<16) | ((g&255)<<8) | (r&255);
}

inline u8	COL_GETA( u32 abgr )
{
	return (abgr >> 24) & 255;
}
inline u8	COL_GETB( u32 abgr )
{
	return (abgr >> 16) & 255;
}
inline u8	COL_GETG( u32 abgr )
{
	return (abgr >>  8) & 255;
}
inline u8	COL_GETR( u32 abgr )
{
	return (abgr      ) & 255;
}

inline float COL_GETA_F( u32 abgr ) { return COL_GETA( abgr ) / 255.0f; }
inline float COL_GETB_F( u32 abgr ) { return COL_GETB( abgr ) / 255.0f; }
inline float COL_GETG_F( u32 abgr ) { return COL_GETG( abgr ) / 255.0f; }
inline float COL_GETR_F( u32 abgr ) { return COL_GETR( abgr ) / 255.0f; }


#endif

#ifndef SAFE_DELETE
#define SAFE_DELETE(p)  { if(p) { delete (p);     (p)=NULL; } }
#endif

#ifndef SAFE_RELEASE
#define SAFE_RELEASE(p)  { if(p) { (p)->Release();     (p)=NULL; } }
#endif

#ifndef SAFE_DELETEARRAY
#define SAFE_DELETEARRAY(p)  { if(p) { delete [](p);     (p)=NULL; } }
#endif

#define ARRAYSIZE(arr)   (sizeof(arr) / sizeof(arr[0]))

#define DAEDLAUS_VERSION_NO		0x008b
#define DAEDALUS_VERSION		"0.08b"
#define DAEDALUS_SITE			"http://daedalus.boob.co.uk/"

// DAEDALUS_RELEASE_BUILD is now only defined for "Win32 Release" configuration
// It is disabled in Win32 Development Release etc

extern const TCHAR * __cdecl DSPrintf( const TCHAR * fmt, ... );

#ifndef DAEDALUS_PSP
extern HINSTANCE	g_hInstance;
#endif
extern TCHAR		g_szDaedalusName[256];
extern TCHAR		g_szDaedalusExeDir[MAX_PATH+1];

#include "ConfigOptions.h"

#define ALIGNED_TYPE(type, alignval) type
#define ALIGNED_EXTERN(type, var, alignval) extern type var
#define ALIGNED_GLOBAL(type, var, alignval) type var
#define ALIGNED_MEMBER(type, var, alignval) type var

#if defined(DAEDALUS_W32) || defined(DAEDALUS_XB)

#ifdef _MSC_VER
#if defined(_MSC_FULL_VER) && (_MSC_FULL_VER >= 12008804) 
#undef ALIGNED_TYPE
#undef ALIGNED_GLOBAL
#undef ALIGNED_MEMBER
#define ALIGNED_TYPE(type, alignval) __declspec(align(alignval)) type
#define ALIGNED_GLOBAL(type, var, alignval) __declspec(align(alignval)) type var
#define ALIGNED_MEMBER(type, var, alignval) __declspec(align(alignval)) type var
#else
#pragma message("You should install Microsoft Visual C++ 6.0 Service Pack 5 and Processor Pack (both free of charge)")
#endif
#endif

#ifdef __GNUC__
// GCC has __attribute__((aligned)) but it complains that the maximum alignment is 16
#undef ALIGNED_GLOBAL
#undef ALIGNED_MEMBER
#define ALIGNED_GLOBAL(type, var, alignval) __asm__(".balign " #alignval ) type var;
#define ALIGNED_MEMBER(type, var, alignval) __attribute__((aligned(alignval))) type var;
#endif
#endif // defined(DAEDALUS_W32) || defined(DAEDALUS_XB)

// Pentium 4 has 64-byte cachelines
#define DATA_ALIGN	16
#define CACHE_ALIGN	64
#define PAGE_ALIGN	4096


// Windowsx.h type defines - hope these don't conflict with anything!
#define CheckBox_GetCheck(hwnd)			SendMessage(hwnd, BM_GETCHECK, 0,0)
#define CheckBox_SetCheck(hwnd, state)	SendMessage(hwnd, BM_SETCHECK, (WPARAM)state,0)

#define CheckBox_IsChecked( hwnd )		(CheckBox_GetCheck( (hwnd) ) == BST_CHECKED)

#endif // PRECOMPILED_H_
